@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,60 @@
  *
  **********************************************************************
  *
- * $Id: SQLAllocHandle.c,v 1.1.1.1 2001/10/17 16:40:03 lurcher Exp $
+ * $Id: SQLAllocHandle.c,v 1.13 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLAllocHandle.c,v $
+ * Revision 1.13  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.12  2007/12/17 13:13:03  lurcher
+ * Fix a couple of descriptor typo's
+ *
+ * Revision 1.11  2007/02/12 11:49:34  lurcher
+ * Add QT4 support to existing GUI parts
+ *
+ * Revision 1.10  2006/06/28 08:08:41  lurcher
+ * Add timestamp with timezone to Postgres7.1 driver
+ *
+ * Revision 1.9  2005/11/21 17:25:43  lurcher
+ * A few DM fixes for Oracle's ODBC driver
+ *
+ * Revision 1.8  2005/11/08 09:37:10  lurcher
+ * Allow the driver and application to have different length handles
+ *
+ * Revision 1.7  2005/10/06 08:50:58  lurcher
+ * Fix problem with SQLDrivers not returning first entry
+ *
+ * Revision 1.6  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.5  2003/02/25 13:28:21  lurcher
+ *
+ * Allow errors on the drivers AllocHandle to be reported
+ * Fix a problem that caused errors to not be reported in the log
+ * Remove a redundant line from the spec file
+ *
+ * Revision 1.4  2002/12/12 18:21:21  lurcher
+ *
+ * Fix bug where if a SQLAllocHandle in the driver failed, the driver manager
+ * seg faulted
+ *
+ * Revision 1.3  2002/08/12 13:17:52  lurcher
+ *
+ * Replicate the way the MS DM handles loading of driver libs, and allocating
+ * handles in the driver. usage counting in the driver means that dlopen is
+ * only called for the first use, and dlclose for the last. AllocHandle for
+ * the driver environment is only called for the first time per driver
+ * per application environment.
+ *
+ * Revision 1.2  2002/07/25 09:30:26  lurcher
+ *
+ * Additional unicode and iconv changes
+ *
  * Revision 1.1.1.1  2001/10/17 16:40:03  lurcher
  *
  * First upload to SourceForge
@@ -194,13 +245,14 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 #if defined ( COLLECT_STATS ) && defined( HAVE_SYS_SEM_H )
 #include "__stats.h"
 #include <uodbc_stats.h>
 #endif
 
-static char const rcsid[]= "$RCSfile: SQLAllocHandle.c,v $ $Revision: 1.1.1.1 $";
+static char const rcsid[]= "$RCSfile: SQLAllocHandle.c,v $ $Revision: 1.13 $";
 
 /*
  * connection pooling stuff
@@ -221,12 +273,20 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
     switch( handle_type )
     {
       case SQL_HANDLE_ENV:
+      case SQL_HANDLE_SENV:
         {
             DMHENV environment;
             char pooling_string[ 128 ];
 
-            if ( !output_handle )
+            if ( !output_handle ) 
+            {
                 return SQL_ERROR;
+            }
+
+            if ( input_handle )
+            {
+                return SQL_INVALID_HANDLE;
+            }
 
             /*
              * check connection pooling attributes
@@ -261,6 +321,8 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             environment -> state = STATE_E1;
             environment -> requested_version = requested_version;
+            environment -> version_set = !!requested_version;
+        	environment -> sql_driver_count = -1;
 
             /*
              * if SQLAllocEnv is called then it's probable that
@@ -330,9 +392,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY009, NULL, 
                         SQL_OV_ODBC3 );
 
-                thread_release( SQL_HANDLE_ENV, environment );
-
-                return function_return( environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             /*
@@ -351,9 +411,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         SQL_OV_ODBC3 );
 
-                thread_release( SQL_HANDLE_ENV, environment );
+                *output_handle = SQL_NULL_HDBC;
 
-                return function_return( environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             connection = __alloc_dbc();
@@ -369,9 +429,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                     ERROR_HY013, NULL,
                     environment -> requested_version );
 
-                thread_release( SQL_HANDLE_ENV, environment );
+                *output_handle = SQL_NULL_HDBC;
 
-                return function_return( environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             /*
@@ -389,9 +449,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
             connection -> cursors = SQL_CUR_DEFAULT;
             connection -> login_timeout = SQL_LOGIN_TIMEOUT_DEFAULT;
             connection -> login_timeout_set = 0;
-            connection -> auto_commit = 0;
-            connection -> auto_commit_set = 0;
-            connection -> auto_commit = 0;
+            connection -> auto_commit = SQL_AUTOCOMMIT_ON;
             connection -> auto_commit_set = 0;
             connection -> async_enable = 0;
             connection -> async_enable_set = 0;
@@ -417,6 +475,12 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
             connection -> stmt_attribute.list = NULL;
             connection -> save_attr = NULL;
 
+#ifdef HAVE_ICONV
+            connection -> iconv_cd_uc_to_ascii = (iconv_t)-1;
+            connection -> iconv_cd_ascii_to_uc = (iconv_t)-1;
+            strcpy( connection -> unicode_string, DEFAULT_ICONV_ENCODING );
+#endif
+
             *output_handle = (SQLHANDLE) connection;
 
             if ( log_info.log_flag )
@@ -491,9 +555,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY009, NULL, 
                         connection -> environment -> requested_version  );
 
-                thread_release( SQL_HANDLE_DBC, connection );
-
-                return function_return( connection , SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection , SQL_ERROR );
             }
 
             if ( connection -> state == STATE_C1 ||
@@ -510,9 +572,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_08003, NULL,
                         connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_DBC, connection );
+                *output_handle = SQL_NULL_HSTMT;
 
-                return function_return( connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             statement = __alloc_stmt();
@@ -528,16 +590,16 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY013, NULL,
                         connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_DBC, connection );
+                *output_handle = SQL_NULL_HSTMT;
 
-                return function_return( connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             /*
              * pass the call on
              */
 
-            if ( requested_version == SQL_OV_ODBC3 )
+            if ( requested_version >= SQL_OV_ODBC3 )
             {
                 if ( CHECK_SQLALLOCHANDLE( connection ))
                 {
@@ -574,9 +636,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                     __release_stmt( statement );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
+                    *output_handle = SQL_NULL_HSTMT;
 
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -616,9 +678,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                     __release_stmt( statement );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    *output_handle = SQL_NULL_HSTMT;
+                
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
 
@@ -632,9 +694,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                 if ( connection -> state == STATE_C4 )
                     connection -> state = STATE_C5;
 
-                connection -> statement_count ++;
-
-                statement -> connection = connection;
+                __register_stmt ( connection, statement );
 
                 *output_handle = (SQLHANDLE) statement;
 
@@ -649,7 +709,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                 if ( connection -> driver_act_ver == 3 &&
                         CHECK_SQLGETSTMTATTR( connection ))
                 {
-                    SQLHDESC desc;
+                    DRV_SQLHDESC desc;
 
                     /*
                      * ARD
@@ -683,12 +743,11 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             __release_stmt( statement );
 
-                            thread_release( SQL_HANDLE_DBC, connection );
-
-                            return function_return( connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
                         }
                         statement -> implicit_ard = statement -> ard;
                         statement -> ard -> implicit = 1;
+                        statement -> ard -> associated_with = statement;
                         statement -> ard -> state = STATE_D1i;
                         statement -> ard -> driver_desc = desc;
                         statement -> ard -> connection = connection;
@@ -712,7 +771,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                          * to wrap around this
                          */
                         statement -> apd = __alloc_desc();
-                        if ( !statement -> ard )
+                        if ( !statement -> apd )
                         {
                             dm_log_write( __FILE__, 
                                 __LINE__, 
@@ -726,12 +785,13 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
         
                             __release_stmt( statement );
 
-                            thread_release( SQL_HANDLE_DBC, connection );
+                            *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
                         }
                         statement -> implicit_apd = statement -> apd;
                         statement -> apd -> implicit = 1;
+                        statement -> apd -> associated_with = statement;
                         statement -> apd -> state = STATE_D1i;
                         statement -> apd -> driver_desc = desc;
                         statement -> apd -> connection = connection;
@@ -769,12 +829,13 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             __release_stmt( statement );
 
-                            thread_release( SQL_HANDLE_DBC, connection );
+                            *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
                         }
                         statement -> implicit_ird = statement -> ird;
                         statement -> ird -> implicit = 1;
+                        statement -> ird -> associated_with = statement;
                         statement -> ird -> state = STATE_D1i;
                         statement -> ird -> driver_desc = desc;
                         statement -> ird -> connection = connection;
@@ -810,25 +871,209 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                                     ERROR_HY013, NULL,
                                     connection -> environment -> requested_version );
 
-                            thread_release( SQL_HANDLE_DBC, connection );
+                            __release_stmt( statement );
 
-                            return function_return( connection, SQL_ERROR );
+                            *output_handle = SQL_NULL_HSTMT;
+
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
                         }
                         statement -> implicit_ipd = statement -> ipd;
                         statement -> ipd -> implicit = 1;
+                        statement -> ipd -> associated_with = statement;
                         statement -> ipd -> state = STATE_D1i;
                         statement -> ipd -> driver_desc = desc;
                         statement -> ipd -> connection = connection;
                     }
                 }
+                /* Driver may only have unicode API's */
+                else if ( CHECK_SQLGETSTMTATTRW( connection ))
+                {
+                    DRV_SQLHDESC desc;
+
+                    /*
+                     * ARD
+                     */
+
+                    ret1 = SQLGETSTMTATTRW( connection,
+                            statement -> driver_stmt,
+                            SQL_ATTR_APP_ROW_DESC,
+                            &desc,
+                            sizeof( desc ),
+                            NULL );
+
+                    if ( SQL_SUCCEEDED( ret1 ))
+                    {
+                        /*
+                         * allocate one of our descriptors
+                         * to wrap around this
+                         */
+                        statement -> ard = __alloc_desc();
+                        if ( !statement -> ard )
+                        {
+                            dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY013" );
+
+                            __post_internal_error( &connection -> error,
+                                    ERROR_HY013, NULL,
+                                    connection -> environment -> requested_version );
+
+                            __release_stmt( statement );
+
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        }
+                        statement -> implicit_ard = statement -> ard;
+                        statement -> ard -> implicit = 1;
+                        statement -> ard -> associated_with = statement;
+                        statement -> ard -> state = STATE_D1i;
+                        statement -> ard -> driver_desc = desc;
+                        statement -> ard -> connection = connection;
+                    }
+
+                    /*
+                     * APD
+                     */
+
+                    ret1 = SQLGETSTMTATTRW( connection,
+                            statement -> driver_stmt,
+                            SQL_ATTR_APP_PARAM_DESC,
+                            &desc,
+                            sizeof( desc ),
+                            NULL );
+
+                    if ( SQL_SUCCEEDED( ret1 ))
+                    {
+                        /*
+                         * allocate one of our descriptors
+                         * to wrap around this
+                         */
+                        statement -> apd = __alloc_desc();
+                        if ( !statement -> apd )
+                        {
+                            dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY013" );
+
+                            __post_internal_error( &connection -> error,
+                                    ERROR_HY013, NULL,
+                                    connection -> environment -> requested_version );
+        
+                            __release_stmt( statement );
+
+                            *output_handle = SQL_NULL_HSTMT;
+
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        }
+                        statement -> implicit_apd = statement -> apd;
+                        statement -> apd -> implicit = 1;
+                        statement -> apd -> associated_with = statement;
+                        statement -> apd -> state = STATE_D1i;
+                        statement -> apd -> driver_desc = desc;
+                        statement -> apd -> connection = connection;
+                    }
+
+                    /*
+                     * IRD
+                     */
+
+                    ret1 = SQLGETSTMTATTRW( connection,
+                            statement -> driver_stmt,
+                            SQL_ATTR_IMP_ROW_DESC,
+                            &desc,
+                            sizeof( desc ),
+                            NULL );
+
+                    if ( SQL_SUCCEEDED( ret1 ))
+                    {
+                        /*
+                         * allocate one of our descriptors
+                         * to wrap around this
+                         */
+                        statement -> ird = __alloc_desc();
+                        if ( !statement -> ird )
+                        {
+                            dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY013" );
+
+                            __post_internal_error( &connection -> error,
+                                    ERROR_HY013, NULL, 
+                                    connection -> environment -> requested_version );
+
+                            __release_stmt( statement );
+
+                            *output_handle = SQL_NULL_HSTMT;
+
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        }
+                        statement -> implicit_ird = statement -> ird;
+                        statement -> ird -> implicit = 1;
+                        statement -> ird -> associated_with = statement;
+                        statement -> ird -> state = STATE_D1i;
+                        statement -> ird -> driver_desc = desc;
+                        statement -> ird -> connection = connection;
+                    }
+
+                    /*
+                     * IPD
+                     */
+
+                    ret1 = SQLGETSTMTATTRW( connection,
+                            statement -> driver_stmt,
+                            SQL_ATTR_IMP_PARAM_DESC,
+                            &desc,
+                            sizeof( desc ),
+                            NULL );
+
+                    if ( SQL_SUCCEEDED( ret1 ))
+                    {
+                        /*
+                         * allocate one of our descriptors
+                         * to wrap around this
+                         */
+                        statement -> ipd = __alloc_desc();
+                        if ( !statement -> ipd )
+                        {
+                            dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY013" );
+
+                            __post_internal_error( &connection -> error,
+                                    ERROR_HY013, NULL,
+                                    connection -> environment -> requested_version );
+
+                            __release_stmt( statement );
+
+                            *output_handle = SQL_NULL_HSTMT;
+
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        }
+                        statement -> implicit_ipd = statement -> ipd;
+                        statement -> ipd -> implicit = 1;
+                        statement -> ipd -> associated_with = statement;
+                        statement -> ipd -> state = STATE_D1i;
+                        statement -> ipd -> driver_desc = desc;
+                        statement -> ipd -> connection = connection;
+                    }
+                } 
             }
 
             /*
              * set any preset statement attributes
              */
 
-
-            __set_attributes( statement, SQL_HANDLE_STMT );
+            if ( SQL_SUCCEEDED( ret ))
+            {
+                __set_attributes( statement, SQL_HANDLE_STMT );
+            }
 
             if ( log_info.log_flag )
             {
@@ -847,9 +1092,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                                UODBC_STATS_TYPE_HSTMT, (void *)1);
 #endif
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, ret );
+            return function_return( SQL_HANDLE_DBC, connection, ret );
         }
         break;
 
@@ -903,9 +1146,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                     ERROR_HY009, NULL, 
                     connection -> environment -> requested_version  );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection , SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection , SQL_ERROR );
         }
 
         if ( connection -> state == STATE_C1 ||
@@ -922,9 +1163,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                     ERROR_08003, NULL,
                     connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
+            *output_handle = SQL_NULL_HDESC;
 
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         descriptor = __alloc_desc();
@@ -940,9 +1181,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                     ERROR_HY013, NULL,
                     connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
+            *output_handle = SQL_NULL_HDESC;
 
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         /*
@@ -974,9 +1215,9 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             __release_desc( descriptor );
 
-            thread_release( SQL_HANDLE_DBC, connection );
+            *output_handle = SQL_NULL_HDESC;
 
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         if ( SQL_SUCCEEDED( ret ))
@@ -987,6 +1228,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             descriptor -> state = STATE_D1e;
             descriptor -> implicit = 0;
+            descriptor -> associated_with = NULL;
 
             connection -> statement_count ++;
 
@@ -1012,15 +1254,40 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                            (void *)1);
 #endif
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, ret );
+        return function_return( SQL_HANDLE_DBC, connection, ret );
         }
         break;
 
       default:
-        return SQL_ERROR;
-        break;
+		if ( __validate_env( (DMHENV) input_handle ))
+		{
+			DMHENV environment = (DMHENV) input_handle;
+
+            thread_protect( SQL_HANDLE_ENV, environment );
+
+			__post_internal_error( &environment -> error,
+						ERROR_HY092, NULL,
+						environment -> requested_version );
+
+			return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
+		}
+		else if ( __validate_dbc( (DMHDBC) input_handle ))
+		{
+			DMHDBC connection = (DMHDBC) input_handle;
+
+            thread_protect( SQL_HANDLE_DBC, connection );
+
+			__post_internal_error( &connection -> error,
+					ERROR_HY092, NULL,
+					connection -> environment -> requested_version );
+	
+			return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+		}
+		else
+		{
+			return SQL_INVALID_HANDLE;
+		}
+		break;
     }
 }
 
