@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,54 @@
  *
  **********************************************************************
  *
- * $Id: SQLBrowseConnectW.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLBrowseConnectW.c,v 1.13 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLBrowseConnectW.c,v $
+ * Revision 1.13  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.12  2007/10/19 10:14:05  lurcher
+ * Pull errors from SQLBrowseConnect when it returns SQL_NEED_DATA
+ *
+ * Revision 1.11  2007/02/28 15:37:46  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.10  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.9  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.8  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.7  2002/08/15 08:10:33  lurcher
+ *
+ * Couple of small fixes from John L Miller
+ *
+ * Revision 1.6  2002/07/25 09:30:26  lurcher
+ *
+ * Additional unicode and iconv changes
+ *
+ * Revision 1.5  2002/07/24 08:49:51  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.4  2002/02/08 17:59:40  lurcher
+ *
+ * Fix threading problem in SQLBrowseConnect
+ *
+ * Revision 1.3  2002/01/21 18:00:51  lurcher
+ *
+ * Assorted fixed and changes, mainly UNICODE/bug fixes
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -72,6 +117,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLBrowseConnectW.c,v $";
@@ -95,6 +141,7 @@ SQLRETURN SQLBrowseConnectW(
     SQLRETURN ret;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ], s2[ 100 + LOG_MESSAGE_LEN ];
     SQLWCHAR *uc_in_str;
+    int warnings = 0;
 
     /*
      * check connection
@@ -108,6 +155,39 @@ SQLRETURN SQLBrowseConnectW(
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDBC parent_connection;
+
+			parent_connection = find_parent_handle( connection, SQL_HANDLE_DBC );
+
+			if ( parent_connection ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLBROWSECONNECTW( parent_connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLBROWSECONNECTW( parent_connection, 
+							connection, 
+							conn_str_in,
+							len_conn_str_in,
+							conn_str_out,
+							conn_str_out_max,
+							ptr_conn_str_out );
+				}
+			}
+		}
+#endif
+
         return SQL_INVALID_HANDLE;
     }
 
@@ -116,10 +196,10 @@ SQLRETURN SQLBrowseConnectW(
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tStr In = %s\
-            \n\t\t\tStr Out = %s\
-            \n\t\t\tPtr Conn Str Out = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tStr In = %s\
+\n\t\t\tStr Out = %s\
+\n\t\t\tPtr Conn Str Out = %p",
                 connection,
                 __wstring_with_length( s1, conn_str_in, len_conn_str_in ), 
                 __wstring_with_length( s2, conn_str_out, conn_str_out_max ), 
@@ -149,13 +229,16 @@ SQLRETURN SQLBrowseConnectW(
         __post_internal_error( &connection -> error,
                 ERROR_08002, NULL, 
                 connection -> environment -> requested_version );
-        return function_return( connection, SQL_ERROR );
+
+        return function_return_nodrv( IGNORE_THREAD, connection, SQL_ERROR );
     }
 
     /*
      * are we at the start of a connection
      */
 
+    thread_protect( SQL_HANDLE_DBC, connection );
+
     if ( connection -> state == STATE_C2 )
     {
         /*
@@ -191,7 +274,8 @@ SQLRETURN SQLBrowseConnectW(
                         ERROR_IM002, NULL,
                         connection -> environment -> requested_version );
                 __release_conn( &con_struct );
-                return function_return( connection, SQL_ERROR );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             __generate_connection_string( &con_struct, in_str, sizeof( in_str ));
@@ -217,7 +301,8 @@ SQLRETURN SQLBrowseConnectW(
                 __post_internal_error( &connection -> error,
                         ERROR_IM012, NULL,
                         connection -> environment -> requested_version );
-                return function_return( connection, SQL_ERROR );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             sprintf( in_str, "DSN=%s;", dsn );
@@ -238,7 +323,8 @@ SQLRETURN SQLBrowseConnectW(
                         ERROR_IM002, NULL,
                         connection -> environment -> requested_version );
                 __release_conn( &con_struct );
-                return function_return( connection, SQL_ERROR );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             __generate_connection_string( &con_struct, in_str, sizeof( in_str ));
@@ -251,8 +337,18 @@ SQLRETURN SQLBrowseConnectW(
          * we now have a driver to connect to
          */
 
-        if ( !__connect_part_one( connection, lib_name, driver_name ))
-            return function_return( connection, SQL_ERROR );
+        if ( !__connect_part_one( connection, lib_name, driver_name, &warnings ))
+        {
+            __disconnect_part_four( connection );       /* release unicode handles */
+
+            dm_log_write( __FILE__,
+                    __LINE__,
+                    LOG_INFO,
+                    LOG_INFO,
+                    "Error: connect_part_one fails" );
+
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        }
 
         if ( !CHECK_SQLBROWSECONNECTW( connection ) &&
             !CHECK_SQLBROWSECONNECT( connection ))
@@ -264,10 +360,12 @@ SQLRETURN SQLBrowseConnectW(
                         "Error: IM001" );
 
             __disconnect_part_one( connection );
+            __disconnect_part_four( connection );       /* release unicode handles */
             __post_internal_error( &connection -> error,
                     ERROR_IM001, NULL,
                     connection -> environment -> requested_version );
-            return function_return( connection, SQL_ERROR );
+
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else
@@ -296,7 +394,7 @@ SQLRETURN SQLBrowseConnectW(
 
     if ( CHECK_SQLBROWSECONNECTW( connection ))
     {
-        uc_in_str = ansi_to_unicode((SQLCHAR*) in_str, SQL_NTS );
+        uc_in_str = ansi_to_unicode_alloc((SQLCHAR*) in_str, SQL_NTS, connection, NULL );
 
         ret = SQLBROWSECONNECTW( connection,
                 connection -> driver_dbc,
@@ -313,24 +411,57 @@ SQLRETURN SQLBrowseConnectW(
     }
     else
     {
-        ret = SQLBROWSECONNECT( connection,
-                connection -> driver_dbc,
-                in_str,
-                SQL_NTS,
-                conn_str_out,
-                conn_str_out_max,
-                ptr_conn_str_out );
+        if ( conn_str_out )
+        {
+            if ( conn_str_out_max > 0 )
+            {
+                SQLCHAR *ob = malloc( conn_str_out_max + 1 );
+                SQLINTEGER len;
+
+                ret = SQLBROWSECONNECT( connection,
+                        connection -> driver_dbc,
+                        in_str,
+                        SQL_NTS,
+                        ob,
+                        conn_str_out_max,
+                        &len );
+
+                if ( len > 0 )
+                {
+                    ansi_to_unicode_copy( conn_str_out, (char*) ob, len, connection, NULL );
+                }
+
+                if ( ptr_conn_str_out )
+                {
+                    *ptr_conn_str_out = len;
+                }
+            }
+            else
+            {
+                ret = SQLBROWSECONNECT( connection,
+                        connection -> driver_dbc,
+                        in_str,
+                        SQL_NTS,
+                        conn_str_out,
+                        conn_str_out_max,
+                        ptr_conn_str_out );
+            }
+        }
+        else
+        {
+            ret = SQLBROWSECONNECT( connection,
+                    connection -> driver_dbc,
+                    in_str,
+                    SQL_NTS,
+                    conn_str_out,
+                    conn_str_out_max,
+                    ptr_conn_str_out );
+        }
 
         connection -> unicode_driver = 0;
-        ansi_to_unicode_in_place((SQLCHAR*) conn_str_out, *ptr_conn_str_out );
     }
 
-    if ( ret == SQL_NEED_DATA )
-    {
-        connection -> state = STATE_C3;
-        return function_return( connection, ret );
-    }
-    else if ( !SQL_SUCCEEDED( ret ))
+    if ( !SQL_SUCCEEDED( ret ) || ret == SQL_NEED_DATA )
     {
         if ( connection -> unicode_driver )
         {
@@ -338,7 +469,7 @@ SQLRETURN SQLBrowseConnectW(
             SQLINTEGER native_error;
             SQLSMALLINT ind;
             SQLWCHAR message_text[ SQL_MAX_MESSAGE_LENGTH + 1 ];
-            SQLRETURN ret;
+            SQLRETURN eret;
 
             /*
              * get the error from the driver before
@@ -349,7 +480,7 @@ SQLRETURN SQLBrowseConnectW(
             {
                 do
                 {
-                    ret = SQLERRORW( connection,
+                    eret = SQLERRORW( connection,
                             SQL_NULL_HENV,
                             connection -> driver_dbc,
                             SQL_NULL_HSTMT,
@@ -360,15 +491,16 @@ SQLRETURN SQLBrowseConnectW(
                             &ind );
 
 
-                    if ( SQL_SUCCEEDED( ret ))
+                    if ( SQL_SUCCEEDED( eret ))
                     {
                         __post_internal_error_ex_w( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
-                while( SQL_SUCCEEDED( ret ));
+                while( SQL_SUCCEEDED( eret ));
             }
             else if ( CHECK_SQLGETDIAGRECW( connection ))
             {
@@ -376,7 +508,7 @@ SQLRETURN SQLBrowseConnectW(
 
                 do
                 {
-                    ret = SQLGETDIAGRECW( connection,
+                    eret = SQLGETDIAGRECW( connection,
                             SQL_HANDLE_DBC,
                             connection -> driver_dbc,
                             rec ++,
@@ -386,19 +518,28 @@ SQLRETURN SQLBrowseConnectW(
                             sizeof( message_text ),
                             &ind );
 
-                    if ( SQL_SUCCEEDED( ret ))
+                    if ( SQL_SUCCEEDED( eret ))
                     {
                         __post_internal_error_ex_w( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
-                while( SQL_SUCCEEDED( ret ));
+                while( SQL_SUCCEEDED( eret ));
             }
 
-            __disconnect_part_one( connection );
-            connection -> state = STATE_C2;
+    		if ( ret != SQL_NEED_DATA ) 
+			{
+        		__disconnect_part_one( connection );
+                __disconnect_part_four( connection );       /* release unicode handles */
+        		connection -> state = STATE_C2;
+			}
+			else 
+			{
+       			connection -> state = STATE_C3;
+			}
         }
         else
         {
@@ -406,7 +547,7 @@ SQLRETURN SQLBrowseConnectW(
             SQLINTEGER native_error;
             SQLSMALLINT ind;
             SQLCHAR message_text[ SQL_MAX_MESSAGE_LENGTH + 1 ];
-            SQLRETURN ret;
+            SQLRETURN eret;
 
             /*
              * get the error from the driver before
@@ -416,7 +557,7 @@ SQLRETURN SQLBrowseConnectW(
             {
                 do
                 {
-                    ret = SQLERROR( connection,
+                    eret = SQLERROR( connection,
                         SQL_NULL_HENV,
                             connection -> driver_dbc,
                             SQL_NULL_HSTMT,
@@ -427,15 +568,16 @@ SQLRETURN SQLBrowseConnectW(
                             &ind );
 
 
-                    if ( SQL_SUCCEEDED( ret ))
+                    if ( SQL_SUCCEEDED( eret ))
                     {
                         __post_internal_error_ex( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
-                while( SQL_SUCCEEDED( ret ));
+                while( SQL_SUCCEEDED( eret ));
             }
             else if ( CHECK_SQLGETDIAGREC( connection ))
             {
@@ -443,7 +585,7 @@ SQLRETURN SQLBrowseConnectW(
 
                 do
                 {
-                    ret = SQLGETDIAGRECW( connection,
+                    eret = SQLGETDIAGRECW( connection,
                             SQL_HANDLE_DBC,
                             connection -> driver_dbc,
                             rec ++,
@@ -453,29 +595,28 @@ SQLRETURN SQLBrowseConnectW(
                             sizeof( message_text ),
                             &ind );
 
-                    if ( SQL_SUCCEEDED( ret ))
+                    if ( SQL_SUCCEEDED( eret ))
                     {
                         __post_internal_error_ex( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
-                while( SQL_SUCCEEDED( ret ));
+                while( SQL_SUCCEEDED( eret ));
             }
 
-            sprintf( connection -> msg,
-                    "\n\t\tExit:[%s]",
-                __get_return_status( ret ));
-
-            dm_log_write( __FILE__,
-                    __LINE__,
-                    LOG_INFO,
-                    LOG_INFO,
-                    connection -> msg );
-
-            __disconnect_part_one( connection );
-            connection -> state = STATE_C2;
+    		if ( ret != SQL_NEED_DATA ) 
+			{
+        		__disconnect_part_one( connection );
+                __disconnect_part_four( connection );       /* release unicode handles */
+        		connection -> state = STATE_C2;
+			}
+			else 
+			{
+       			connection -> state = STATE_C3;
+			}
         }
     }
     else
@@ -485,11 +626,19 @@ SQLRETURN SQLBrowseConnectW(
          */
 
         connection -> state = STATE_C4;
+
+        if( ret == SQL_SUCCESS_WITH_INFO )
+        {
+            function_return_ex( IGNORE_THREAD, connection, ret, TRUE );
+        }
+
         if ( !__connect_part_two( connection ))
         {
             __disconnect_part_two( connection );
             __disconnect_part_one( connection );
-            return function_return( connection, SQL_ERROR );
+            __disconnect_part_four( connection );       /* release unicode handles */
+
+            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -498,7 +647,7 @@ SQLRETURN SQLBrowseConnectW(
         sprintf( connection -> msg, 
                 "\n\t\tExit:[%s]\
                 \n\t\t\tPtr Conn Str Out = %s",
-                    __get_return_status( ret ),
+                    __get_return_status( ret, s2 ),
                     __sptr_as_string( s1, ptr_conn_str_out ));
 
         dm_log_write( __FILE__, 
@@ -508,5 +657,10 @@ SQLRETURN SQLBrowseConnectW(
                 connection -> msg );
     }
 
-    return function_return( connection, ret );
+    if ( warnings && ret == SQL_SUCCESS )
+    {
+        ret = SQL_SUCCESS_WITH_INFO;
+    }
+
+    return function_return_nodrv( SQL_HANDLE_DBC, connection, ret );
 }
