@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,74 @@
  *
  **********************************************************************
  *
- * $Id: SQLColAttribute.c,v 1.2 2001/11/16 11:39:17 lurcher Exp $
+ * $Id: SQLColAttribute.c,v 1.19 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLColAttribute.c,v $
+ * Revision 1.19  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.18  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.17  2008/09/29 14:02:43  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.16  2007/04/02 10:50:17  lurcher
+ * Fix some 64bit problems (only when sizeof(SQLLEN) == 8 )
+ *
+ * Revision 1.15  2006/03/08 09:18:41  lurcher
+ * fix silly typo that was using sizeof( SQL_WCHAR ) instead of SQLWCHAR
+ *
+ * Revision 1.14  2004/11/22 17:02:48  lurcher
+ * Fix unicode/ansi conversion in the SQLGet functions
+ *
+ * Revision 1.13  2004/10/30 20:19:21  peteralexharvey
+ * ODBC spec says last arg for SQLColAttribute() is SQLPOINTER not (SQLEN*).
+ * So switched back to SQLPOINTER.
+ *
+ * Revision 1.12  2004/10/29 10:00:35  lurcher
+ * Fix SQLColAttribute protype
+ *
+ * Revision 1.11  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.10  2003/04/10 13:45:51  lurcher
+ *
+ * Alter the way that SQLDataSources returns the description field (again)
+ *
+ * Revision 1.9  2003/04/09 08:42:18  lurcher
+ *
+ * Allow setting of odbcinstQ lib from odbcinst.ini and Environment
+ *
+ * Revision 1.8  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.7  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.6  2002/11/11 17:10:06  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.5  2002/08/19 09:11:49  lurcher
+ *
+ * Fix Maxor ineffiecny in Postgres Drivers, and fix a return state
+ *
+ * Revision 1.4  2002/07/24 08:49:51  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.3  2002/04/25 15:16:46  lurcher
+ *
+ * Fix bug with SQLCOlAttribute(s)(W) where a column of zero could not be
+ * used to get the count value
+ *
  * Revision 1.2  2001/11/16 11:39:17  lurcher
  *
  * Add mapping between ODBC 2 and ODBC 3 types for SQLColAttribute(s)(W)
@@ -117,9 +182,10 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLColAttribute.c,v $ $Revision: 1.2 $";
+static char const rcsid[]= "$RCSfile: SQLColAttribute.c,v $ $Revision: 1.19 $";
 
 SQLINTEGER map_ca_odbc3_to_2( SQLINTEGER field_identifier )
 {
@@ -160,17 +226,35 @@ SQLINTEGER map_ca_odbc3_to_2( SQLINTEGER field_identifier )
     return field_identifier;
 }
 
+SQLRETURN SQLColAttributeA( SQLHSTMT statement_handle,
+           SQLSMALLINT column_number,
+           SQLSMALLINT field_identifier,
+           SQLPOINTER character_attribute,
+           SQLSMALLINT buffer_length,
+           SQLSMALLINT *string_length,
+           SQLLEN *numeric_attribute )
+{
+    return SQLColAttribute( statement_handle,
+                            (SQLUSMALLINT) column_number,
+                            (SQLUSMALLINT) field_identifier,
+                            character_attribute,
+                            buffer_length,
+                            string_length,
+                            numeric_attribute );
+}
+
 SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
            SQLUSMALLINT column_number,
            SQLUSMALLINT field_identifier,
            SQLPOINTER character_attribute,
            SQLSMALLINT buffer_length,
            SQLSMALLINT *string_length,
-           SQLPOINTER numeric_attribute )
+           SQLLEN *numeric_attribute )
 {
     DMHSTMT statement = (DMHSTMT) statement_handle;
-    SQLRETURN ret;
+    SQLRETURN ret = 0;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
+    int isStringAttr;
 
     /*
      * check statement
@@ -191,13 +275,13 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, "\n\t\tEntry:\
-            \n\t\t\tStatement = %p\
-            \n\t\t\tColumn Number = %d\
-            \n\t\t\tField Identifier = %s\
-            \n\t\t\tCharacter Attr = %p\
-            \n\t\t\tBuffer Length = %d\
-            \n\t\t\tString Length = %p\
-            \n\t\t\tNumeric Attribute = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tColumn Number = %d\
+\n\t\t\tField Identifier = %s\
+\n\t\t\tCharacter Attr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Length = %p\
+\n\t\t\tNumeric Attribute = %p",
                 statement,
                 column_number,
                 __col_attr_as_string( s1, field_identifier ),
@@ -213,8 +297,11 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
+    thread_protect( SQL_HANDLE_STMT, statement );
+
     if ( column_number == 0 &&
-            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF )
+            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF &&
+            field_identifier != SQL_DESC_COUNT )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -222,10 +309,12 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                 LOG_INFO, 
                 "Error: 07009" );
 
-        __post_internal_error( &statement -> error,
+        __post_internal_error_api( &statement -> error,
                 ERROR_07009, NULL,
-                statement -> connection -> environment -> requested_version );
-        return function_return( statement, SQL_ERROR );
+                statement -> connection -> environment -> requested_version,
+                SQL_API_SQLCOLATTRIBUTE );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
 	/*
@@ -258,8 +347,10 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
         __post_internal_error( &statement -> error,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
-        return function_return( statement, SQL_ERROR );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
+    /* MS Driver manager passes this to driver
     else if ( statement -> state == STATE_S2 &&
             field_identifier != SQL_DESC_COUNT )
     {
@@ -272,8 +363,10 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
         __post_internal_error( &statement -> error,
                 ERROR_07005, NULL,
                 statement -> connection -> environment -> requested_version );
-        return function_return( statement, SQL_ERROR );
+
+        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
+    */
     else if ( statement -> state == STATE_S4 )
     {
         dm_log_write( __FILE__, 
@@ -285,11 +378,15 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
         __post_internal_error( &statement -> error,
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
-        return function_return( statement, SQL_ERROR );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -300,7 +397,8 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
         __post_internal_error( &statement -> error,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
-        return function_return( statement, SQL_ERROR );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -317,27 +415,100 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
             __post_internal_error( &statement -> error,
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
-            return function_return( statement, SQL_ERROR );
+
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
+    switch( field_identifier )
+    {
+        case SQL_DESC_AUTO_UNIQUE_VALUE:
+        case SQL_DESC_CASE_SENSITIVE:
+        case SQL_DESC_CONCISE_TYPE:
+        case SQL_DESC_COUNT:
+        case SQL_DESC_DISPLAY_SIZE:
+        case SQL_DESC_FIXED_PREC_SCALE:
+        case SQL_DESC_LENGTH:
+        case SQL_DESC_NULLABLE:
+        case SQL_DESC_NUM_PREC_RADIX:
+        case SQL_DESC_OCTET_LENGTH:
+        case SQL_DESC_PRECISION:
+        case SQL_DESC_SCALE:
+        case SQL_DESC_SEARCHABLE:
+        case SQL_DESC_TYPE:
+        case SQL_DESC_UNNAMED:
+        case SQL_DESC_UNSIGNED:
+        case SQL_DESC_UPDATABLE:
+            isStringAttr = 0;
+            break;
+        case SQL_COLUMN_QUALIFIER_NAME:
+        case SQL_COLUMN_NAME:
+        case SQL_COLUMN_LABEL:
+        case SQL_COLUMN_OWNER_NAME:
+        case SQL_COLUMN_TABLE_NAME:
+        case SQL_COLUMN_TYPE_NAME:
+        case SQL_DESC_BASE_COLUMN_NAME:
+        case SQL_DESC_BASE_TABLE_NAME:
+        case SQL_DESC_LITERAL_PREFIX:
+        case SQL_DESC_LITERAL_SUFFIX:
+        case SQL_DESC_LOCAL_TYPE_NAME:
+        case SQL_DESC_NAME:
+            if ( buffer_length < 0 && buffer_length != SQL_NTS )
+            {
+                __post_internal_error( &statement -> error,
+                    ERROR_HY090, NULL,
+                    statement -> connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            }
+        default:
+            isStringAttr = buffer_length >= 0;
+            break;
+    }
+
     if ( statement -> connection -> unicode_driver )
     {
         if ( !CHECK_SQLCOLATTRIBUTEW( statement -> connection ))
         {
             if ( CHECK_SQLCOLATTRIBUTESW( statement -> connection ))
             {
+                SQLWCHAR *s1 = NULL;
+
                 /*
                  * map to the ODBC2 function
                  */
 
                 field_identifier = map_ca_odbc3_to_2( field_identifier );
 
+                switch( field_identifier )
+                {
+                  case SQL_COLUMN_QUALIFIER_NAME:
+                  case SQL_COLUMN_NAME:
+                  case SQL_COLUMN_LABEL:
+                  case SQL_COLUMN_OWNER_NAME:
+                  case SQL_COLUMN_TABLE_NAME:
+                  case SQL_COLUMN_TYPE_NAME:
+                  case SQL_DESC_BASE_COLUMN_NAME:
+                  case SQL_DESC_BASE_TABLE_NAME:
+                  case SQL_DESC_LITERAL_PREFIX:
+                  case SQL_DESC_LITERAL_SUFFIX:
+                  case SQL_DESC_LOCAL_TYPE_NAME:
+                  case SQL_DESC_NAME:
+                    if ( SQL_SUCCEEDED( ret ) && character_attribute && buffer_length > 0 )
+                    {
+                        s1 = calloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ), 1);
+                    }
+                    break;
+
+                  default:
+                    break;
+                }
+
                 ret = SQLCOLATTRIBUTESW( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
+                    s1 ? s1 : character_attribute,
                     buffer_length,
                     string_length,
                     numeric_attribute );
@@ -356,15 +527,24 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                   case SQL_DESC_LITERAL_SUFFIX:
                   case SQL_DESC_LOCAL_TYPE_NAME:
                   case SQL_DESC_NAME:
-                    if ( SQL_SUCCEEDED( ret ) && character_attribute )
+                    if ( SQL_SUCCEEDED( ret ) && character_attribute && s1 )
                     {
-                        unicode_to_ansi( character_attribute, SQL_NTS );
+                        unicode_to_ansi_copy( character_attribute, buffer_length, s1,  SQL_NTS, statement -> connection, NULL );
                     }
+					if ( SQL_SUCCEEDED( ret ) && string_length ) 
+					{
+						*string_length /= sizeof( SQLWCHAR );	
+					}
                     break;
 
                   default:
                     break;
                 }
+
+                if ( s1 )
+                {
+                    free( s1 );
+                }
             }
             else
             {
@@ -377,42 +557,71 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                 __post_internal_error( &statement -> error,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
-                return function_return( statement, SQL_ERROR );
+
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
         {
+            SQLWCHAR *s1 = NULL;
+
+            SQLSMALLINT unibuf_len;
+            if ( isStringAttr && character_attribute && buffer_length > 0 )
+                {
+                    s1 = calloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ), 1);
+                /* Do not overflow, since SQLSMALLINT can only hold -32768 <= x <= 32767 */
+                unibuf_len = buffer_length > 16383 ? buffer_length : sizeof( SQLWCHAR ) * buffer_length;
+            }
+
             ret = SQLCOLATTRIBUTEW( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
-                    buffer_length,
+                    s1 ? s1 : character_attribute,
+                    s1 ? unibuf_len : buffer_length,
                     string_length,
                     numeric_attribute );
 
-            switch( field_identifier )
+            if ( SQL_SUCCEEDED( ret ) && isStringAttr && buffer_length > 0 )
             {
+                if ( character_attribute && s1 )
+            {
+                    unicode_to_ansi_copy( character_attribute, buffer_length, s1, SQL_NTS, statement -> connection, NULL );
+                }
+                /*
+                    BUGBUG: Windows DM returns the number of bytes for the Unicode string
+                    but only for certain ODBC-defined string fields, and when truncating
+                */
+                switch ( field_identifier )
+                {
+                case SQL_COLUMN_QUALIFIER_NAME:
+                case SQL_COLUMN_NAME:
+                case SQL_COLUMN_LABEL:
+                case SQL_COLUMN_OWNER_NAME:
+                case SQL_COLUMN_TABLE_NAME:
+                case SQL_COLUMN_TYPE_NAME:
               case SQL_DESC_BASE_COLUMN_NAME:
               case SQL_DESC_BASE_TABLE_NAME:
-              case SQL_DESC_CATALOG_NAME:
-              case SQL_DESC_LABEL:
               case SQL_DESC_LITERAL_PREFIX:
               case SQL_DESC_LITERAL_SUFFIX:
               case SQL_DESC_LOCAL_TYPE_NAME:
               case SQL_DESC_NAME:
-              case SQL_DESC_SCHEMA_NAME:
-              case SQL_DESC_TABLE_NAME:
-              case SQL_DESC_TYPE_NAME:
-              case SQL_COLUMN_NAME:
-                if ( SQL_SUCCEEDED( ret ) && character_attribute )
-                {
-                    unicode_to_ansi( character_attribute, SQL_NTS );
+                    if ( ret == SQL_SUCCESS && string_length )
+                    {
+                        *string_length /= sizeof( SQLWCHAR );	
+                    }
+                    break;
+                default:
+                    if ( string_length )
+				{
+					*string_length /= sizeof( SQLWCHAR );	
+				}
                 }
-                break;
+            }
 
-              default:
-                break;
+            if ( s1 )
+            {
+                free( s1 );
             }
         }
     }
@@ -452,7 +661,7 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                 __post_internal_error( &statement -> error,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
@@ -494,7 +703,7 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
     {
         sprintf( statement -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -503,5 +712,5 @@ SQLRETURN SQLColAttribute ( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    return function_return( statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret );
 }
