@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,56 @@
  *
  **********************************************************************
  *
- * $Id: SQLColAttributes.c,v 1.3 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLColAttributes.c,v 1.15 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLColAttributes.c,v $
+ * Revision 1.15  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.14  2008/09/29 14:02:43  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.13  2006/03/08 09:18:41  lurcher
+ * fix silly typo that was using sizeof( SQL_WCHAR ) instead of SQLWCHAR
+ *
+ * Revision 1.12  2006/01/06 18:44:35  lurcher
+ * Couple of unicode fixes
+ *
+ * Revision 1.11  2004/11/22 17:02:48  lurcher
+ * Fix unicode/ansi conversion in the SQLGet functions
+ *
+ * Revision 1.10  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.9  2003/08/15 17:34:43  lurcher
+ *
+ * Remove some unneeded ODBC2->3 attribute conversions
+ *
+ * Revision 1.8  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.7  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.6  2002/08/19 09:11:49  lurcher
+ *
+ * Fix Maxor ineffiecny in Postgres Drivers, and fix a return state
+ *
+ * Revision 1.5  2002/07/24 08:49:51  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.4  2002/04/25 15:16:46  lurcher
+ *
+ * Fix bug with SQLCOlAttribute(s)(W) where a column of zero could not be
+ * used to get the count value
+ *
  * Revision 1.3  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -120,9 +167,10 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLColAttributes.c,v $ $Revision: 1.3 $";
+static char const rcsid[]= "$RCSfile: SQLColAttributes.c,v $ $Revision: 1.15 $";
 
 SQLINTEGER  map_ca_odbc2_to_3( SQLINTEGER field_identifier )
 {
@@ -132,6 +180,7 @@ SQLINTEGER  map_ca_odbc2_to_3( SQLINTEGER field_identifier )
         field_identifier = SQL_DESC_COUNT; 
         break;
 
+        /*
       case SQL_COLUMN_TYPE:
         field_identifier = SQL_DESC_TYPE; 
         break;
@@ -147,6 +196,7 @@ SQLINTEGER  map_ca_odbc2_to_3( SQLINTEGER field_identifier )
       case SQL_COLUMN_SCALE:
         field_identifier = SQL_DESC_SCALE; 
         break;
+        */
 
       case SQL_COLUMN_NULLABLE:
         field_identifier = SQL_DESC_NULLABLE; 
@@ -163,6 +213,23 @@ SQLINTEGER  map_ca_odbc2_to_3( SQLINTEGER field_identifier )
     return field_identifier;
 }
 
+SQLRETURN SQLColAttributesA( SQLHSTMT statement_handle,
+           SQLUSMALLINT column_number,
+           SQLUSMALLINT field_identifier,
+           SQLPOINTER   character_attribute,
+           SQLSMALLINT  buffer_length,
+           SQLSMALLINT  *string_length,
+           SQLLEN       *numeric_attribute )
+{
+    return SQLColAttributes( statement_handle,
+                            column_number,
+                            field_identifier,
+                            character_attribute,
+                            buffer_length,
+                            string_length,
+                            numeric_attribute );
+}
+
 SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
            SQLUSMALLINT column_number,
            SQLUSMALLINT field_identifier,
@@ -174,6 +241,7 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
     DMHSTMT statement = (DMHSTMT) statement_handle;
     SQLRETURN ret;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
+    int isStringAttr;
 
     /*
      * check statement
@@ -194,13 +262,13 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, "\n\t\tEntry:\
-            \n\t\t\tStatement = %p\
-            \n\t\t\tColumn Number = %d\
-            \n\t\t\tField Identifier = %s\
-            \n\t\t\tCharacter Attr = %p\
-            \n\t\t\tBuffer Length = %d\
-            \n\t\t\tString Length = %p\
-            \n\t\t\tNumeric Attribute = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tColumn Number = %d\
+\n\t\t\tField Identifier = %s\
+\n\t\t\tCharacter Attr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Length = %p\
+\n\t\t\tNumeric Attribute = %p",
                 statement,
                 column_number,
                 __col_attr_as_string( s1, field_identifier ),
@@ -219,7 +287,8 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
     thread_protect( SQL_HANDLE_STMT, statement );
 
     if ( column_number == 0 &&
-            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF )
+            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF &&
+            field_identifier != SQL_COLUMN_COUNT )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -227,13 +296,13 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 LOG_INFO, 
                 "Error: 07009" );
 
-        __post_internal_error( &statement -> error,
+        __post_internal_error_api( &statement -> error,
                 ERROR_07009, NULL,
-                statement -> connection -> environment -> requested_version );
+                statement -> connection -> environment -> requested_version,
+                
+                SQL_API_SQLCOLATTRIBUTES );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
 	/*
@@ -254,9 +323,7 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 ERROR_07009, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 	*/
 
@@ -275,10 +342,9 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
+    /*
     else if ( statement -> state == STATE_S2 &&
             field_identifier != SQL_DESC_COUNT )
     {
@@ -292,10 +358,9 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 ERROR_07005, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
+    */
     else if ( statement -> state == STATE_S4 )
     {
         dm_log_write( __FILE__, 
@@ -308,13 +373,14 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -326,9 +392,7 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -346,36 +410,88 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
+    switch( field_identifier )
+    {
+        case SQL_COLUMN_AUTO_INCREMENT:
+        case SQL_COLUMN_CASE_SENSITIVE:
+        case SQL_COLUMN_COUNT:
+        case SQL_COLUMN_DISPLAY_SIZE:
+        case SQL_COLUMN_LENGTH:
+        case SQL_COLUMN_MONEY:
+        case SQL_COLUMN_NULLABLE:
+        case SQL_COLUMN_PRECISION:
+        case SQL_COLUMN_SCALE:
+        case SQL_COLUMN_SEARCHABLE:
+        case SQL_COLUMN_TYPE:
+        case SQL_COLUMN_UNSIGNED:
+        case SQL_COLUMN_UPDATABLE:
+            isStringAttr = 0;
+            break;
+        case SQL_COLUMN_LABEL:
+        case SQL_COLUMN_NAME:
+        case SQL_COLUMN_OWNER_NAME:
+        case SQL_COLUMN_QUALIFIER_NAME:
+        case SQL_COLUMN_TABLE_NAME:
+        case SQL_COLUMN_TYPE_NAME:
+            if ( buffer_length < 0 )
+            {
+                __post_internal_error( &statement -> error,
+                    ERROR_HY090, NULL,
+                    statement -> connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            }
+        default:
+            isStringAttr = buffer_length >= 0;
+            break;
+    }
+
     if ( statement -> connection -> unicode_driver )
     {
         if ( !CHECK_SQLCOLATTRIBUTESW( statement -> connection ))
         {
             if ( CHECK_SQLCOLATTRIBUTEW( statement -> connection ))
             {
+                SQLWCHAR *s1 = NULL;
+                SQLSMALLINT unibuf_len;
                 /*
                  * map to the ODBC3 function
                  */
 
                 field_identifier = map_ca_odbc2_to_3( field_identifier );
 
+                if ( isStringAttr && character_attribute && buffer_length > 0 )
+                {
+                    s1 = calloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ), 1);
+                    /* Do not overflow, since SQLSMALLINT can only hold -32768 <= x <= 32767 */
+                    unibuf_len = buffer_length > 16383 ? buffer_length : sizeof( SQLWCHAR ) * buffer_length;
+                }
+
                 ret = SQLCOLATTRIBUTEW( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
-                    buffer_length,
+                    s1 ? s1 : character_attribute,
+                    s1 ? unibuf_len : buffer_length,
                     string_length,
                     numeric_attribute );
 
-                if ( SQL_SUCCEEDED( ret ) && character_attribute )
+                if ( SQL_SUCCEEDED( ret ) && isStringAttr && character_attribute && buffer_length > 0 && s1 )
                 {
-                    unicode_to_ansi( character_attribute, SQL_NTS );
+                    if ( string_length && ret == SQL_SUCCESS )
+                    {
+                        *string_length /= sizeof ( SQLWCHAR );
+                    }
+                    unicode_to_ansi_copy( character_attribute, buffer_length, s1, SQL_NTS, statement -> connection, NULL );
+                }
+
+                if ( s1 )
+                {
+                    free( s1 );
                 }
             }
             else
@@ -390,25 +506,39 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
         {
+            SQLWCHAR *s1 = NULL;
+
+            if ( character_attribute && buffer_length > 0 )
+            {
+                s1 = calloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ), 1);
+            }
+
             ret = SQLCOLATTRIBUTESW( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
+                    s1 ? s1 : character_attribute,
                     buffer_length,
                     string_length,
                     numeric_attribute );
 
             if ( SQL_SUCCEEDED( ret ) && character_attribute )
             {
-                unicode_to_ansi( character_attribute, SQL_NTS );
+                unicode_to_ansi_copy( character_attribute, buffer_length, s1, SQL_NTS, statement -> connection, NULL );
+			}
+			if ( SQL_SUCCEEDED( ret ) && string_length && character_attribute ) 
+			{
+				*string_length /= sizeof( SQLWCHAR );	
+			}
+
+            if ( s1 )
+            {
+                free( s1 );
             }
         }
     }
@@ -445,9 +575,7 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
@@ -508,7 +636,7 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
     {
         sprintf( statement -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -517,7 +645,5 @@ SQLRETURN SQLColAttributes( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    thread_release( SQL_HANDLE_STMT, statement );
-
-    return function_return( statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret );
 }
