@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,53 @@
  *
  **********************************************************************
  *
- * $Id: SQLColAttributesW.c,v 1.3 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLColAttributesW.c,v 1.14 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLColAttributesW.c,v $
+ * Revision 1.14  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.13  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.12  2008/08/29 08:01:38  lurcher
+ * Alter the way W functions are passed to the driver
+ *
+ * Revision 1.11  2007/02/28 15:37:47  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.10  2004/11/22 17:02:48  lurcher
+ * Fix unicode/ansi conversion in the SQLGet functions
+ *
+ * Revision 1.9  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.8  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.7  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.6  2002/08/19 09:11:49  lurcher
+ *
+ * Fix Maxor ineffiecny in Postgres Drivers, and fix a return state
+ *
+ * Revision 1.5  2002/07/24 08:49:51  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.4  2002/04/25 15:16:46  lurcher
+ *
+ * Fix bug with SQLCOlAttribute(s)(W) where a column of zero could not be
+ * used to get the count value
+ *
  * Revision 1.3  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -60,6 +104,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLColAttributesW.c,v $";
@@ -87,6 +132,39 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHSTMT parent_statement;
+
+			parent_statement = find_parent_handle( statement, SQL_HANDLE_STMT );
+
+			if ( parent_statement ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLCOLATTRIBUTEW( parent_statement -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+            		return  SQLCOLATTRIBUTEW( parent_statement -> connection,
+							statement_handle,
+                    		column_number,
+                    		field_identifier,
+                    		character_attribute,
+                    		buffer_length,
+                    		string_length,
+                    		numeric_attribute );
+				}
+			}
+		}
+#endif
         return SQL_INVALID_HANDLE;
     }
 
@@ -95,13 +173,13 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, "\n\t\tEntry:\
-            \n\t\t\tStatement = %p\
-            \n\t\t\tColumn Number = %d\
-            \n\t\t\tField Identifier = %s\
-            \n\t\t\tCharacter Attr = %p\
-            \n\t\t\tBuffer Length = %d\
-            \n\t\t\tString Length = %p\
-            \n\t\t\tNumeric Attribute = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tColumn Number = %d\
+\n\t\t\tField Identifier = %s\
+\n\t\t\tCharacter Attr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Length = %p\
+\n\t\t\tNumeric Attribute = %p",
                 statement,
                 column_number,
                 __col_attr_as_string( s1, field_identifier ),
@@ -120,7 +198,8 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
     thread_protect( SQL_HANDLE_STMT, statement );
 
     if ( column_number == 0 &&
-            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF )
+            statement -> bookmarks_on == SQL_UB_OFF && statement -> connection -> bookmarks_on == SQL_UB_OFF &&
+            field_identifier != SQL_COLUMN_COUNT )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -128,13 +207,12 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 LOG_INFO, 
                 "Error: 07009" );
 
-        __post_internal_error( &statement -> error,
+        __post_internal_error_api( &statement -> error,
                 ERROR_07009, NULL,
-                statement -> connection -> environment -> requested_version );
+                statement -> connection -> environment -> requested_version,
+                SQL_API_SQLCOLATTRIBUTES );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
 	/*
@@ -155,9 +233,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 ERROR_07009, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 	*/
 
@@ -176,9 +252,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S2 &&
             field_identifier != SQL_DESC_COUNT )
@@ -193,9 +267,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 ERROR_07005, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S4 )
     {
@@ -209,13 +281,14 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -227,9 +300,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -247,13 +318,31 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
-    if ( statement -> connection -> unicode_driver )
+    switch ( field_identifier )
+    {
+        case SQL_COLUMN_LABEL:
+        case SQL_COLUMN_NAME:
+        case SQL_COLUMN_OWNER_NAME:
+        case SQL_COLUMN_QUALIFIER_NAME:
+        case SQL_COLUMN_TABLE_NAME:
+        case SQL_COLUMN_TYPE_NAME:
+            if ( buffer_length < 0 && buffer_length != SQL_NTS )
+            {
+                __post_internal_error( &statement -> error,
+                    ERROR_HY090, NULL,
+                    statement -> connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            }
+    }
+
+    if ( statement -> connection -> unicode_driver || 
+			    CHECK_SQLCOLATTRIBUTESW( statement -> connection ) || 
+			    CHECK_SQLCOLATTRIBUTEW( statement -> connection ))
     {
         if ( !CHECK_SQLCOLATTRIBUTESW( statement -> connection ))
         {
@@ -286,9 +375,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
@@ -309,17 +396,44 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
         {
             if ( CHECK_SQLCOLATTRIBUTE( statement -> connection ))
             {
+                SQLCHAR *as1 = NULL;
+
                 /*
                  * map to the ODBC3 function
                  */
 
                 field_identifier = map_ca_odbc2_to_3( field_identifier );
 
+                switch( field_identifier )
+                {
+                  case SQL_DESC_BASE_COLUMN_NAME:
+                  case SQL_DESC_BASE_TABLE_NAME:
+                  case SQL_DESC_CATALOG_NAME:
+                  case SQL_DESC_LABEL:
+                  case SQL_DESC_LITERAL_PREFIX:
+                  case SQL_DESC_LITERAL_SUFFIX:
+                  case SQL_DESC_LOCAL_TYPE_NAME:
+                  case SQL_DESC_NAME:
+                  case SQL_DESC_SCHEMA_NAME:
+                  case SQL_DESC_TABLE_NAME:
+                  case SQL_DESC_TYPE_NAME:
+                  case SQL_COLUMN_NAME:
+                    buffer_length = buffer_length / 2;
+                    if ( buffer_length > 0 )
+                    {
+                        as1 = malloc( buffer_length + 1 );
+                    }
+                    break;
+
+                  default:
+                    break;
+                }
+
                 ret = SQLCOLATTRIBUTE( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
+                    as1 ? as1 : character_attribute,
                     buffer_length,
                     string_length,
                     numeric_attribute );
@@ -338,10 +452,19 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                   case SQL_DESC_TABLE_NAME:
                   case SQL_DESC_TYPE_NAME:
                   case SQL_COLUMN_NAME:
-                    if ( SQL_SUCCEEDED( ret ) && character_attribute )
+                    if ( SQL_SUCCEEDED( ret ) && character_attribute && as1 )
                     {
-                        ansi_to_unicode_in_place( character_attribute, SQL_NTS );
+                        ansi_to_unicode_copy( character_attribute, (char*) as1, SQL_NTS, statement -> connection, NULL );
                     }
+
+                    if ( as1 ) 
+                    {
+                        free( as1 );
+                    }
+					if ( SQL_SUCCEEDED( ret ) && string_length )
+					{
+						*string_length *= sizeof( SQLWCHAR );
+					}
                     break;
 
                   default:
@@ -360,18 +483,43 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                         ERROR_IM001, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
         else
         {
+            SQLCHAR *as1 = NULL;
+
+            switch( field_identifier )
+            {
+              case SQL_DESC_BASE_COLUMN_NAME:
+              case SQL_DESC_BASE_TABLE_NAME:
+              case SQL_DESC_CATALOG_NAME:
+              case SQL_DESC_LABEL:
+              case SQL_DESC_LITERAL_PREFIX:
+              case SQL_DESC_LITERAL_SUFFIX:
+              case SQL_DESC_LOCAL_TYPE_NAME:
+              case SQL_DESC_NAME:
+              case SQL_DESC_SCHEMA_NAME:
+              case SQL_DESC_TABLE_NAME:
+              case SQL_DESC_TYPE_NAME:
+              case SQL_COLUMN_NAME:
+                buffer_length = buffer_length / 2;
+                if ( buffer_length > 0 )
+                {
+                    as1 = malloc( buffer_length + 1 );
+                }
+                break;
+
+              default:
+                break;
+            }
+
             ret = SQLCOLATTRIBUTES( statement -> connection,
                     statement -> driver_stmt,
                     column_number,
                     field_identifier,
-                    character_attribute,
+                    as1 ? as1 : character_attribute,
                     buffer_length,
                     string_length,
                     numeric_attribute );
@@ -390,10 +538,18 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
               case SQL_DESC_TABLE_NAME:
               case SQL_DESC_TYPE_NAME:
               case SQL_COLUMN_NAME:
-                if ( SQL_SUCCEEDED( ret ) && character_attribute )
+                if ( SQL_SUCCEEDED( ret ) && character_attribute && as1 )
                 {
-                    ansi_to_unicode_in_place( character_attribute, SQL_NTS );
+                    ansi_to_unicode_copy( character_attribute, (char*) as1, SQL_NTS, statement -> connection, NULL );
                 }
+                if ( as1 ) 
+                {
+                    free( as1 );
+                }
+				if ( SQL_SUCCEEDED( ret ) && string_length && character_attribute )
+				{
+					*string_length *= sizeof( SQLWCHAR );
+				}
                 break;
 
               default:
@@ -447,7 +603,7 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
     {
         sprintf( statement -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -456,7 +612,5 @@ SQLRETURN SQLColAttributesW( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    thread_release( SQL_HANDLE_STMT, statement );
-
-    return function_return( statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret );
 }
