@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,36 @@
  *
  **********************************************************************
  *
- * $Id: SQLCopyDesc.c,v 1.1.1.1 2001/10/17 16:40:05 lurcher Exp $
+ * $Id: SQLCopyDesc.c,v 1.8 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLCopyDesc.c,v $
+ * Revision 1.8  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.7  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.6  2004/03/30 13:20:11  lurcher
+ *
+ *
+ * Fix problem with SQLCopyDesc
+ * Add additional target for iconv
+ *
+ * Revision 1.5  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.4  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.3  2002/09/18 14:49:32  lurcher
+ *
+ * DataManagerII additions and some more threading fixes
+ *
  * Revision 1.1.1.1  2001/10/17 16:40:05  lurcher
  *
  * First upload to SourceForge
@@ -88,9 +115,10 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLCopyDesc.c,v $ $Revision: 1.1.1.1 $";
+static char const rcsid[]= "$RCSfile: SQLCopyDesc.c,v $ $Revision: 1.8 $";
 
 struct cdesc
 {
@@ -133,6 +161,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
 {
     DMHDESC src_descriptor = (DMHDESC)source_desc_handle;
     DMHDESC target_descriptor = (DMHDESC)target_desc_handle;
+    SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
 
     /*
      * check descriptor
@@ -165,8 +194,8 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
     if ( log_info.log_flag )
     {
         sprintf( src_descriptor -> msg, "\n\t\tEntry:\
-            \n\t\t\tSource Descriptor = %p\
-            \n\t\t\tTarget Descriptor = %p", 
+\n\t\t\tSource Descriptor = %p\
+\n\t\t\tTarget Descriptor = %p", 
                 src_descriptor,
                 target_descriptor );
 
@@ -177,13 +206,64 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 src_descriptor -> msg );
     }
 
+	if ( src_descriptor -> associated_with ) {
+    	DMHSTMT statement = (DMHSTMT) src_descriptor -> associated_with;
+
+    	if ( statement -> state == STATE_S8 ||
+            statement -> state == STATE_S9 ||
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S11 ||
+            statement -> state == STATE_S12 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
+    	{
+       	 	dm_log_write( __FILE__, 
+                __LINE__, 
+                LOG_INFO, 
+                LOG_INFO, 
+                "Error: HY010" );
+
+        	__post_internal_error( &src_descriptor -> error,
+                ERROR_HY010, NULL,
+                src_descriptor -> connection -> environment -> requested_version );
+
+			function_return_nodrv( SQL_HANDLE_DESC, target_descriptor, SQL_SUCCESS );
+        	return function_return_nodrv( SQL_HANDLE_DESC, src_descriptor, SQL_ERROR );
+    	}
+	}
+
+	if ( target_descriptor -> associated_with ) {
+    	DMHSTMT statement = (DMHSTMT) target_descriptor -> associated_with;
+
+    	if ( statement -> state == STATE_S8 ||
+            statement -> state == STATE_S9 ||
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S11 ||
+            statement -> state == STATE_S12 )
+    	{
+       	 	dm_log_write( __FILE__, 
+                __LINE__, 
+                LOG_INFO, 
+                LOG_INFO, 
+                "Error: HY010" );
+
+        	__post_internal_error( &target_descriptor -> error,
+                ERROR_HY010, NULL,
+                target_descriptor -> connection -> environment -> requested_version );
+
+        	return function_return_nodrv( IGNORE_THREAD, src_descriptor, SQL_ERROR );
+    	}
+	}
+
     /*
-     * if both descriptors are from the same connection the we can just
+     * if both descriptors are from the same driver then we can just
      * pass it on
      */
 
-    if ( src_descriptor -> connection == 
-            target_descriptor -> connection && 
+    if ( (src_descriptor -> connection == target_descriptor -> connection ||
+          !strcmp(src_descriptor -> connection -> dl_name,
+                  target_descriptor -> connection -> dl_name) ) && 
             CHECK_SQLCOPYDESC( src_descriptor -> connection ))
     {
         SQLRETURN ret;
@@ -200,20 +280,18 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
 
         if ( log_info.log_flag )
         {
-            sprintf( src_descriptor -> msg, 
+            sprintf( target_descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
                     LOG_INFO, 
                     LOG_INFO, 
-                    src_descriptor -> msg );
+                    target_descriptor -> msg );
         }
 
-        thread_release( SQL_HANDLE_DBC, src_descriptor -> connection );
-
-        return function_return( target_descriptor, ret );
+        return function_return( IGNORE_THREAD, target_descriptor, ret );
     }
     else
     {
@@ -251,14 +329,14 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
             else
                 thread_release( SQL_HANDLE_ENV, src_descriptor -> connection -> environment );
 
-            return function_return( target_descriptor, SQL_ERROR );
+            return function_return_nodrv( IGNORE_THREAD, target_descriptor, SQL_ERROR );
         }
 
         /*
          * get the count from the source field
          */
 
-        ret == SQLGETDESCFIELD( src_descriptor -> connection,
+        ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                 src_descriptor -> driver_desc,
                                 0, 
                                 SQL_DESC_COUNT,
@@ -277,7 +355,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
             {
                 if ( header_fields[ i ].field_type == SQL_INTEGER )
                 {
-                    ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                    ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                 src_descriptor -> driver_desc,
                                 0, 
                                 header_fields[ i ].field_identifier,
@@ -287,7 +365,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 }
                 else if ( header_fields[ i ].field_type == SQL_SMALLINT )
                 {
-                    ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                    ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                 src_descriptor -> driver_desc,
                                 0, 
                                 header_fields[ i ].field_identifier,
@@ -297,7 +375,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 }
                 if ( header_fields[ i ].field_type == SQL_VARCHAR )
                 {
-                    ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                    ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                 src_descriptor -> driver_desc,
                                 0, 
                                 header_fields[ i ].field_identifier,
@@ -310,7 +388,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 {
                     if ( header_fields[ i ].field_type == SQL_INTEGER )
                     {
-                        ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                        ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                         target_descriptor -> driver_desc,
                                         0, 
                                         header_fields[ i ].field_identifier,
@@ -319,7 +397,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     }
                     else if ( header_fields[ i ].field_type == SQL_SMALLINT )
                     {
-                        ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                        ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                         target_descriptor -> driver_desc,
                                         0, 
                                         header_fields[ i ].field_identifier,
@@ -328,7 +406,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     }
                     else if ( header_fields[ i ].field_type == SQL_VARCHAR )
                     {
-                        ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                        ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                         target_descriptor -> driver_desc,
                                         0, 
                                         header_fields[ i ].field_identifier,
@@ -355,7 +433,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 {
                     if ( rec_fields[ i ].field_type == SQL_INTEGER )
                     {
-                        ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                        ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                     src_descriptor -> driver_desc,
                                     rec, 
                                     rec_fields[ i ].field_identifier,
@@ -365,7 +443,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     }
                     else if ( rec_fields[ i ].field_type == SQL_SMALLINT )
                     {
-                        ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                        ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                     src_descriptor -> driver_desc,
                                     rec, 
                                     rec_fields[ i ].field_identifier,
@@ -375,7 +453,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     }
                     if ( rec_fields[ i ].field_type == SQL_VARCHAR )
                     {
-                        ret == SQLGETDESCFIELD( src_descriptor -> connection,
+                        ret = SQLGETDESCFIELD( src_descriptor -> connection,
                                     src_descriptor -> driver_desc,
                                     rec, 
                                     rec_fields[ i ].field_identifier,
@@ -388,7 +466,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     {
                         if ( rec_fields[ i ].field_type == SQL_INTEGER )
                         {
-                            ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                            ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                             target_descriptor -> driver_desc,
                                             rec, 
                                             rec_fields[ i ].field_identifier,
@@ -397,7 +475,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                         }
                         else if ( rec_fields[ i ].field_type == SQL_SMALLINT )
                         {
-                            ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                            ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                             target_descriptor -> driver_desc,
                                             rec, 
                                             rec_fields[ i ].field_identifier,
@@ -406,7 +484,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                         }
                         else if ( rec_fields[ i ].field_type == SQL_VARCHAR )
                         {
-                            ret == SQLSETDESCFIELD( target_descriptor -> connection,
+                            ret = SQLSETDESCFIELD( target_descriptor -> connection,
                                             target_descriptor -> driver_desc,
                                             rec, 
                                             rec_fields[ i ].field_identifier,
@@ -427,7 +505,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
         {
             sprintf( src_descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -441,6 +519,6 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
         else
             thread_release( SQL_HANDLE_ENV, src_descriptor -> connection -> environment );
 
-        return function_return( target_descriptor, ret );
+        return function_return( IGNORE_THREAD, target_descriptor, ret );
     }
 }
