@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,68 @@
  *
  **********************************************************************
  *
- * $Id: SQLDriverConnectW.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLDriverConnectW.c,v 1.18 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLDriverConnectW.c,v $
+ * Revision 1.18  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.17  2009/01/16 11:02:38  lurcher
+ * Interface to GUI for DSN selection
+ *
+ * Revision 1.16  2009/01/12 15:18:15  lurcher
+ * Add interface into odbcinstQ to allow for a dialog if SQLDriverConnect is called without a DSN=
+ *
+ * Revision 1.15  2008/09/29 14:02:44  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.14  2007/02/28 15:37:47  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.13  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.12  2003/08/08 11:14:21  lurcher
+ *
+ * Fix UNICODE problem in SQLDriverConnectW
+ *
+ * Revision 1.11  2002/12/20 11:36:46  lurcher
+ *
+ * Update DMEnvAttr code to allow setting in the odbcinst.ini entry
+ *
+ * Revision 1.10  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.9  2002/10/14 09:46:10  lurcher
+ *
+ * Remove extra return
+ *
+ * Revision 1.8  2002/07/25 09:30:26  lurcher
+ *
+ * Additional unicode and iconv changes
+ *
+ * Revision 1.7  2002/07/24 08:49:51  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.6  2002/07/04 17:27:56  lurcher
+ *
+ * Small bug fixes
+ *
+ * Revision 1.4  2002/05/24 12:42:50  lurcher
+ *
+ * Alter NEWS and ChangeLog to match their correct usage
+ * Additional UNICODE tweeks
+ *
+ * Revision 1.3  2002/01/21 18:00:51  lurcher
+ *
+ * Assorted fixed and changes, mainly UNICODE/bug fixes
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -64,6 +123,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include <string.h>
 #include "drivermanager.h"
 
@@ -74,21 +134,31 @@ int __parse_connection_string_w( struct con_struct *con_str,
 {
 struct con_pair *cp;
 char *local_str, *ptr;
+int len;
 int got_dsn = 0;    /* if we have a DSN then ignore any DRIVER or FILEDSN */
 int got_driver = 0;    /* if we have a DRIVER or FILEDSN then ignore any DSN */
 
     con_str -> count = 0;
     con_str -> list = NULL;
 
-    local_str = unicode_to_ansi( str, str_len );
+    if ( str_len == SQL_NTS )
+    {
+        len = wide_strlen( str );
+        local_str = malloc( len + 1 );
+    }
+    else
+    {
+        len = str_len;
+        local_str = malloc( len + 1 );
+    }
+
+    unicode_to_ansi_copy( local_str, len+1, str, len, NULL, NULL );
 
     if ( !local_str || strlen( local_str ) == 0 ||
         ( strlen( local_str ) == 1 && *local_str == ';' ))
     {
         /* connection-string ::= empty-string [;] */
         free( local_str );
-
-        __append_pair( con_str, "DSN", "DEFAULT" );
         return 0;
     }
 
@@ -104,7 +174,7 @@ int got_driver = 0;    /* if we have a DRIVER or FILEDSN then ignore any DSN */
             got_dsn = 1;
         }
         else if ( strcasecmp( cp -> keyword, "DRIVER" ) == 0 ||
-            strcmp( cp -> keyword, "FILEDSN" ) == 0 )
+            strcasecmp( cp -> keyword, "FILEDSN" ) == 0 )
         {
             if ( got_dsn )
                 continue;
@@ -118,12 +188,6 @@ int got_driver = 0;    /* if we have a DRIVER or FILEDSN then ignore any DSN */
         free( cp );
     }
 
-    /* if no dsn or DRIVER, then set DSN=DEFAULT */
-    if ( !got_driver && !got_dsn )
-    {
-        __append_pair( con_str, "DSN", "DEFAULT" );
-    }
-
     free( local_str );
 
     return 0;
@@ -141,16 +205,21 @@ SQLRETURN SQLDriverConnectW(
 {
     DMHDBC connection = (DMHDBC)hdbc;
     struct con_struct con_struct;
-    char *driver, *dsn;
+    char *driver = NULL, *dsn = NULL;
     char lib_name[ INI_MAX_PROPERTY_VALUE + 1 ];
     char driver_name[ INI_MAX_PROPERTY_VALUE + 1 ];
+	SQLWCHAR local_conn_string[ 1024 ];
+	SQLCHAR local_conn_str_in[ 1024 ];
     SQLRETURN ret_from_connect;
     SQLCHAR s1[ 2048 ];
+    int warnings = 0;
 
     /*
      * check connection
      */
 
+    strcpy( driver_name, "" );
+
     if ( !__validate_dbc( connection ))
     {
         dm_log_write( __FILE__, 
@@ -159,6 +228,40 @@ SQLRETURN SQLDriverConnectW(
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDBC parent_connection;
+
+			parent_connection = find_parent_handle( connection, SQL_HANDLE_DBC );
+
+			if ( parent_connection ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLDRIVERCONNECTW( parent_connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLDRIVERCONNECTW( parent_connection, 
+							connection, 
+							hwnd, 
+							conn_str_in, 
+							len_conn_str_in,
+							conn_str_out,
+							conn_str_out_max,
+							ptr_conn_str_out,
+							driver_completion );
+				}
+			}
+		}
+#endif
         return SQL_INVALID_HANDLE;
     }
 
@@ -167,13 +270,13 @@ SQLRETURN SQLDriverConnectW(
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tWindow Hdl = %p\
-            \n\t\t\tStr In = %s\
-            \n\t\t\tStr Out = %p\
-            \n\t\t\tStr Out Max = %d\
-            \n\t\t\tStr Out Ptr = %p\
-            \n\t\t\tCompletion = %d",
+\n\t\t\tConnection = %p\
+\n\t\t\tWindow Hdl = %p\
+\n\t\t\tStr In = %s\
+\n\t\t\tStr Out = %p\
+\n\t\t\tStr Out Max = %d\
+\n\t\t\tStr Out Ptr = %p\
+\n\t\t\tCompletion = %d",
                 connection,
                 hwnd,
                 __wstring_with_length_hide_pwd( s1, conn_str_in, 
@@ -204,9 +307,7 @@ SQLRETURN SQLDriverConnectW(
                 ERROR_HY090, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( driver_completion == SQL_DRIVER_PROMPT &&
@@ -222,9 +323,7 @@ SQLRETURN SQLDriverConnectW(
                 ERROR_HY092, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( driver_completion != SQL_DRIVER_PROMPT &&
@@ -242,9 +341,7 @@ SQLRETURN SQLDriverConnectW(
                 ERROR_HY110, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     /*
@@ -263,34 +360,90 @@ SQLRETURN SQLDriverConnectW(
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     /*
      * parse the connection string
      */
 
-    if ( !conn_str_in )
-    {
+	if ( driver_completion == SQL_DRIVER_NOPROMPT ) 
+	{
         char *ansi_conn_str_in;
 
-        /*
-         * not quite by the book, but better than nothing
-         */
+    	if ( !conn_str_in )
+    	{
+        	ansi_conn_str_in = "DSN=DEFAULT;";
+        	len_conn_str_in = strlen( ansi_conn_str_in );
 
-        ansi_conn_str_in = "DSN=DEFAULT;";
-        len_conn_str_in = strlen( ansi_conn_str_in );
+			ansi_to_unicode_copy( local_conn_string, ansi_conn_str_in, len_conn_str_in, connection, NULL );
+			conn_str_in = local_conn_string;
 
-        __parse_connection_string( &con_struct,
-                ansi_conn_str_in, len_conn_str_in );
-    }
-    else
-    {
-        __parse_connection_string_w( &con_struct,
-                conn_str_in, len_conn_str_in );
+			__parse_connection_string( &con_struct,
+				ansi_conn_str_in, len_conn_str_in );
+		}
+		else 
+		{
+			__parse_connection_string_w( &con_struct,
+				conn_str_in, len_conn_str_in );
+		}
     }
+	else {
+        char *ansi_conn_str_in;
+
+    	if ( !conn_str_in )
+    	{
+        	ansi_conn_str_in = "";
+        	len_conn_str_in = strlen( ansi_conn_str_in );
+
+    		__parse_connection_string( &con_struct,
+            		ansi_conn_str_in, len_conn_str_in );
+		}
+		else {
+    		__parse_connection_string_w( &con_struct,
+            		conn_str_in, len_conn_str_in );
+		}
+
+		if ( !__get_attribute_value( &con_struct, "DSN" ) && 
+			!__get_attribute_value( &con_struct, "DRIVER" ) && 
+			!__get_attribute_value( &con_struct, "FILEDSN" ))
+		{
+			int ret;
+			SQLCHAR returned_dsn[ 1025 ], *prefix, *target;
+
+			/*
+			 * try and call GUI to obtain a DSN
+			 */
+
+			ret = _SQLDriverConnectPrompt( hwnd, returned_dsn, sizeof( returned_dsn ));
+			if ( !ret || returned_dsn[ 0 ] == '\0' ) 
+			{
+        		__append_pair( &con_struct, "DSN", "DEFAULT" );
+			}
+			else 
+			{
+				prefix = returned_dsn;
+				target = (SQLCHAR*)strchr( (char*)returned_dsn, '=' );
+				if ( target ) 
+				{
+					*target = '\0';
+					target ++;
+        			__append_pair( &con_struct, (char*)prefix, (char*)target );
+				}
+				else {
+        			__append_pair( &con_struct, "DSN", (char*)returned_dsn );
+				}
+			}
+
+			/*
+			 * regenerate to pass to driver
+			 */
+			__generate_connection_string( &con_struct, (char*)local_conn_str_in, sizeof( local_conn_str_in ));
+        	len_conn_str_in = strlen((char*) local_conn_str_in );
+			ansi_to_unicode_copy( local_conn_string, (char*)local_conn_str_in, len_conn_str_in, connection, NULL );
+			conn_str_in = local_conn_string;
+		}
+	}
 
     /*
      * look for some keywords
@@ -309,38 +462,35 @@ SQLRETURN SQLDriverConnectW(
 
         strcpy( driver_name, driver );
 
-        SQLGetPrivateProfileString( driver, "Driver", "",
+#ifdef PLATFORM64
+        SQLGetPrivateProfileString( driver, "Driver64", "",
                 lib_name, sizeof( lib_name ), "ODBCINST.INI" );
 
-        if ( lib_name[ 0 ] == '\0' )
-        {
-            /*
-             * at this point a box could pop up to allow the selection of a driver
-             *
-             * do this later
-             */
+		if ( lib_name[ 0 ] == '\0' )
+		{
+        	SQLGetPrivateProfileString( driver, "Driver", "",
+                	lib_name, sizeof( lib_name ), "ODBCINST.INI" );
+		}
+#else
+        SQLGetPrivateProfileString( driver, "Driver", "",
+                lib_name, sizeof( lib_name ), "ODBCINST.INI" );
+#endif
 
-            dm_log_write( __FILE__, 
-                    __LINE__, 
-                    LOG_INFO, 
-                    LOG_INFO, 
-                    "Error: IM002" );
+        /*
+         * Assume if it's not in a odbcinst.ini then it's a direct reference
+         */
 
-            __post_internal_error( &connection -> error,
-                    ERROR_IM002, NULL,
-                    connection -> environment -> requested_version );
-            __release_conn( &con_struct );
-
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+        if ( lib_name[ 0 ] == '\0' ) {
+            strcpy( lib_name, driver );
         }
 
         strcpy( connection -> dsn, "" );
+        __handle_attr_extensions( connection, NULL, driver_name );
     }
     else
     {
         dsn = __get_attribute_value( &con_struct, "DSN" );
+
         if ( !dsn )
         {
             dm_log_write( __FILE__, 
@@ -354,9 +504,7 @@ SQLRETURN SQLDriverConnectW(
                     connection -> environment -> requested_version );
             __release_conn( &con_struct );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         if ( strlen( dsn ) > SQL_MAX_DSN_LENGTH )
@@ -371,9 +519,7 @@ SQLRETURN SQLDriverConnectW(
                     ERROR_IM012, NULL,
                     connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         /*
@@ -393,12 +539,26 @@ SQLRETURN SQLDriverConnectW(
                     connection -> environment -> requested_version );
             __release_conn( &con_struct );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         strcpy( connection -> dsn, dsn );
+        __handle_attr_extensions( connection, dsn, driver_name );
+    }
+
+    if ( dsn )
+    {
+        /*
+         * do we have any Environment, Connection, or Statement attributes set in the ini ?
+         */
+
+        __handle_attr_extensions( connection, dsn, driver_name );
+    }
+    else {
+        /* 
+         * the attributes may be in the connection string
+         */
+        __handle_attr_extensions_cs( connection, &con_struct );
     }
 
     __release_conn( &con_struct );
@@ -406,10 +566,10 @@ SQLRETURN SQLDriverConnectW(
     /*
      * we have now got the name of a lib to load
      */
-    if ( !__connect_part_one( connection, lib_name, driver_name ))
+    if ( !__connect_part_one( connection, lib_name, driver_name, &warnings ))
     {
-        thread_release( SQL_HANDLE_DBC, connection );
-        return function_return( connection, SQL_ERROR );
+        __disconnect_part_four( connection );       /* release unicode handles */
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( !CHECK_SQLDRIVERCONNECTW( connection ) &&
@@ -422,17 +582,27 @@ SQLRETURN SQLDriverConnectW(
                 "Error: IM001" );
 
         __disconnect_part_one( connection );
+        __disconnect_part_four( connection );       /* release unicode handles */
         __post_internal_error( &connection -> error,
                 ERROR_IM001, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( CHECK_SQLDRIVERCONNECTW( connection ))
     {
+        if ( CHECK_SQLSETCONNECTATTR( connection ))
+        {
+            int lret;
+                
+            lret = SQLSETCONNECTATTR( connection,
+                    connection -> driver_dbc,
+                    SQL_ATTR_ANSI_APP,
+                    SQL_AA_FALSE,
+                    0 );
+        }
+
         ret_from_connect = SQLDRIVERCONNECTW( connection,
                 connection -> driver_dbc,
                 hwnd,
@@ -443,7 +613,7 @@ SQLRETURN SQLDriverConnectW(
                 ptr_conn_str_out,
                 driver_completion );
 
-        if ( !SQL_SUCCEEDED( ret_from_connect ))
+        if ( ret_from_connect != SQL_SUCCESS )
         {
             SQLWCHAR sqlstate[ 6 ];
             SQLINTEGER native_error;
@@ -473,10 +643,11 @@ SQLRETURN SQLDriverConnectW(
 
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex_w( &connection -> error,
+                        __post_internal_error_ex_w_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
                 while( SQL_SUCCEEDED( ret ));
@@ -500,10 +671,11 @@ SQLRETURN SQLDriverConnectW(
 
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex_w( &connection -> error,
+                        __post_internal_error_ex_w_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
                 while( SQL_SUCCEEDED( ret ));
@@ -517,12 +689,11 @@ SQLRETURN SQLDriverConnectW(
             if ( !SQL_SUCCEEDED( ret_from_connect ))
             {
                 __disconnect_part_one( connection );
-
-                thread_release( SQL_HANDLE_DBC, connection );
+                __disconnect_part_four( connection );       /* release unicode handles */
 
                 sprintf( connection -> msg,
                         "\n\t\tExit:[%s]",
-                            __get_return_status( ret_from_connect ));
+                            __get_return_status( ret_from_connect, s1 ));
 
                 dm_log_write( __FILE__,
                         __LINE__,
@@ -530,24 +701,69 @@ SQLRETURN SQLDriverConnectW(
                         LOG_INFO,
                         connection -> msg );
 
-                return function_return( connection, ret_from_connect );
+                return function_return( SQL_HANDLE_DBC, connection, ret_from_connect );
             }
         }
         connection -> unicode_driver = 1;
     }
     else
     {
+        char *in_str, *out_str;
+        int len;
+
+        if ( conn_str_in )
+        {
+            if ( len_conn_str_in == SQL_NTS )
+            {
+                len = wide_strlen( conn_str_in ) + sizeof( SQLWCHAR );
+            }
+            else
+            {
+                len = len_conn_str_in + sizeof( SQLWCHAR );
+            }
+            in_str = malloc( len );
+            unicode_to_ansi_copy( in_str, len, conn_str_in, len, connection, NULL );
+        }
+        else
+        {
+            in_str = NULL;
+        }
+
+        if ( conn_str_out && conn_str_out_max > 0 )
+        {
+            out_str = malloc( conn_str_out_max + sizeof( SQLWCHAR ) );
+        }
+        else
+        {
+            out_str = NULL;
+        }
+       
         ret_from_connect = SQLDRIVERCONNECT( connection,
                 connection -> driver_dbc,
                 hwnd,
-                conn_str_in,
+                (SQLCHAR*)in_str,
                 len_conn_str_in,
-                conn_str_out,
+                (SQLCHAR*)out_str,
                 conn_str_out_max,
                 ptr_conn_str_out,
                 driver_completion );
 
-        if ( !SQL_SUCCEEDED( ret_from_connect ))
+        if ( in_str )
+        {
+            free( in_str );
+        }
+
+        if ( out_str )
+        {
+            if ( SQL_SUCCEEDED( ret_from_connect ))
+            {
+                ansi_to_unicode_copy( conn_str_out, out_str, SQL_NTS, connection, NULL );
+            }
+
+            free( out_str );
+        }
+
+        if ( ret_from_connect != SQL_SUCCESS )
         {
             SQLCHAR sqlstate[ 6 ];
             SQLINTEGER native_error;
@@ -574,13 +790,13 @@ SQLRETURN SQLDriverConnectW(
                             sizeof( message_text ),
                             &ind );
 
-
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex( &connection -> error,
+                        __post_internal_error_ex_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
                 while( SQL_SUCCEEDED( ret ));
@@ -601,13 +817,13 @@ SQLRETURN SQLDriverConnectW(
                             sizeof( message_text ),
                             &ind );
 
-
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex( &connection -> error,
+                        __post_internal_error_ex_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
-                                message_text );
+                                message_text,
+                                SUBCLASS_ODBC, SUBCLASS_ODBC );
                     }
                 }
                 while( SQL_SUCCEEDED( ret ));
@@ -620,12 +836,11 @@ SQLRETURN SQLDriverConnectW(
             if ( !SQL_SUCCEEDED( ret_from_connect ))
             {
                 __disconnect_part_one( connection );
-
-                thread_release( SQL_HANDLE_DBC, connection );
+                __disconnect_part_four( connection );       /* release unicode handles */
 
                 sprintf( connection -> msg,
                         "\n\t\tExit:[%s]",
-                            __get_return_status( ret_from_connect ));
+                            __get_return_status( ret_from_connect, s1 ));
 
                 dm_log_write( __FILE__,
                         __LINE__,
@@ -633,7 +848,7 @@ SQLRETURN SQLDriverConnectW(
                         LOG_INFO,
                         connection -> msg );
 
-                return function_return( connection, ret_from_connect );
+                return function_return( SQL_HANDLE_DBC, connection, ret_from_connect );
             }
         }
         connection -> unicode_driver = 0;
@@ -665,10 +880,9 @@ SQLRETURN SQLDriverConnectW(
     {
         __disconnect_part_two( connection );
         __disconnect_part_one( connection );
+        __disconnect_part_four( connection );       /* release unicode handles */
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( log_info.log_flag )
@@ -678,8 +892,8 @@ SQLRETURN SQLDriverConnectW(
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]\
                     \n\t\t\tConnection Out [%.64s...]",
-                        __get_return_status( ret_from_connect ),
-                        conn_str_out );
+                        __get_return_status( ret_from_connect, s1 ),
+                        __wstring_with_length_hide_pwd( s1, conn_str_out, SQL_NTS ));
         }
         else
         {
@@ -690,9 +904,8 @@ SQLRETURN SQLDriverConnectW(
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]\
                     \n\t\t\tConnection Out [%s]",
-                        __get_return_status( ret_from_connect ),
-                        __wstring_with_length_hide_pwd( s1, 
-                            conn_str_out ? conn_str_out : ansi_to_unicode_in_place((SQLCHAR*) null, SQL_NTS ), SQL_NTS ));
+                        __get_return_status( ret_from_connect, s1 ),
+                        __wstring_with_length_hide_pwd( s1, conn_str_out, SQL_NTS ));
         }
 
         dm_log_write( __FILE__, 
@@ -702,7 +915,10 @@ SQLRETURN SQLDriverConnectW(
                 connection -> msg );
     }
 
-    thread_release( SQL_HANDLE_DBC, connection );
+    if ( warnings && ret_from_connect == SQL_SUCCESS )
+    {
+        ret_from_connect = SQL_SUCCESS_WITH_INFO;
+    }
 
-    return function_return( connection, ret_from_connect );
+    return function_return_nodrv( SQL_HANDLE_DBC, connection, ret_from_connect );
 }
