@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,43 @@
  *
  **********************************************************************
  *
- * $Id: SQLFreeHandle.c,v 1.1.1.1 2001/10/17 16:40:05 lurcher Exp $
+ * $Id: SQLFreeHandle.c,v 1.12 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLFreeHandle.c,v $
+ * Revision 1.12  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.11  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.10  2007/12/17 13:13:03  lurcher
+ * Fix a couple of descriptor typo's
+ *
+ * Revision 1.9  2007/01/02 10:27:50  lurcher
+ * Fix descriptor leak with unicode only driver
+ *
+ * Revision 1.8  2006/04/18 10:24:47  lurcher
+ * Add a couple of changes from Mark Vanderwiel
+ *
+ * Revision 1.7  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.6  2003/05/14 09:42:25  lurcher
+ *
+ * Fix bug in stats collection
+ *
+ * Revision 1.5  2003/04/09 08:42:18  lurcher
+ *
+ * Allow setting of odbcinstQ lib from odbcinst.ini and Environment
+ *
+ * Revision 1.4  2002/09/18 14:49:32  lurcher
+ *
+ * DataManagerII additions and some more threading fixes
+ *
  * Revision 1.1.1.1  2001/10/17 16:40:05  lurcher
  *
  * First upload to SourceForge
@@ -139,13 +173,14 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 #if defined ( COLLECT_STATS ) && defined( HAVE_SYS_SEM_H )
 #include "__stats.h"
 #include <uodbc_stats.h>
 #endif
 
-static char const rcsid[]= "$RCSfile: SQLFreeHandle.c,v $ $Revision: 1.1.1.1 $";
+static char const rcsid[]= "$RCSfile: SQLFreeHandle.c,v $ $Revision: 1.12 $";
 
 SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
         SQLHANDLE handle )
@@ -153,6 +188,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
     switch( handle_type )
     {
       case SQL_HANDLE_ENV:
+      case SQL_HANDLE_SENV:
         {
             DMHENV environment = (DMHENV)handle;
 
@@ -204,13 +240,16 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         environment -> requested_version );
 
-                thread_release( SQL_HANDLE_ENV, environment );
-
-                return function_return( environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             thread_release( SQL_HANDLE_ENV, environment );
 
+            /*
+             * release any pooled connections that are using this environment
+             */
+            __strip_from_pool( environment );
+
             __release_env( environment );
             return SQL_SUCCESS;
         }
@@ -224,6 +263,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
             /*
              * check connection
              */
+
             if ( !__validate_dbc( connection ))
             {
                 dm_log_write( __FILE__, 
@@ -270,9 +310,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_ENV, environment);
-
-                return function_return( environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             environment -> connection_count --;
@@ -304,7 +342,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         environment -> msg );
             }
 #if defined ( COLLECT_STATS ) && defined( HAVE_SYS_SEM_H )
-            uodbc_update_stats(environment->sh, UODBC_STATS_TYPE_HDESC,
+            uodbc_update_stats(environment->sh, UODBC_STATS_TYPE_HDBC,
                                (void *)-1);
 #endif
 
@@ -361,7 +399,10 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                     statement -> state == STATE_S9 ||
                     statement -> state == STATE_S10 ||
                     statement -> state == STATE_S11 ||
-                    statement -> state == STATE_S12 )
+                    statement -> state == STATE_S12 ||
+                    statement -> state == STATE_S13 ||
+                    statement -> state == STATE_S14 ||
+                    statement -> state == STATE_S15 )
             {
                 dm_log_write( __FILE__,
                         __LINE__,
@@ -373,9 +414,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                           ERROR_HY010, NULL,
                           statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
 
             if ( !CHECK_SQLFREEHANDLE( statement -> connection ))
@@ -392,9 +431,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                             ERROR_IM001, NULL,
                             statement -> connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_STMT, statement );
-
-                    return function_return( statement, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
                 }
                 else
                 {
@@ -413,9 +450,12 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
             if ( SQL_SUCCEEDED( ret ))
             {
                 /*
-                 * release the implicit descriptors
+                 * release the implicit descriptors, 
+				 * this matches the tests in SQLAllocHandle
                  */
-                if ( statement -> connection -> driver_act_ver == 3 )
+                if (( statement -> connection -> driver_act_ver == 3 &&
+						CHECK_SQLGETSTMTATTR( connection )) ||
+						CHECK_SQLGETSTMTATTRW( connection ))
                 {
                     if ( statement -> implicit_ard )
                         __release_desc( statement -> implicit_ard );
@@ -453,7 +493,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         connection -> msg );
             }
 
-            return function_return( connection, ret ); 
+            return function_return( IGNORE_THREAD, connection, ret ); 
         }
         break;
 
@@ -462,7 +502,6 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
             DMHDESC descriptor = (DMHDESC)handle;
             DMHDBC connection;
             SQLRETURN ret;
-            SQLCHAR s0[ 20 ];
 
             /*
              * check descriptor
@@ -490,6 +529,21 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         descriptor -> msg );
             }
 
+			if ( descriptor -> implicit )
+			{
+				dm_log_write( __FILE__,
+						__LINE__,
+						LOG_INFO,
+						LOG_INFO,
+						"Error: HY017" );
+		
+				__post_internal_error( &descriptor -> error,
+						ERROR_HY017, NULL,
+						connection -> environment -> requested_version );
+		
+				return function_return_nodrv( IGNORE_THREAD, descriptor, SQL_ERROR );
+			}
+		
             thread_protect( SQL_HANDLE_DESC, descriptor );
 
             if ( !CHECK_SQLFREEHANDLE( connection ))
@@ -504,9 +558,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_IM001, NULL,
                         connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_DESC, descriptor );
-
-                return function_return( descriptor, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
             }
             else
             {
@@ -515,6 +567,29 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         descriptor -> driver_desc );
             }
 
+            /*
+             * check status of statements associated with this descriptor
+             */
+
+            if( __check_stmt_from_desc( descriptor, STATE_S8 ) ||
+                __check_stmt_from_desc( descriptor, STATE_S9 ) ||
+                __check_stmt_from_desc( descriptor, STATE_S10 ) ||
+                __check_stmt_from_desc( descriptor, STATE_S11 ) ||
+                __check_stmt_from_desc( descriptor, STATE_S12 )) {
+
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY010" );
+
+                __post_internal_error( &descriptor -> error,
+                        ERROR_HY010, NULL,
+                        descriptor -> connection -> environment -> requested_version );
+
+                return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+            }
+
             thread_release( SQL_HANDLE_DESC, descriptor );
 
             __release_desc( descriptor );
@@ -535,7 +610,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                                UODBC_STATS_TYPE_HDESC, (void *)-1);
 #endif
 
-            return function_return( connection, SQL_SUCCESS ); 
+            return function_return( IGNORE_THREAD, connection, SQL_SUCCESS ); 
         }
         break;
 
@@ -543,7 +618,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
         /*
          * there is nothing to report a error on
          */
-        return SQL_ERROR;
+        return SQL_INVALID_HANDLE;
     }
 }
 
