@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,57 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetConnectAttr.c,v 1.3 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetConnectAttr.c,v 1.15 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetConnectAttr.c,v $
+ * Revision 1.15  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.14  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.13  2008/09/29 14:02:45  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.12  2006/03/08 09:18:41  lurcher
+ * fix silly typo that was using sizeof( SQL_WCHAR ) instead of SQLWCHAR
+ *
+ * Revision 1.11  2004/11/22 17:02:48  lurcher
+ * Fix unicode/ansi conversion in the SQLGet functions
+ *
+ * Revision 1.10  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.9  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.8  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.7  2002/11/11 17:10:09  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.6  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.5  2002/07/16 13:08:18  lurcher
+ *
+ * Filter attribute values from SQLSetStmtAttr to SQLSetStmtOption to fit
+ * within ODBC 2
+ * Make DSN's double clickable in ODBCConfig
+ *
+ * Revision 1.4  2002/01/30 12:20:02  lurcher
+ *
+ * Add MyODBC 3 driver source
+ *
  * Revision 1.3  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -113,9 +161,23 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLGetConnectAttr.c,v $ $Revision: 1.3 $";
+static char const rcsid[]= "$RCSfile: SQLGetConnectAttr.c,v $ $Revision: 1.15 $";
+
+SQLRETURN SQLGetConnectAttrA( SQLHDBC connection_handle,
+           SQLINTEGER attribute,
+           SQLPOINTER value,
+           SQLINTEGER buffer_length,
+           SQLINTEGER *string_length )
+{
+    return SQLGetConnectAttr( connection_handle,
+                        attribute,
+                        value,
+                        buffer_length,
+                        string_length );
+}
 
 SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
            SQLINTEGER attribute,
@@ -142,7 +204,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
             }
             else
             {
-                *((SQLINTEGER*)value) = SQL_OPT_TRACE_ON;
+                *((SQLINTEGER*)value) = SQL_OPT_TRACE_OFF;
             }
         }
 
@@ -215,11 +277,11 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tBuffer Length = %d\
-            \n\t\t\tStrLen = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tStrLen = %p",
                 connection,
                 __con_attr_as_string( s1, attribute ),
                 value, 
@@ -247,9 +309,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                 ERROR_HY010, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( connection -> state == STATE_C2 )
@@ -263,15 +323,60 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
           case SQL_ATTR_TRACE:
           case SQL_ATTR_TRACEFILE:
           case SQL_ATTR_ASYNC_ENABLE:
-          case SQL_ATTR_AUTO_IPD:
-          case SQL_ATTR_CONNECTION_TIMEOUT:
-          case SQL_ATTR_METADATA_ID:
-          case SQL_ATTR_PACKET_SIZE:
-          case SQL_ATTR_QUIET_MODE:
-          case SQL_ATTR_TXN_ISOLATION:
             break;
 
+          case SQL_ATTR_PACKET_SIZE:
+            if ( connection -> packet_size_set )
+                break;
+          case SQL_ATTR_QUIET_MODE:
+            if ( connection -> quite_mode_set )
+                break;
+
           default:
+            {
+                struct save_attr *sa = connection -> save_attr;
+                while (sa)
+                {
+                    if (sa -> attr_type == attribute)
+                    {
+                        SQLRETURN rc = SQL_SUCCESS;
+                        if (sa -> str_len == SQL_NTS || sa -> str_len > 0)
+                        {
+                            SQLLEN realLen = sa->str_attr ? strlen(sa->str_attr) : 0;
+                            if(value && sa->str_attr)
+                            {
+                                strncpy(value, sa->str_attr, buffer_length - 1);
+                                ((SQLCHAR*)value)[buffer_length - 1] = 0;
+                            }
+                            if(string_length)
+                            {
+                                *string_length = realLen;
+                            }
+                            if(realLen > buffer_length - 1)
+                            {
+                                __post_internal_error( &connection -> error,
+                                ERROR_01004, NULL,
+                                connection -> environment -> requested_version );
+                                rc = SQL_SUCCESS_WITH_INFO;
+                            }
+                        }
+                        else if(buffer_length >= sizeof(SQLLEN))
+                        {
+                            *(SQLLEN*)value = sa -> int_attr;
+                            if(string_length)
+                            {
+                                *string_length = sizeof(SQLLEN);
+                            }
+                        }
+                        else
+                        {
+                            memcpy(value, &sa->int_attr, buffer_length);
+                        }
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, rc );
+                    }
+                    sa = sa -> next;
+                }
+            }
             dm_log_write( __FILE__, 
                     __LINE__, 
                     LOG_INFO, 
@@ -282,9 +387,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                     ERROR_08003, NULL,
                     connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -336,7 +439,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         break;
 
       case SQL_ATTR_ODBC_CURSORS:
-        *((SQLINTEGER*)value) = connection -> cursors;
+        *((SQLULEN*)value) = connection -> cursors;
         type = 1;
         break;
 
@@ -360,7 +463,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         }
         else
         {
-            *((SQLINTEGER*)value) = connection -> async_enable;
+            *((SQLULEN*)value) = connection -> async_enable;
             type = 1;
         }
         break;
@@ -435,7 +538,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         }
         else
         {
-            *((SQLINTEGER*)value) = connection -> quite_mode;
+            *((SQLLEN*)value) = connection -> quite_mode;
             type = 1;
         }
         break;
@@ -496,7 +599,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
 
         sprintf( connection -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -504,13 +607,11 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, ret );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, ret );
     }
     else
     {
-        SQLRETURN ret;
+        SQLRETURN ret = 0;
 
         /*
          * call the driver
@@ -519,12 +620,29 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         {
             if ( !CHECK_SQLGETCONNECTATTRW( connection ))
             {
-                if ( CHECK_SQLGETCONNECTOPTIONW( connection ))
+                if (( ret = CHECK_SQLGETCONNECTOPTIONW( connection )))
                 {
-                    ret = SQLGETCONNECTOPTIONW( connection,
-                        connection -> driver_dbc,
-                        attribute,
-                        value );
+                    SQLWCHAR *s1 = NULL;
+
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
 
                     switch( attribute )
                     {
@@ -533,10 +651,36 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                       case SQL_ATTR_TRANSLATE_LIB:
                         if ( SQL_SUCCEEDED( ret ) && value )
                         {
-                            unicode_to_ansi( value, SQL_NTS );
+                            s1 = malloc( sizeof( SQLWCHAR ) * ( 1024 ));
                         }
                         break;
                     }
+
+                    ret = SQLGETCONNECTOPTIONW( connection,
+                        connection -> driver_dbc,
+                        attribute,
+                        s1 ? s1 : value );
+
+                    switch( attribute )
+                    {
+                      case SQL_ATTR_CURRENT_CATALOG:
+                      case SQL_ATTR_TRACEFILE:
+                      case SQL_ATTR_TRANSLATE_LIB:
+                        if ( SQL_SUCCEEDED( ret ) && value && s1 )
+                        {
+                            unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection, NULL );
+                        }
+						if ( SQL_SUCCEEDED( ret ) && string_length ) 
+						{
+							*string_length /= sizeof( SQLWCHAR );	
+						}
+                        break;
+                    }
+
+                    if ( s1 )
+                    {
+                        free( s1 );
+                    }
                 }
                 else
                 {
@@ -550,18 +694,30 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
             {
+                SQLWCHAR *s1 = NULL;
+
+                switch( attribute )
+                {
+                  case SQL_ATTR_CURRENT_CATALOG:
+                  case SQL_ATTR_TRACEFILE:
+                  case SQL_ATTR_TRANSLATE_LIB:
+                    if ( SQL_SUCCEEDED( ret ) && value && buffer_length > 0 )
+                    {
+                        s1 = malloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ));
+                    }
+                    break;
+                }
+
                 ret = SQLGETCONNECTATTRW( connection,
                     connection -> driver_dbc,
                     attribute,
-                    value,
-                    buffer_length,
+                    s1 ? s1 : value,
+                    s1 ? sizeof( SQLWCHAR ) * buffer_length : buffer_length,
                     string_length );
 
                 switch( attribute )
@@ -569,12 +725,21 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                   case SQL_ATTR_CURRENT_CATALOG:
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
-                    if ( SQL_SUCCEEDED( ret ) && value )
+                    if ( SQL_SUCCEEDED( ret ) && value && s1 )
                     {
-                        unicode_to_ansi( value, SQL_NTS );
+                        unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection, NULL );
                     }
+					if ( SQL_SUCCEEDED( ret ) && string_length ) 
+					{
+						*string_length /= sizeof( SQLWCHAR );	
+					}
                     break;
                 }
+
+                if ( s1 )
+                {
+                    free( s1 );
+                }
             }
         }
         else
@@ -583,6 +748,26 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
             {
                 if ( CHECK_SQLGETCONNECTOPTION( connection ))
                 {
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     ret = SQLGETCONNECTOPTION( connection,
                         connection -> driver_dbc,
                         attribute,
@@ -600,9 +785,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -620,7 +803,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -629,8 +812,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                     connection -> msg );
         }
 
-        thread_release( SQL_HANDLE_DBC, connection );
 
-        return function_return( connection, ret );
+        return function_return( SQL_HANDLE_DBC, connection, ret );
     }
 }
