@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,59 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetConnectAttrW.c,v 1.3 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetConnectAttrW.c,v 1.14 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetConnectAttrW.c,v $
+ * Revision 1.14  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.13  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.12  2008/08/29 08:01:38  lurcher
+ * Alter the way W functions are passed to the driver
+ *
+ * Revision 1.11  2004/11/22 17:02:48  lurcher
+ * Fix unicode/ansi conversion in the SQLGet functions
+ *
+ * Revision 1.10  2003/10/30 18:20:45  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.9  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.8  2002/11/11 17:10:10  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.7  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.6  2002/08/12 13:17:52  lurcher
+ *
+ * Replicate the way the MS DM handles loading of driver libs, and allocating
+ * handles in the driver. usage counting in the driver means that dlopen is
+ * only called for the first use, and dlclose for the last. AllocHandle for
+ * the driver environment is only called for the first time per driver
+ * per application environment.
+ *
+ * Revision 1.5  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.4  2002/07/16 13:08:18  lurcher
+ *
+ * Filter attribute values from SQLSetStmtAttr to SQLSetStmtOption to fit
+ * within ODBC 2
+ * Make DSN's double clickable in ODBCConfig
+ *
  * Revision 1.3  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -67,6 +117,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLGetConnectAttrW.c,v $";
@@ -96,7 +147,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
             }
             else
             {
-                *((SQLINTEGER*)value) = SQL_OPT_TRACE_ON;
+                *((SQLINTEGER*)value) = SQL_OPT_TRACE_OFF;
             }
         }
 
@@ -108,22 +159,23 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
 
         ptr = log_info.log_file_name;
 
-        if ( log_info.log_file_name )
+        if ( ptr )
         {
+            int len = strlen( ptr ) * sizeof( SQLWCHAR );
             if ( string_length )
             {
-                *string_length = strlen( ptr );
+                *string_length = len;
             }
             if ( value )
             {
-                if ( buffer_length > strlen((char*) ptr ) + 1 )
+                if ( buffer_length > len + sizeof( SQLWCHAR ))
                 {
-                    strcpy( value, ptr );
+                    ansi_to_unicode_copy( value, ptr, SQL_NTS, connection, NULL );
                 }
                 else
                 {
-                    memcpy( value, ptr, buffer_length - 1 );
-                    ((char*)value)[ buffer_length - 1 ] = '\0';
+                    ansi_to_unicode_copy( value, ptr, buffer_length - 1, connection, NULL );
+                    ((SQLWCHAR*)value)[( buffer_length - 1 ) / sizeof( SQLWCHAR )] = 0;
                     ret = SQL_SUCCESS_WITH_INFO;
                 }
             }
@@ -136,9 +188,9 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
             }
             if ( value )
             {
-                if ( buffer_length >= 1 )
+                if ( buffer_length > 0 )
                 {
-                    strcpy( value, "" );
+                    ((SQLWCHAR*)value)[ 0 ] = 0;
                 }
                 else
                 {
@@ -146,7 +198,6 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                 }
             }
         }
-        ansi_to_unicode_in_place( value, SQL_NTS );
         return ret;
     }
 
@@ -170,11 +221,11 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tBuffer Length = %d\
-            \n\t\t\tStrLen = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tStrLen = %p",
                 connection,
                 __con_attr_as_string( s1, attribute ),
                 value, 
@@ -202,9 +253,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                 ERROR_HY010, NULL,
                 connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( connection -> state == STATE_C2 )
@@ -218,15 +267,60 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
           case SQL_ATTR_TRACE:
           case SQL_ATTR_TRACEFILE:
           case SQL_ATTR_ASYNC_ENABLE:
-          case SQL_ATTR_AUTO_IPD:
-          case SQL_ATTR_CONNECTION_TIMEOUT:
-          case SQL_ATTR_METADATA_ID:
-          case SQL_ATTR_PACKET_SIZE:
-          case SQL_ATTR_QUIET_MODE:
-          case SQL_ATTR_TXN_ISOLATION:
             break;
 
+          case SQL_ATTR_PACKET_SIZE:
+            if ( connection -> packet_size_set )
+                break;
+          case SQL_ATTR_QUIET_MODE:
+            if ( connection -> quite_mode_set )
+                break;
+
           default:
+            {
+                struct save_attr *sa = connection -> save_attr;
+                while (sa)
+                {
+                    if (sa -> attr_type == attribute)
+                    {
+                        SQLRETURN rc = SQL_SUCCESS;
+                        if (sa -> str_len == SQL_NTS || sa -> str_len > 0)
+                        {
+                            SQLLEN realLen = sa->str_attr ? strlen(sa->str_attr) : 0;
+                            if(value && sa->str_attr)
+                            {
+                                strncpy(value, sa->str_attr, buffer_length - 1);
+                                ((SQLCHAR*)value)[buffer_length - 1] = 0;
+                            }
+                            if(string_length)
+                            {
+                                *string_length = realLen;
+                            }
+                            if(realLen > buffer_length - 1)
+                            {
+                                __post_internal_error( &connection -> error,
+                                ERROR_01004, NULL,
+                                connection -> environment -> requested_version );
+                                rc = SQL_SUCCESS_WITH_INFO;
+                            }
+                        }
+                        else if(buffer_length >= sizeof(SQLLEN))
+                        {
+                            *(SQLLEN*)value = sa -> int_attr;
+                            if(string_length)
+                            {
+                                *string_length = sizeof(SQLLEN);
+                            }
+                        }
+                        else
+                        {
+                            memcpy(value, &sa->int_attr, buffer_length);
+                        }
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, rc );
+                    }
+                    sa = sa -> next;
+                }
+            }
             dm_log_write( __FILE__, 
                     __LINE__, 
                     LOG_INFO, 
@@ -237,9 +331,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                     ERROR_08003, NULL,
                     connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -291,7 +383,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
         break;
 
       case SQL_ATTR_ODBC_CURSORS:
-        *((SQLINTEGER*)value) = connection -> cursors;
+        *((SQLULEN*)value) = connection -> cursors;
         type = 1;
         break;
 
@@ -315,7 +407,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
         }
         else
         {
-            *((SQLINTEGER*)value) = connection -> async_enable;
+            *((SQLULEN*)value) = connection -> async_enable;
             type = 1;
         }
         break;
@@ -451,7 +543,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
 
         sprintf( connection -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -459,23 +551,43 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, ret );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, ret );
     }
     else
     {
-        SQLRETURN ret;
+        SQLRETURN ret = 0;
 
         /*
          * call the driver
          */
-        if ( connection -> unicode_driver )
+        if ( connection -> unicode_driver ||
+			CHECK_SQLGETCONNECTATTRW( connection ) ||
+			CHECK_SQLGETCONNECTOPTIONW( connection ))
         {
             if ( !CHECK_SQLGETCONNECTATTRW( connection ))
             {
                 if ( CHECK_SQLGETCONNECTOPTIONW( connection ))
                 {
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     ret = SQLGETCONNECTOPTIONW( connection,
                         connection -> driver_dbc,
                         attribute,
@@ -493,9 +605,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -512,24 +622,66 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
         {
             if ( !CHECK_SQLGETCONNECTATTR( connection ))
             {
-                if ( CHECK_SQLGETCONNECTOPTION( connection ))
+                if (( ret = CHECK_SQLGETCONNECTOPTION( connection )))
                 {
-                    ret = SQLGETCONNECTOPTION( connection,
-                        connection -> driver_dbc,
-                        attribute,
-                        value );
+                    SQLCHAR *as1 = NULL;
+
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
 
                     switch( attribute )
                     {
                       case SQL_ATTR_CURRENT_CATALOG:
                       case SQL_ATTR_TRACEFILE:
                       case SQL_ATTR_TRANSLATE_LIB:
-                        if ( SQL_SUCCEEDED( ret ) && value )
+                        if ( SQL_SUCCEEDED( ret ) && value && buffer_length > 0 )
                         {
-                            ansi_to_unicode_in_place( value, SQL_NTS );
+                            as1 = malloc( buffer_length + 1 );
                         }
                         break;
                     }
+
+                    ret = SQLGETCONNECTOPTION( connection,
+                        connection -> driver_dbc,
+                        attribute,
+                        as1 ? as1 : value );
+
+                    switch( attribute )
+                    {
+                      case SQL_ATTR_CURRENT_CATALOG:
+                      case SQL_ATTR_TRACEFILE:
+                      case SQL_ATTR_TRANSLATE_LIB:
+                        if ( SQL_SUCCEEDED( ret ) && value && buffer_length > 0 && as1 )
+                        {
+                            ansi_to_unicode_copy( value, (char*) as1, SQL_NTS, connection, NULL );
+                        }
+                        if ( as1 )
+                        {
+                            free( as1 );
+                        }
+						if ( SQL_SUCCEEDED( ret ) && string_length )
+						{
+							*string_length *= sizeof( SQLWCHAR );
+						}
+                        break;
+                    }
                 }
                 else
                 {
@@ -543,26 +695,30 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
             {
+                SQLCHAR *as1 = NULL;
+
                 switch( attribute )
                 {
                   case SQL_ATTR_CURRENT_CATALOG:
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
                     buffer_length = buffer_length / 2;
+                    if ( buffer_length > 0 )
+                    {
+                        as1 = malloc( buffer_length + 1 );
+                    }
                     break;
                 }
 
                 ret = SQLGETCONNECTATTR( connection,
                     connection -> driver_dbc,
                     attribute,
-                    value,
+                    as1 ? as1 : value,
                     buffer_length,
                     string_length );
 
@@ -571,10 +727,18 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                   case SQL_ATTR_CURRENT_CATALOG:
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
-                    if ( SQL_SUCCEEDED( ret ) && value )
+                    if ( SQL_SUCCEEDED( ret ) && value && buffer_length > 0 && as1 )
                     {
-                        ansi_to_unicode_in_place( value, SQL_NTS );
+                        ansi_to_unicode_copy( value, (char*)as1, SQL_NTS, connection, NULL );
                     }
+                    if ( as1 )
+                    {
+                        free( as1 );
+                    }		
+					if ( SQL_SUCCEEDED( ret ) && string_length )
+					{
+						*string_length *= sizeof( SQLWCHAR );
+					}
                     break;
                 }
             }
@@ -584,7 +748,7 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -593,8 +757,6 @@ SQLRETURN SQLGetConnectAttrW( SQLHDBC connection_handle,
                     connection -> msg );
         }
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, ret );
+        return function_return( SQL_HANDLE_DBC, connection, ret );
     }
 }
