@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,60 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetDiagField.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetDiagField.c,v 1.17 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetDiagField.c,v $
+ * Revision 1.17  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.16  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.15  2009/02/04 09:30:01  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.14  2008/09/29 14:02:45  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.13  2007/03/05 09:49:24  lurcher
+ * Get it to build on VMS again
+ *
+ * Revision 1.12  2006/11/27 14:08:33  lurcher
+ * Sync up dirs
+ *
+ * Revision 1.11  2006/05/31 17:35:34  lurcher
+ * Add unicode ODBCINST entry points
+ *
+ * Revision 1.10  2006/03/08 09:18:41  lurcher
+ * fix silly typo that was using sizeof( SQL_WCHAR ) instead of SQLWCHAR
+ *
+ * Revision 1.9  2005/12/19 18:43:26  lurcher
+ * Add new parts to contrib and alter how the errors are returned from the driver
+ *
+ * Revision 1.8  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.7  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.6  2002/11/11 17:10:12  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.5  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.4  2002/01/30 12:20:02  lurcher
+ *
+ * Add MyODBC 3 driver source
+ *
+ * Revision 1.3  2002/01/21 18:00:51  lurcher
+ *
+ * Assorted fixed and changes, mainly UNICODE/bug fixes
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -143,15 +194,39 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLGetDiagField.c,v $ $Revision: 1.2 $";
+static char const rcsid[]= "$RCSfile: SQLGetDiagField.c,v $ $Revision: 1.17 $";
 
 #define ODBC30_SUBCLASS        "01S00,01S01,01S02,01S06,01S07,07S01,08S01,21S01,\
 21S02,25S01,25S02,25S03,42S01,42S02,42S11,42S12,42S21,42S22,HY095,HY097,HY098,\
 HY099,HY100,HY101,HY105,HY107,HY109,HY110,HY111,HYT00,HYT01,IM001,IM002,IM003,\
 IM004,IM005,IM006,IM007,IM008,IM010,IM011,IM012"
 
+extern int __is_env( EHEAD * head );
+
+/*
+ * is it a diag identifier that we have to convert from unicode to ansi
+ */
+
+static int is_char_diag( int diag_identifier )
+{
+    switch( diag_identifier ) {
+        case SQL_DIAG_CLASS_ORIGIN:
+        case SQL_DIAG_CONNECTION_NAME:
+        case SQL_DIAG_DYNAMIC_FUNCTION:
+        case SQL_DIAG_MESSAGE_TEXT:
+        case SQL_DIAG_SERVER_NAME:
+        case SQL_DIAG_SQLSTATE:
+        case SQL_DIAG_SUBCLASS_ORIGIN:
+            return 1;
+
+        default:
+            return 0;
+    }
+}
+
 static SQLRETURN extract_sql_error_field( EHEAD *head,
                 SQLSMALLINT rec_number,
                 SQLSMALLINT diag_identifier,
@@ -161,6 +236,11 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 {
     ERROR *ptr;
 
+    if ( is_char_diag( diag_identifier ) && buffer_length < 0 )
+    {
+        return SQL_ERROR;
+    }
+    
     /*
      * check the header fields first
      */
@@ -170,10 +250,10 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
       case SQL_DIAG_CURSOR_ROW_COUNT:
       case SQL_DIAG_ROW_COUNT:
         {
-            SQLINTEGER val;
+            SQLLEN val;
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 || head -> handle_type != SQL_HANDLE_STMT )
             {
                 return SQL_ERROR;
             }
@@ -184,14 +264,14 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
                   case SQL_DIAG_CURSOR_ROW_COUNT:
                     if ( SQL_SUCCEEDED( head -> diag_cursor_row_count_ret ) && diag_info_ptr )
                     {
-                        *((SQLINTEGER*)diag_info_ptr) = head -> diag_cursor_row_count;
+                        *((SQLLEN*)diag_info_ptr) = head -> diag_cursor_row_count;
                     }
                     return head -> diag_cursor_row_count_ret;
 
                   case SQL_DIAG_ROW_COUNT:
                     if ( SQL_SUCCEEDED( head -> diag_row_count_ret ) && diag_info_ptr )
                     {
-                        *((SQLINTEGER*)diag_info_ptr) = head -> diag_row_count;
+                        *((SQLLEN*)diag_info_ptr) = head -> diag_row_count;
                     }
                     return head -> diag_row_count_ret;
                 }
@@ -251,16 +331,27 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
         {
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 )
             {
                 return SQL_ERROR;
             }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                if ( diag_info_ptr )
+                {
+                    strcpy( diag_info_ptr, "" );
+                }
+                if ( string_length_ptr )
+                {
+                    *string_length_ptr = 0;
+                }
+                return SQL_SUCCESS;
+            }
             else if ( head -> header_set )
             {
                 if ( SQL_SUCCEEDED( head -> diag_dynamic_function_ret ) && diag_info_ptr )
                 {
-                    wide_strcpy( diag_info_ptr, head -> diag_dynamic_function );
-                    unicode_to_ansi( diag_info_ptr, buffer_length );
+                    unicode_to_ansi_copy( diag_info_ptr, buffer_length, head -> diag_dynamic_function, SQL_NTS, __get_connection( head ), NULL );
                     if ( string_length_ptr )
                     {
                         *string_length_ptr = wide_strlen( head -> diag_dynamic_function );
@@ -271,18 +362,35 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
             else if ( __get_connection( head ) -> unicode_driver &&
                 CHECK_SQLGETDIAGFIELDW( __get_connection( head )))
             {
+                SQLWCHAR *s1 = NULL;
+
+                if ( buffer_length > 0 )
+                {
+                    s1 = malloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ));
+                }
+                
                 ret = SQLGETDIAGFIELDW( __get_connection( head ),
                         SQL_HANDLE_STMT,
                         __get_driver_handle( head ),
                         0,
                         diag_identifier,
-                        diag_info_ptr,
-                        buffer_length,
+                        s1 ? s1 : diag_info_ptr,
+                        buffer_length * sizeof ( SQLWCHAR ),
                         string_length_ptr );
 
-                if ( SQL_SUCCEEDED( ret ) && diag_info_ptr )
+                if ( SQL_SUCCEEDED( ret ) && diag_info_ptr && s1 )
                 {
-                    unicode_to_ansi( diag_info_ptr, buffer_length );
+                    unicode_to_ansi_copy( diag_info_ptr, buffer_length, s1, buffer_length, __get_connection( head ), NULL );
+                }
+
+                if ( string_length_ptr && *string_length_ptr > 0 )
+                {
+                    *string_length_ptr /= sizeof ( SQLWCHAR );
+                }
+
+                if ( s1 )
+                {
+                    free( s1 );
                 }
 
                 return ret;
@@ -313,10 +421,15 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
             SQLINTEGER val;
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0  )
             {
                 return SQL_ERROR;
             }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                *((SQLINTEGER*)diag_info_ptr) = 0;
+                return SQL_SUCCESS;
+            }
             else if ( head -> header_set )
             {
                 if ( SQL_SUCCEEDED( head -> diag_dynamic_function_code_ret ) && diag_info_ptr )
@@ -369,6 +482,10 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
         {
             SQLINTEGER val;
             
+            if ( rec_number > 0 )
+            {
+                return SQL_ERROR;
+            }
             val = head -> sql_diag_head.internal_count + 
                 head -> sql_diag_head.error_count;
 
@@ -394,7 +511,9 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
      * else check the records
      */
 
-    if ( rec_number < 1 )
+    if ( rec_number < 1 ||
+        ( diag_identifier == SQL_DIAG_COLUMN_NUMBER ||
+          diag_identifier == SQL_DIAG_ROW_NUMBER ) && head -> handle_type != SQL_HANDLE_STMT )
     {
         return SQL_ERROR;
     }
@@ -411,26 +530,50 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
             ptr = ptr -> next;
             rec_number --;
         }
+		if ( !ptr ) 
+		{
+			return SQL_NO_DATA;
+		}
     }
-    else if ( rec_number <= head -> sql_diag_head.internal_count + 
-            head -> sql_diag_head.error_count )
+    else if ( !__is_env( head ) && __get_connection( head ) -> state != STATE_C2 )
     {
-        rec_number -= head -> sql_diag_head.internal_count;
+		rec_number -= head -> sql_diag_head.internal_count;
 
         if ( __get_connection( head ) -> unicode_driver &&
             CHECK_SQLGETDIAGFIELDW( __get_connection( head )))
         {
             SQLRETURN ret;
+            SQLWCHAR *s1 = NULL;
+
+            if ( diag_info_ptr && buffer_length > 0 )
+            {
+                s1 = malloc( ( buffer_length + 1 ) * sizeof( SQLWCHAR ));
+            }
 
             ret = SQLGETDIAGFIELDW( __get_connection( head ),
                     head -> handle_type,
                     __get_driver_handle( head ),
                     rec_number,
                     diag_identifier,
-                    diag_info_ptr,
-                    buffer_length,
+                    s1 ? s1 : diag_info_ptr,
+                    s1 ? ( buffer_length + 1 ) * sizeof( SQLWCHAR ) : buffer_length,
                     string_length_ptr );
 
+            if ( SQL_SUCCEEDED( ret ) && s1 && diag_info_ptr )
+            {
+                unicode_to_ansi_copy( diag_info_ptr, buffer_length, s1, SQL_NTS, __get_connection( head ), NULL );
+
+                if ( string_length_ptr && *string_length_ptr > 0 ) 
+            {
+                    *string_length_ptr /= sizeof( SQLWCHAR );
+                }
+            }
+
+            if ( s1 )
+			{
+                free( s1 );
+			}
+
             if ( SQL_SUCCEEDED( ret ) && diag_identifier == SQL_DIAG_SQLSTATE )
             {
                 /*
@@ -438,8 +581,6 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
                  */
                 if ( diag_info_ptr )
                 {
-                    unicode_to_ansi( diag_info_ptr, buffer_length );
-
                     if ( diag_info_ptr )
                         __map_error_state( diag_info_ptr, __get_version( head ));
                 }
@@ -447,7 +588,7 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
             return ret;
         }
-        else if ( !__get_connection( head ) -> unicode_driver &&
+		else if ( !__get_connection( head ) -> unicode_driver &&
             CHECK_SQLGETDIAGFIELD( __get_connection( head )))
         {
             SQLRETURN ret;
@@ -483,15 +624,19 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
                 ptr = ptr -> next;
                 rec_number --;
             }
+			if ( !ptr ) 
+			{
+	    		return SQL_NO_DATA;
+			}
         }
     }
-    else
+    else 
     {
-        return SQL_NO_DATA;
+	    return SQL_NO_DATA;
     }
 
     /*
-     * if we are here ptr should point to the error
+     * if we are here ptr should point to the local error
      * record
      */
 
@@ -499,66 +644,27 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
     {
       case SQL_DIAG_CLASS_ORIGIN:
         {
-            char *str;
-            int ret = SQL_SUCCESS;
-
             if ( SQL_SUCCEEDED( ptr -> diag_class_origin_ret ))
             {
-                wide_strcpy( diag_info_ptr, ptr -> diag_class_origin );
-                unicode_to_ansi( diag_info_ptr, buffer_length );
+                unicode_to_ansi_copy( diag_info_ptr, buffer_length, ptr -> diag_class_origin, SQL_NTS, __get_connection( head ), NULL );
                 if ( string_length_ptr )
                 {
                     *string_length_ptr = wide_strlen( ptr -> diag_class_origin );
                 }
                 return ptr -> diag_class_origin_ret;
             }
-
-            if ( wide_ansi_strncmp( ptr -> sqlstate, "IM", 2 ) != 0 )
-            {
-                str = "ISO 9075";
-            }
             else
             {
-                str = "ODBC 3.0";
+                return ptr -> diag_class_origin_ret;
             }
-
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= strlen( str ) + 1 )
-                {
-                    strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, buffer_length - 1 );
-                    (( char * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = strlen( str );
-            }
-            return ret;
         }
         break;
 
       case SQL_DIAG_COLUMN_NUMBER:
         {
-            SQLINTEGER val = SQL_COLUMN_NUMBER_UNKNOWN;
-
-            if ( SQL_SUCCEEDED( ptr -> diag_column_number_ret ))
-            {
-                if ( diag_info_ptr )
-                {
-                    memcpy( diag_info_ptr, &ptr -> diag_column_number, sizeof( val ));
-                }
-                return ptr -> diag_column_number_ret;
-            }
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> diag_column_number, sizeof( SQLINTEGER ));
             }
             return SQL_SUCCESS;
         }
@@ -566,40 +672,19 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
       case SQL_DIAG_CONNECTION_NAME:
         {
-            char *str;
-            int ret = SQL_SUCCESS;
-
-            str = "";
-
             if ( SQL_SUCCEEDED( ptr -> diag_connection_name_ret ))
             {
-                wide_strcpy( diag_info_ptr, ptr -> diag_connection_name );
-                unicode_to_ansi( diag_info_ptr, buffer_length );
+                unicode_to_ansi_copy( diag_info_ptr, buffer_length, ptr -> diag_connection_name, SQL_NTS, __get_connection( head ), NULL );
                 if ( string_length_ptr )
                 {
                     *string_length_ptr = wide_strlen( ptr -> diag_connection_name );
                 }
                 return ptr -> diag_connection_name_ret;
             }
-
-            if ( diag_info_ptr )
+            else
             {
-                if ( buffer_length >= strlen( str ) + 1 )
-                {
-                    strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, buffer_length - 1 );
-                    (( char * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
+                return ptr -> diag_connection_name_ret;
             }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = strlen( str );
-            }
-            return ret;
         }
         break;
 
@@ -608,7 +693,7 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
             char *str;
             int ret = SQL_SUCCESS;
 
-            str = unicode_to_ansi( ptr -> msg, SQL_NTS );
+            str = unicode_to_ansi_alloc( ptr -> msg, SQL_NTS, __get_connection( head ), NULL );
 
             if ( diag_info_ptr )
             {
@@ -628,7 +713,10 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
                 *string_length_ptr = strlen( str );
             }
 
-            ansi_back_to_unicode((char*) ptr -> msg , SQL_NTS );
+            if ( str )
+            {
+                free( str );
+            }
 
             return ret;
         }
@@ -636,11 +724,9 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
       case SQL_DIAG_NATIVE:
         {
-            SQLINTEGER val = ptr -> native_error;
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> native_error, sizeof( SQLINTEGER ));
             }
             return SQL_SUCCESS;
         }
@@ -648,20 +734,9 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
       case SQL_DIAG_ROW_NUMBER:
         {
-            SQLINTEGER val = SQL_ROW_NUMBER_UNKNOWN;
-
-            if ( SQL_SUCCEEDED( ptr -> diag_row_number_ret ))
-            {
-                if ( diag_info_ptr )
-                {
-                    memcpy( diag_info_ptr, &ptr -> diag_row_number, sizeof( val ));
-                }
-                return ptr -> diag_row_number_ret;
-            }
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> diag_row_number, sizeof( SQLLEN ));
             }
             return SQL_SUCCESS;
         }
@@ -669,66 +744,19 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
       case SQL_DIAG_SERVER_NAME:
         {
-            char *str;
-            int ret = SQL_SUCCESS;
-
-            str = "";
-
             if ( SQL_SUCCEEDED( ptr -> diag_server_name_ret ))
             {
-                wide_strcpy( diag_info_ptr, ptr -> diag_server_name );
-                unicode_to_ansi( diag_info_ptr, buffer_length );
+                unicode_to_ansi_copy( diag_info_ptr, buffer_length, ptr -> diag_server_name, SQL_NTS, __get_connection( head ), NULL );
                 if ( string_length_ptr )
                 {
                     *string_length_ptr = wide_strlen( ptr -> diag_server_name );
                 }
                 return ptr -> diag_server_name_ret;
             }
-
-            if ( head -> handle_type == SQL_HANDLE_DBC )              
-            {   
-                if (strlen(((DMHDBC)head->owning_handle)->dsn) > 0)
-                {
-                    str = ((DMHDBC)head->owning_handle)->dsn;
-                }               
-            }
-            else if ( head -> handle_type == SQL_HANDLE_STMT )
+            else
             {
-                DMHSTMT stmt = (DMHSTMT)head->owning_handle;
-                DMHDBC dbc = stmt -> connection;
-                if (strlen(dbc->dsn) > 0)
-                {
-                    str = dbc->dsn;
-                }               
+                return ptr -> diag_server_name_ret;
             }
-            else if ( head -> handle_type == SQL_HANDLE_DESC )
-            {
-                DMHDESC stmt = (DMHDESC)head->owning_handle;
-                DMHDBC dbc = stmt -> connection;
-                if (strlen(dbc->dsn) > 0)
-                {
-                    str = dbc->dsn;
-                }               
-            }
-
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= strlen( str ) + 1 )
-                {
-                    strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, buffer_length - 1 );
-                    (( char * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = strlen( str );
-            }
-            return ret;
         }
         break;
 
@@ -737,7 +765,7 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
             char *str;
             int ret = SQL_SUCCESS;
 
-            str = unicode_to_ansi( ptr -> sqlstate, SQL_NTS );
+            str = unicode_to_ansi_alloc( ptr -> sqlstate, SQL_NTS, __get_connection( head ), NULL );
 
             if ( diag_info_ptr )
             {
@@ -764,7 +792,10 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
                 *string_length_ptr = strlen( str );
             }
 
-            ansi_back_to_unicode((char*) ptr -> sqlstate, SQL_NTS );
+            if ( str )
+            {
+                free( str );
+            }
 
             return ret;
         }
@@ -772,51 +803,19 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
 
       case SQL_DIAG_SUBCLASS_ORIGIN:
         {
-            char *str, *astr;
-            int ret = SQL_SUCCESS;
-
             if ( SQL_SUCCEEDED( ptr -> diag_subclass_origin_ret ))
             {
-                wide_strcpy( diag_info_ptr, ptr -> diag_subclass_origin );
-                unicode_to_ansi( diag_info_ptr, buffer_length );
+                unicode_to_ansi_copy( diag_info_ptr, buffer_length, ptr -> diag_subclass_origin, SQL_NTS, __get_connection( head ), NULL );
                 if ( string_length_ptr )
                 {
                     *string_length_ptr = wide_strlen( ptr -> diag_subclass_origin );
                 }
                 return ptr -> diag_subclass_origin_ret;
             }
-
-            astr = unicode_to_ansi( ptr -> sqlstate, SQL_NTS );
-
-            if ( strstr( ODBC30_SUBCLASS, astr ) != 0 )
-            {
-                str = "ODBC 3.0";
-            }
             else
             {
-                str = "ISO 9075";
+                return ptr -> diag_subclass_origin_ret;
             }
-
-            ansi_back_to_unicode((char*) ptr -> sqlstate, SQL_NTS );
-                                                       
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= strlen( str ) + 1 )
-                {
-                    strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, buffer_length - 1 );
-                    (( char * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = strlen( str );
-            }
-            return ret;
         }
         break;
     }
@@ -824,6 +823,23 @@ static SQLRETURN extract_sql_error_field( EHEAD *head,
     return SQL_SUCCESS;
 }
 
+SQLRETURN SQLGetDiagFieldA( SQLSMALLINT handle_type,
+        SQLHANDLE handle,
+        SQLSMALLINT rec_number,
+        SQLSMALLINT diag_identifier,
+        SQLPOINTER diag_info_ptr,
+        SQLSMALLINT buffer_length,
+        SQLSMALLINT *string_length_ptr )
+{
+    return SQLGetDiagField( handle_type,
+                        handle,
+                        rec_number,
+                        diag_identifier,
+                        diag_info_ptr,
+                        buffer_length,
+                        string_length_ptr );
+}
+
 SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         SQLHANDLE handle,
         SQLSMALLINT rec_number,
@@ -856,12 +872,12 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tEnvironment = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tEnvironment = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     environment,
                     rec_number,
                     diag_identifier,
@@ -887,7 +903,7 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -915,12 +931,12 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tConnection = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     connection,
                     rec_number,
                     diag_identifier,
@@ -946,7 +962,7 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -974,12 +990,12 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tStatement = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     statement,
                     rec_number,
                     diag_identifier,
@@ -1005,7 +1021,7 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -1033,12 +1049,12 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tDescriptor = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tDescriptor = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     descriptor,
                     rec_number,
                     diag_identifier,
@@ -1064,7 +1080,7 @@ SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
