@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,39 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetDiagFieldW.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetDiagFieldW.c,v 1.10 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetDiagFieldW.c,v $
+ * Revision 1.10  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.9  2009/02/04 09:30:02  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.8  2007/02/28 15:37:48  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.7  2002/12/05 17:44:30  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.6  2002/11/11 17:10:14  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.5  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.4  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.3  2002/01/21 18:00:51  lurcher
+ *
+ * Assorted fixed and changes, mainly UNICODE/bug fixes
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -61,6 +91,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLGetDiagFieldW.c,v $";
@@ -70,6 +101,27 @@ static char const rcsid[]= "$RCSfile: SQLGetDiagFieldW.c,v $";
 HY099,HY100,HY101,HY105,HY107,HY109,HY110,HY111,HYT00,HYT01,IM001,IM002,IM003,\
 IM004,IM005,IM006,IM007,IM008,IM010,IM011,IM012"
 
+/*
+ * is it a diag identifier that we have to convert from unicode to ansi
+ */
+
+static int is_char_diag( int diag_identifier )
+{
+    switch( diag_identifier ) {
+        case SQL_DIAG_CLASS_ORIGIN:
+        case SQL_DIAG_CONNECTION_NAME:
+        case SQL_DIAG_DYNAMIC_FUNCTION:
+        case SQL_DIAG_MESSAGE_TEXT:
+        case SQL_DIAG_SERVER_NAME:
+        case SQL_DIAG_SQLSTATE:
+        case SQL_DIAG_SUBCLASS_ORIGIN:
+            return 1;
+
+        default:
+            return 0;
+    }
+}
+
 static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 SQLSMALLINT rec_number,
                 SQLSMALLINT diag_identifier,
@@ -79,6 +131,11 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 {
     ERROR *ptr;
 
+    if ( is_char_diag( diag_identifier ) && buffer_length < 0 )
+    {
+        return SQL_ERROR;
+    }
+
     /*
      * check the header fields first
      */
@@ -91,7 +148,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             SQLINTEGER val;
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 || head -> handle_type != SQL_HANDLE_STMT )
             {
                 return SQL_ERROR;
             }
@@ -169,18 +226,30 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
         {
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 )
             {
                 return SQL_ERROR;
             }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                if ( diag_info_ptr )
+                {
+                    *((SQLWCHAR*)diag_info_ptr) = 0;
+                }
+                if ( string_length_ptr )
+                {
+                    *string_length_ptr = 0;
+                }
+                return SQL_SUCCESS;
+            }
             else if ( head -> header_set )
             {
                 if ( SQL_SUCCEEDED( head -> diag_dynamic_function_ret ) && diag_info_ptr )
                 {
-                    wide_strcpy( diag_info_ptr, head -> diag_dynamic_function );
+                    wide_strncpy( diag_info_ptr, head -> diag_dynamic_function, buffer_length );
                     if ( string_length_ptr )
                     {
-                        *string_length_ptr = wide_strlen( head -> diag_dynamic_function );
+                        *string_length_ptr = wide_strlen( head -> diag_dynamic_function ) * sizeof( SQLWCHAR );
                     }
                 }
                 return head -> diag_dynamic_function_ret;
@@ -202,15 +271,36 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             else if ( !__get_connection( head ) -> unicode_driver &&
                 CHECK_SQLGETDIAGFIELD( __get_connection( head )))
             {
-                ret = SQLGETDIAGFIELDW( __get_connection( head ),
+                SQLCHAR *as1 = NULL;
+
+                if ( buffer_length > 0 && diag_info_ptr )
+                {
+                    as1 = malloc( buffer_length + 1 );
+                }
+
+                ret = SQLGETDIAGFIELD( __get_connection( head ),
                         SQL_HANDLE_STMT,
                         __get_driver_handle( head ),
                         0,
                         diag_identifier,
-                        diag_info_ptr,
-                        buffer_length,
+                        as1 ? as1 : diag_info_ptr,
+                        buffer_length / 2,
                         string_length_ptr );
 
+                if ( SQL_SUCCEEDED( ret ) && as1 && diag_info_ptr )
+                {
+                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ), NULL);
+                }
+                if ( SQL_SUCCEEDED( ret ) && string_length_ptr )
+                {
+                    *string_length_ptr *= sizeof( SQLWCHAR );
+                }
+
+                if ( as1 )
+                {
+                    free( as1 );
+                }
+
                 return ret;
             }
             if ( diag_info_ptr )
@@ -225,10 +315,15 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             SQLINTEGER val;
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0  )
             {
                 return SQL_ERROR;
             }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                *((SQLINTEGER*)diag_info_ptr) = 0;
+                return SQL_SUCCESS;
+            }
             else if ( head -> header_set )
             {
                 if ( SQL_SUCCEEDED( head -> diag_dynamic_function_code_ret ) && diag_info_ptr )
@@ -260,14 +355,9 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                         0,
                         diag_identifier,
                         diag_info_ptr,
-                        buffer_length / 2,
+                        buffer_length,
                         string_length_ptr );
 
-                if ( SQL_SUCCEEDED( ret ))
-                {
-                    ansi_to_unicode_in_place( diag_info_ptr, SQL_NTS );
-                }
-
                 return ret;
             }
             else
@@ -286,6 +376,10 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
         {
             SQLINTEGER val;
             
+            if ( rec_number > 0 )
+            {
+                return SQL_ERROR;
+            }
             val = head -> sql_diag_head.internal_count + 
                 head -> sql_diag_head.error_count;
 
@@ -311,7 +405,9 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
      * else check the records
      */
 
-    if ( rec_number < 1 )
+    if ( rec_number < 1 ||
+        ( diag_identifier == SQL_DIAG_COLUMN_NUMBER ||
+          diag_identifier == SQL_DIAG_ROW_NUMBER ) && head -> handle_type != SQL_HANDLE_STMT )
     {
         return SQL_ERROR;
     }
@@ -328,6 +424,10 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             ptr = ptr -> next;
             rec_number --;
         }
+		if ( !ptr ) 
+		{
+			return SQL_NO_DATA;
+		}
     }
     else if ( rec_number <= head -> sql_diag_head.internal_count + 
             head -> sql_diag_head.error_count )
@@ -364,13 +464,19 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             CHECK_SQLGETDIAGFIELD( __get_connection( head )))
         {
             SQLRETURN ret;
+            SQLCHAR *as1 = NULL;
+
+            if ( is_char_diag( diag_identifier ) && diag_info_ptr && buffer_length > 0 )
+            {
+                as1 = malloc( buffer_length + 1 );
+            }
 
             ret = SQLGETDIAGFIELD( __get_connection( head ),
                     head -> handle_type,
                     __get_driver_handle( head ),
                     rec_number,
                     diag_identifier,
-                    diag_info_ptr,
+                    as1 ? as1 : diag_info_ptr,
                     buffer_length,
                     string_length_ptr );
 
@@ -380,13 +486,18 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                  * map 3 to 2 if required
                  */
 
-                if ( diag_info_ptr )
+                if ( diag_info_ptr && as1 )
                 {
-                    __map_error_state( diag_info_ptr, __get_version( head ));
-                    ansi_to_unicode_in_place( diag_info_ptr, SQL_NTS );
+                    __map_error_state( (char*) as1, __get_version( head ));
+                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ), NULL );
                 }
             }
 
+            if ( as1 )
+            {
+                free( as1 );
+            }
+
             return ret;
         }
         else
@@ -397,6 +508,11 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 ptr = ptr -> next;
                 rec_number --;
             }
+
+			if ( !ptr ) 
+			{
+	    		return SQL_NO_DATA;
+			}
         }
     }
     else
@@ -413,59 +529,27 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
     {
       case SQL_DIAG_CLASS_ORIGIN:
         {
-            SQLWCHAR *str;
-            int ret = SQL_SUCCESS;
-
             if ( SQL_SUCCEEDED( ptr -> diag_class_origin_ret ))
             {
-                wide_strcpy( diag_info_ptr, ptr -> diag_class_origin );
+                wide_strncpy( diag_info_ptr, ptr -> diag_class_origin, buffer_length );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_class_origin );
+                    *string_length_ptr = wide_strlen( ptr -> diag_class_origin ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_class_origin_ret;
             }
-
-            if ( wide_ansi_strncmp( ptr -> sqlstate, "IM", 2 ) != 0 )
-            {
-                str = ansi_to_unicode((SQLCHAR*) "ISO 9075", SQL_NTS );
-            }
             else
             {
-                str = ansi_to_unicode((SQLCHAR*) "ODBC 3.0", SQL_NTS );
+                return ptr -> diag_class_origin_ret;
             }
-
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= wide_strlen( str ) + 1 )
-                {
-                    wide_strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, ( buffer_length - 1 ) * 2 );
-                    (( SQLWCHAR * ) diag_info_ptr )[ buffer_length - 1 ] = 0;
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = wide_strlen( str );
-            }
-
-            free( str );
-
-            return ret;
         }
         break;
 
       case SQL_DIAG_COLUMN_NUMBER:
         {
-            SQLINTEGER val = SQL_COLUMN_NUMBER_UNKNOWN;
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> diag_column_number, sizeof( SQLINTEGER ));
             }
             return SQL_SUCCESS;
         }
@@ -473,41 +557,19 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
       case SQL_DIAG_CONNECTION_NAME:
         {
-            SQLWCHAR *str;
-            int ret = SQL_SUCCESS;
-
-            str = ansi_to_unicode((SQLCHAR*) "", SQL_NTS );
-
             if ( SQL_SUCCEEDED( ptr -> diag_connection_name_ret ))
             {
                 wide_strcpy( diag_info_ptr, ptr -> diag_connection_name );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_connection_name );
+                    *string_length_ptr = wide_strlen( ptr -> diag_connection_name ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_connection_name_ret;
             }
-
-            if ( diag_info_ptr )
+            else
             {
-                if ( buffer_length >= wide_strlen( str ) + 1 )
-                {
-                    wide_strcpy( diag_info_ptr, str );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, str, ( buffer_length - 1 ) * 2 );
-                    (( SQLWCHAR * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
+                return ptr -> diag_connection_name_ret;
             }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = wide_strlen( str );
-            }
-
-            free( str );
-            return ret;
         }
         break;
 
@@ -533,21 +595,18 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             }
             if ( string_length_ptr )
             {
-                *string_length_ptr = wide_strlen( str );
+                *string_length_ptr = wide_strlen( str ) * sizeof( SQLWCHAR );
             }
 
-            free( str );
             return ret;
         }
         break;
 
       case SQL_DIAG_NATIVE:
         {
-            SQLINTEGER val = ptr -> native_error;
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> native_error, sizeof( SQLINTEGER ));
             }
             return SQL_SUCCESS;
         }
@@ -555,20 +614,9 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
       case SQL_DIAG_ROW_NUMBER:
         {
-            SQLINTEGER val = SQL_ROW_NUMBER_UNKNOWN;
-
-            if ( SQL_SUCCEEDED( ptr -> diag_row_number_ret ))
-            {
-                if ( diag_info_ptr )
-                {
-                    memcpy( diag_info_ptr, &ptr -> diag_row_number, sizeof( val ));
-                }
-                return ptr -> diag_row_number_ret;
-            }
-
             if ( diag_info_ptr )
             {
-                memcpy( diag_info_ptr, &val, sizeof( val ));
+                memcpy( diag_info_ptr, &ptr -> diag_row_number, sizeof( SQLINTEGER ));
             }
             return SQL_SUCCESS;
         }
@@ -576,70 +624,19 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
       case SQL_DIAG_SERVER_NAME:
         {
-            char *str;
-            SQLWCHAR *ustr;
-            int ret = SQL_SUCCESS;
-
-            str = "";
-
             if ( SQL_SUCCEEDED( ptr -> diag_server_name_ret ))
             {
                 wide_strcpy( diag_info_ptr, ptr -> diag_server_name );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_server_name );
+                    *string_length_ptr = wide_strlen( ptr -> diag_server_name ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_server_name_ret;
             }
-
-            if ( head -> handle_type == SQL_HANDLE_DBC )              
-            {   
-                if (strlen(((DMHDBC)head->owning_handle)->dsn) > 0)
-                {
-                    str = ((DMHDBC)head->owning_handle)->dsn;
-                }               
-            }
-            else if ( head -> handle_type == SQL_HANDLE_STMT )
+            else
             {
-                DMHSTMT stmt = (DMHSTMT)head->owning_handle;
-                DMHDBC dbc = stmt -> connection;
-                if (strlen(dbc->dsn) > 0)
-                {
-                    str = dbc->dsn;
-                }               
+                return ptr -> diag_server_name_ret;
             }
-            else if ( head -> handle_type == SQL_HANDLE_DESC )
-            {
-                DMHDESC stmt = (DMHDESC)head->owning_handle;
-                DMHDBC dbc = stmt -> connection;
-                if (strlen(dbc->dsn) > 0)
-                {
-                    str = dbc->dsn;
-                }               
-            }
-
-            ustr = ansi_to_unicode((SQLCHAR*) str, SQL_NTS );
-
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= wide_strlen( ustr ) + 1 )
-                {
-                    wide_strcpy( diag_info_ptr, ustr );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, ustr, ( buffer_length - 1 ) * 2 );
-                    (( SQLWCHAR * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = wide_strlen( ustr );
-            }
-
-            free( ustr );
-            return ret;
         }
         break;
 
@@ -672,7 +669,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             }
             if ( string_length_ptr )
             {
-                *string_length_ptr = wide_strlen( str );
+                *string_length_ptr = wide_strlen( str ) * sizeof( SQLWCHAR );
             }
             return ret;
         }
@@ -680,55 +677,19 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
       case SQL_DIAG_SUBCLASS_ORIGIN:
         {
-            char *str, *state;
-            SQLWCHAR *ustr;
-            int ret = SQL_SUCCESS;
-
             if ( SQL_SUCCEEDED( ptr -> diag_subclass_origin_ret ))
             {
                 wide_strcpy( diag_info_ptr, ptr -> diag_subclass_origin );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_subclass_origin );
+                    *string_length_ptr = wide_strlen( ptr -> diag_subclass_origin ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_subclass_origin_ret;
             }
-
-            state = unicode_to_ansi( ptr -> sqlstate, SQL_NTS );
-
-            if ( strstr( ODBC30_SUBCLASS, state ) != 0 )
-            {
-                str = "ODBC 3.0";
-            }
             else
             {
-                str = "ISO 9075";
+                return ptr -> diag_subclass_origin_ret;
             }
-
-            ansi_back_to_unicode((char*) ptr -> sqlstate, SQL_NTS );
-
-            ustr = ansi_to_unicode((SQLCHAR*) str, SQL_NTS );
-                                                       
-            if ( diag_info_ptr )
-            {
-                if ( buffer_length >= wide_strlen( ustr ) + 1 )
-                {
-                    wide_strcpy( diag_info_ptr, ustr );
-                }
-                else
-                {
-                    ret = SQL_SUCCESS_WITH_INFO;
-                    memcpy( diag_info_ptr, ustr, ( buffer_length - 1 ) * 2 );
-                    (( SQLWCHAR * ) diag_info_ptr )[ buffer_length - 1 ] = '\0';
-                }
-            }
-            if ( string_length_ptr )
-            {
-                *string_length_ptr = wide_strlen( ustr );
-            }
-
-            free( ustr );
-            return ret;
         }
         break;
     }
@@ -768,12 +729,12 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tEnvironment = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tEnvironment = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     environment,
                     rec_number,
                     diag_identifier,
@@ -799,7 +760,7 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -824,6 +785,39 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDBC parent_connection;
+
+			parent_connection = find_parent_handle( connection, SQL_HANDLE_DBC );
+
+			if ( parent_connection ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGFIELDW( parent_connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGFIELDW( parent_connection, 
+							handle_type,
+							connection, 
+        					rec_number,
+        					diag_identifier,
+        					diag_info_ptr,
+        					buffer_length,
+        					string_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -833,12 +827,12 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tConnection = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     connection,
                     rec_number,
                     diag_identifier,
@@ -864,7 +858,7 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -889,6 +883,39 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHSTMT parent_statement;
+
+			parent_statement = find_parent_handle( statement, SQL_HANDLE_STMT );
+
+			if ( parent_statement ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGFIELDW( parent_statement -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGFIELDW( parent_statement -> connection, 
+							handle_type,
+							statement, 
+        					rec_number,
+        					diag_identifier,
+        					diag_info_ptr,
+        					buffer_length,
+        					string_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -898,12 +925,12 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tStatement = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     statement,
                     rec_number,
                     diag_identifier,
@@ -929,7 +956,7 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -954,6 +981,39 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDESC parent_desc;
+
+			parent_desc = find_parent_handle( descriptor, SQL_HANDLE_DESC );
+
+			if ( parent_desc ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGFIELDW( parent_desc -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGFIELDW( parent_desc -> connection, 
+							handle_type,
+							descriptor,
+        					rec_number,
+        					diag_identifier,
+        					diag_info_ptr,
+        					buffer_length,
+        					string_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -963,12 +1023,12 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tDescriptor = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tDiag Ident = %p\
-                \n\t\t\tDiag Info Ptr = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tString Len Ptr = %p",
+\n\t\t\tDescriptor = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tDiag Ident = %d\
+\n\t\t\tDiag Info Ptr = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tString Len Ptr = %p",
                     descriptor,
                     rec_number,
                     diag_identifier,
@@ -994,7 +1054,7 @@ SQLRETURN SQLGetDiagFieldW( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
