@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,78 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetDiagRec.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetDiagRec.c,v 1.21 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetDiagRec.c,v $
+ * Revision 1.21  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.20  2009/02/17 09:47:44  lurcher
+ * Clear up a number of bugs
+ *
+ * Revision 1.19  2009/02/04 09:30:02  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.18  2008/09/29 14:02:45  lurcher
+ * Fix missing dlfcn group option
+ *
+ * Revision 1.17  2008/05/20 13:43:47  lurcher
+ * Vms fixes
+ *
+ * Revision 1.16  2007/02/12 11:49:34  lurcher
+ * Add QT4 support to existing GUI parts
+ *
+ * Revision 1.15  2006/11/27 14:08:34  lurcher
+ * Sync up dirs
+ *
+ * Revision 1.14  2006/05/31 17:35:34  lurcher
+ * Add unicode ODBCINST entry points
+ *
+ * Revision 1.13  2006/04/24 08:42:10  lurcher
+ * Handle resetting statement descriptors to implicit values, by passing in NULL or the implicit descrptor  to SQLSetStmtAttr with the attribute SQL_ATTR_APP_PARAM_DESC or SQL_ATTR_APP_ROW_DESC. Also catch trying to call SQLGetDescField on a closed connection
+ *
+ * Revision 1.12  2005/12/19 18:43:26  lurcher
+ * Add new parts to contrib and alter how the errors are returned from the driver
+ *
+ * Revision 1.11  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.10  2003/02/25 13:28:30  lurcher
+ *
+ * Allow errors on the drivers AllocHandle to be reported
+ * Fix a problem that caused errors to not be reported in the log
+ * Remove a redundant line from the spec file
+ *
+ * Revision 1.9  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.8  2002/11/13 15:59:20  lurcher
+ *
+ * More VMS changes
+ *
+ * Revision 1.7  2002/11/11 17:10:15  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.6  2002/10/14 09:46:10  lurcher
+ *
+ * Remove extra return
+ *
+ * Revision 1.5  2002/10/08 13:36:07  lurcher
+ *
+ * Fix memory leak
+ *
+ * Revision 1.4  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.3  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -111,9 +180,19 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLGetDiagRec.c,v $ $Revision: 1.2 $";
+static char const rcsid[]= "$RCSfile: SQLGetDiagRec.c,v $ $Revision: 1.21 $";
+
+int __is_env( EHEAD * head )
+{
+    int type;
+
+    memcpy( &type, head -> owning_handle, sizeof( type ));
+
+    return type == HENV_MAGIC;
+}
 
 DMHDBC __get_connection( EHEAD * head )
 {
@@ -187,7 +266,7 @@ int __get_version( EHEAD * head )
 }
     
 
-SQLHANDLE __get_driver_handle( EHEAD * head )
+DRV_SQLHANDLE __get_driver_handle( EHEAD * head )
 {
     int type;
 
@@ -235,6 +314,7 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
     if ( rec_number <= head -> sql_diag_head.internal_count )
     {
         ERROR *ptr;
+        SQLCHAR *as1 = NULL;
 
         ptr = head -> sql_diag_head.internal_list_head;
         while( rec_number > 1 )
@@ -243,14 +323,18 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
             rec_number --;
         }
 
-        unicode_to_ansi( ptr -> sqlstate, SQL_NTS );
-        unicode_to_ansi( ptr -> msg, SQL_NTS );
+		if ( !ptr ) 
+		{
+	    	return SQL_NO_DATA;
+		}
+
+        as1 = (SQLCHAR*) unicode_to_ansi_alloc( ptr -> msg, SQL_NTS, __get_connection( head ), NULL );
 
         if ( sqlstate )
         {
-            strcpy((char*) sqlstate, (char*) ptr -> sqlstate );
+            unicode_to_ansi_copy((char*) sqlstate, 6, ptr -> sqlstate, SQL_NTS, __get_connection( head ), NULL );
         }
-        if ( buffer_length < strlen( (char*) ptr -> msg ) + 1 )
+        if ( buffer_length < strlen((char*) as1 ) + 1 )
         {
             ret = SQL_SUCCESS_WITH_INFO;
         }
@@ -259,22 +343,22 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
             ret = SQL_SUCCESS;
         }
 
-        if ( message_text )
+        if ( message_text && as1 )
         {
             if ( ret == SQL_SUCCESS )
             {
-                strcpy((char*) message_text, (char*) ptr -> msg );
+                strcpy((char*) message_text, (char*) as1 );
             }
             else
             {
-                memcpy( message_text, ptr -> msg, buffer_length );
+                memcpy( message_text, as1, buffer_length );
                 message_text[ buffer_length - 1 ] = '\0';
             }
         }
 
-        if ( text_length )
+        if ( text_length && as1 )
         {
-            *text_length = strlen( (char*) ptr -> msg );
+            *text_length = strlen((char*) as1 );
         }
 
         if ( native_error )
@@ -287,18 +371,28 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
          */
 
         if ( SQL_SUCCEEDED( ret ) && sqlstate )
-            __map_error_state( sqlstate, __get_version( head ));
- 
-        ansi_back_to_unicode((char*) ptr -> sqlstate, SQL_NTS );
-        ansi_back_to_unicode((char*) ptr -> msg, SQL_NTS );
+            __map_error_state( (char*) sqlstate, __get_version( head ));
 
+        if ( as1 )
+        {
+            free( as1 );
+        }
         return ret;
     }
-    else if ( rec_number <= head -> sql_diag_head.internal_count + 
-            head -> sql_diag_head.error_count )
+    else if ( !__is_env( head ) && __get_connection( head ) -> state != STATE_C2
+        && head->sql_diag_head.error_count )
     {
         ERROR *ptr;
-        rec_number -= head -> sql_diag_head.internal_count;
+        SQLWCHAR *s1 = NULL, *s2 = NULL;
+
+		rec_number -= head -> sql_diag_head.internal_count;
+
+        s1 = malloc( sizeof( SQLWCHAR ) * ( 6 + 1 ));
+
+        if ( buffer_length > 0 )
+        {
+            s2 = malloc( sizeof( SQLWCHAR ) * ( buffer_length + 1 ));
+        }
 
         if ( __get_connection( head ) -> unicode_driver &&
             CHECK_SQLGETDIAGRECW( __get_connection( head )))
@@ -307,9 +401,9 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
                     head -> handle_type,
                     __get_driver_handle( head ),
                     rec_number,
-                    sqlstate,
+                    s1,
                     native_error,
-                    message_text,
+                    s2,
                     buffer_length,
                     text_length );
 
@@ -321,16 +415,15 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
             {
                 if ( sqlstate )
                 {
-                    unicode_to_ansi((SQLWCHAR*) sqlstate, SQL_NTS );
-                    __map_error_state( sqlstate, __get_version( head ));
+                    unicode_to_ansi_copy((char*) sqlstate, 6, s1, SQL_NTS, __get_connection( head ), NULL );
+                    __map_error_state((char*) sqlstate, __get_version( head ));
                 }
                 if ( message_text )
                 {
-                    unicode_to_ansi((SQLWCHAR*) message_text, SQL_NTS );
+                    unicode_to_ansi_copy((char*) message_text, buffer_length, s2, SQL_NTS, __get_connection( head ), NULL );
                 }
             }
 
-            return ret;
         }
         else if ( !__get_connection( head ) -> unicode_driver &&
             CHECK_SQLGETDIAGREC( __get_connection( head )))
@@ -350,12 +443,12 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
              */
 
             if ( SQL_SUCCEEDED( ret ) && sqlstate )
-                __map_error_state( sqlstate, __get_version( head ));
-
-            return ret;
+                __map_error_state((char*) sqlstate, __get_version( head ));
         }
         else
         {
+            SQLCHAR *as1 = NULL;
+
             ptr = head -> sql_diag_head.error_list_head;
             while( rec_number > 1 )
             {
@@ -363,14 +456,24 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
                 rec_number --;
             }
 
-            unicode_to_ansi( ptr -> sqlstate, SQL_NTS );
-            unicode_to_ansi( ptr -> msg, SQL_NTS );
+			if ( !ptr ) 
+			{
+                if ( s1 )
+                    free( s1 );
+
+                if ( s2 )
+                    free( s2 );
+
+	    		return SQL_NO_DATA;
+			}
+
+            as1 = (SQLCHAR*) unicode_to_ansi_alloc( ptr -> msg, SQL_NTS, __get_connection( head ), NULL );
 
             if ( sqlstate )
             {
-                strcpy((char*) sqlstate, (char*) ptr -> sqlstate );
+                unicode_to_ansi_copy((char*) sqlstate, 6, ptr -> sqlstate, SQL_NTS, __get_connection( head ), NULL );
             }
-            if ( buffer_length < strlen((char*) ptr -> msg ) + 1 )
+            if ( as1 && buffer_length < strlen((char*) as1 ) + 1 )
             {
                 ret = SQL_SUCCESS_WITH_INFO;
             }
@@ -379,22 +482,22 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
                 ret = SQL_SUCCESS;
             }
 
-            if ( message_text )
+            if ( message_text && as1 )
             {
                 if ( ret == SQL_SUCCESS )
                 {
-                    strcpy((char*) message_text,(char*) ptr -> msg );
+                    strcpy((char*) message_text,(char*) as1 );
                 }
                 else
                 {
-                    memcpy( message_text, ptr -> msg, buffer_length );
+                    memcpy( message_text, as1, buffer_length );
                     message_text[ buffer_length - 1 ] = '\0';
                 }
             }
 
-            if ( text_length )
+            if ( text_length && as1 )
             {
-                *text_length = strlen((char*) ptr -> msg );
+                *text_length = strlen((char*) as1 );
             }
 
             if ( native_error )
@@ -407,20 +510,47 @@ static SQLRETURN extract_sql_error_rec( EHEAD *head,
              */
 
             if ( SQL_SUCCEEDED( ret ) && sqlstate )
-                __map_error_state( sqlstate, __get_version( head ));
+                __map_error_state((char*) sqlstate, __get_version( head ));
 
-            ansi_back_to_unicode((char*) ptr -> sqlstate, SQL_NTS );
-            ansi_back_to_unicode((char*) ptr -> msg, SQL_NTS );
-
-            return ret;
+            if ( as1 )
+            {
+                free( as1 );
+            }
         }
+
+        if ( s1 )
+            free( s1 );
+
+        if ( s2 )
+            free( s2 );
+
+        return ret;
     }
-    else
+    else 
     {
-        return SQL_NO_DATA;
+	    return SQL_NO_DATA;
     }
 }
 
+SQLRETURN SQLGetDiagRecA( SQLSMALLINT handle_type,
+        SQLHANDLE   handle,
+        SQLSMALLINT rec_number,
+        SQLCHAR     *sqlstate,
+        SQLINTEGER  *native,
+        SQLCHAR     *message_text,
+        SQLSMALLINT buffer_length,
+        SQLSMALLINT *text_length_ptr )
+{
+    return SQLGetDiagRec( handle_type,
+                        handle,
+                        rec_number,
+                        sqlstate,
+                        native,
+                        message_text,
+                        buffer_length,
+                        text_length_ptr );
+}
+
 SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
         SQLHANDLE   handle,
         SQLSMALLINT rec_number,
@@ -432,6 +562,7 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
 {
     SQLRETURN ret;
     SQLCHAR s0[ 32 ], s1[ 100 + LOG_MESSAGE_LEN ];
+    SQLCHAR s2[ 100 + LOG_MESSAGE_LEN ];
 
     if ( rec_number < 1 )
     {
@@ -459,13 +590,13 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tEnvironment = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tEnvironment = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     environment,
                     rec_number,
                     sqlstate,
@@ -495,12 +626,12 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        sqlstate ? sqlstate : (SQLCHAR*)"NULL",
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
                             text_length_ptr, message_text ));
             }
@@ -508,7 +639,7 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
             }
 
             dm_log_write( __FILE__, 
@@ -537,13 +668,13 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tConnection = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     connection,
                     rec_number,
                     sqlstate,
@@ -573,12 +704,12 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        sqlstate ? sqlstate : (SQLCHAR*)"NULL",
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
                             text_length_ptr, message_text ));
             }
@@ -586,7 +717,7 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
             }
 
             dm_log_write( __FILE__, 
@@ -615,13 +746,13 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tStatement = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     statement,
                     rec_number,
                     sqlstate,
@@ -651,12 +782,12 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        sqlstate ? sqlstate : (SQLCHAR*)"NULL",
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
                             text_length_ptr, message_text ));
             }
@@ -664,7 +795,7 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
             }
 
             dm_log_write( __FILE__, 
@@ -693,13 +824,13 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tDescriptor = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tDescriptor = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     descriptor,
                     rec_number,
                     sqlstate,
@@ -729,12 +860,12 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        sqlstate ? sqlstate : (SQLCHAR*)"NULL",
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
                             text_length_ptr, message_text ));
             }
@@ -742,7 +873,7 @@ SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
             {
                 sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
             }
 
             dm_log_write( __FILE__, 
