@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,46 @@
  *
  **********************************************************************
  *
- * $Id: SQLGetDiagRecW.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLGetDiagRecW.c,v 1.11 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLGetDiagRecW.c,v $
+ * Revision 1.11  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.10  2009/02/04 09:30:02  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.9  2007/11/26 11:37:23  lurcher
+ * Sync up before tag
+ *
+ * Revision 1.8  2007/02/28 15:37:48  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.7  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.6  2002/11/11 17:10:17  lurcher
+ *
+ * VMS changes
+ *
+ * Revision 1.5  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.4  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.3  2002/05/21 14:19:44  lurcher
+ *
+ * * Update libtool to escape from AIX build problem
+ * * Add fix to avoid file handle limitations
+ * * Add more UNICODE changes, it looks like it is native 16 representation
+ *   the old way can be reproduced by defining UCS16BE
+ * * Add iusql, its just the same as isql but uses the wide functions
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -56,10 +93,13 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLGetDiagRecW.c,v $";
 
+extern int __is_env( EHEAD * head );        /* in SQLGetDiagRec.c */
+
 static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
         SQLWCHAR *sqlstate,
         SQLINTEGER rec_number,
@@ -74,7 +114,7 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
     {
         SQLWCHAR *tmp;
 
-        tmp = ansi_to_unicode((SQLCHAR*) "00000", SQL_NTS );
+        tmp = ansi_to_unicode_alloc((SQLCHAR*) "00000", SQL_NTS, __get_connection( head ), NULL );
         wide_strcpy( sqlstate, tmp );
         free( tmp );
     }
@@ -90,6 +130,11 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
             rec_number --;
         }
 
+		if ( !ptr ) 
+		{
+	    	return SQL_NO_DATA;
+		}
+
         if ( sqlstate )
         {
             wide_strcpy( sqlstate, ptr -> sqlstate );
@@ -131,12 +176,12 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
          */
 
         if ( SQL_SUCCEEDED( ret ) && sqlstate )
-            __map_error_state( sqlstate, __get_version( head ));
+            __map_error_state_w(sqlstate, __get_version( head ));
 
         return ret;
     }
-    else if ( rec_number <= head -> sql_diag_head.internal_count + 
-            head -> sql_diag_head.error_count )
+    else if ( !__is_env( head ) && __get_connection( head ) -> state != STATE_C2
+        && head->sql_diag_head.error_count )
     {
         ERROR *ptr;
         rec_number -= head -> sql_diag_head.internal_count;
@@ -168,13 +213,25 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
         else if ( !__get_connection( head ) -> unicode_driver &&
             CHECK_SQLGETDIAGREC( __get_connection( head )))
         {
+            SQLCHAR *as1 = NULL, *as2 = NULL;
+
+            if ( sqlstate )
+            {
+                as1 = malloc( 7 );
+            }
+
+            if ( message_text && buffer_length > 0 )
+            {
+                as2 = malloc( buffer_length + 1 );
+            }
+                
             ret = SQLGETDIAGREC( __get_connection( head ),
                     head -> handle_type,
                     __get_driver_handle( head ),
                     rec_number,
-                    sqlstate,
+                    as1 ? as1 : (SQLCHAR *)sqlstate,
                     native_error,
-                    message_text,
+                    as2 ? as2 : (SQLCHAR *)message_text,
                     buffer_length,
                     text_length );
 
@@ -186,15 +243,24 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
             {
                 if ( sqlstate )
                 {
-                    ansi_to_unicode_in_place((SQLCHAR*) sqlstate, SQL_NTS );
-                    __map_error_state_w( sqlstate, __get_version( head ));
+                    if ( as1 )
+                    {
+                        ansi_to_unicode_copy( sqlstate,(char*) as1, SQL_NTS, __get_connection( head ), NULL );
+                        __map_error_state_w( sqlstate, __get_version( head ));
+                    }
                 }
                 if ( message_text )
                 {
-                    ansi_to_unicode_in_place((SQLCHAR*) message_text, SQL_NTS );
+                    if ( as2 )
+                    {
+                        ansi_to_unicode_copy( message_text,(char*) as2, SQL_NTS, __get_connection( head ), NULL );
+                    }
                 }
             }
 
+            if ( as1 ) free( as1 );
+            if ( as2 ) free( as2 );
+
             return ret;
         }
         else
@@ -206,6 +272,11 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
                 rec_number --;
             }
 
+			if ( !ptr ) 
+			{
+	    		return SQL_NO_DATA;
+			}
+
             if ( sqlstate )
             {
                 wide_strcpy( sqlstate, ptr -> sqlstate );
@@ -269,6 +340,8 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
 {
     SQLRETURN ret;
     SQLCHAR s0[ 32 ], s1[ 100 + LOG_MESSAGE_LEN ];
+    SQLCHAR s2[ 100 + LOG_MESSAGE_LEN ];
+    SQLCHAR s3[ 100 + LOG_MESSAGE_LEN ];
 
     if ( rec_number < 1 )
     {
@@ -296,13 +369,13 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             sprintf( environment -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tEnvironment = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tEnvironment = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     environment,
                     rec_number,
                     sqlstate,
@@ -330,22 +403,32 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             if ( SQL_SUCCEEDED( ret ))
             {
+                char *ts1, *ts2;
+
                 sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, NULL, NULL  )),
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, message_text ));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, NULL, NULL  )));
+
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
                 sprintf( environment -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s2 ));
             }
 
             dm_log_write( __FILE__, 
@@ -371,6 +454,40 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDBC parent_connection;
+
+			parent_connection = find_parent_handle( connection, SQL_HANDLE_DBC );
+
+			if ( parent_connection ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGRECW( parent_connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGRECW( parent_connection, 
+							handle_type,
+							connection, 
+        					rec_number,
+        					sqlstate,
+        					native,
+        					message_text,
+        					buffer_length,
+        					text_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -380,13 +497,13 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             sprintf( connection -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tConnection = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     connection,
                     rec_number,
                     sqlstate,
@@ -414,22 +531,32 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             if ( SQL_SUCCEEDED( ret ))
             {
+                char *ts1, *ts2;
+
                 sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection, NULL  )),
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, message_text ));
+                            	text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, connection, NULL  )));
+
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
                 sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s2 ));
             }
 
             dm_log_write( __FILE__, 
@@ -455,6 +582,40 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHSTMT parent_statement;
+
+			parent_statement = find_parent_handle( statement, SQL_HANDLE_STMT );
+
+			if ( parent_statement ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGRECW( parent_statement -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGRECW( parent_statement -> connection, 
+							handle_type,
+							statement,
+        					rec_number,
+        					sqlstate,
+        					native,
+        					message_text,
+        					buffer_length,
+        					text_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -464,13 +625,13 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             sprintf( statement -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tStatement = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tStatement = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     statement,
                     rec_number,
                     sqlstate,
@@ -498,22 +659,32 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             if ( SQL_SUCCEEDED( ret ))
             {
+                char *ts1, *ts2;
+
                 sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, statement -> connection, NULL  )),
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, message_text ));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, statement -> connection, NULL  )));
+
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
                 sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s2 ));
             }
 
             dm_log_write( __FILE__, 
@@ -539,6 +710,40 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDESC parent_desc;
+
+			parent_desc = find_parent_handle( descriptor, SQL_HANDLE_DESC );
+
+			if ( parent_desc ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLGETDIAGRECW( parent_desc -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLGETDIAGRECW( parent_desc -> connection, 
+							handle_type,
+							descriptor,
+        					rec_number,
+        					sqlstate,
+        					native,
+        					message_text,
+        					buffer_length,
+        					text_length_ptr );
+				}
+			}
+		}
+#endif
             return SQL_INVALID_HANDLE;
         }
 
@@ -548,13 +753,13 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             sprintf( descriptor -> msg, 
                 "\n\t\tEntry:\
-                \n\t\t\tDescriptor = %p\
-                \n\t\t\tRec Number = %d\
-                \n\t\t\tSQLState = %p\
-                \n\t\t\tNative = %p\
-                \n\t\t\tMessage Text = %p\
-                \n\t\t\tBuffer Length = %d\
-                \n\t\t\tText Len Ptr = %p",
+\n\t\t\tDescriptor = %p\
+\n\t\t\tRec Number = %d\
+\n\t\t\tSQLState = %p\
+\n\t\t\tNative = %p\
+\n\t\t\tMessage Text = %p\
+\n\t\t\tBuffer Length = %d\
+\n\t\t\tText Len Ptr = %p",
                     descriptor,
                     rec_number,
                     sqlstate,
@@ -582,22 +787,32 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
         {
             if ( SQL_SUCCEEDED( ret ))
             {
+                char *ts1, *ts2;
+
                 sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]\
-                    \n\t\t\tSQLState = %s\
-                    \n\t\t\tNative = %s\
-                    \n\t\t\tMessage Text = %s",
-                        __get_return_status( ret ),
-                        sqlstate,
-                        __ptr_as_string( s0, native ),
+\n\t\t\tSQLState = %s\
+\n\t\t\tNative = %s\
+\n\t\t\tMessage Text = %s",
+                        __get_return_status( ret, s2 ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, descriptor -> connection, NULL  )),
+                        __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, message_text ));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, descriptor -> connection, NULL  )));
+
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
                 sprintf( descriptor -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s2 ));
             }
 
             dm_log_write( __FILE__, 
