@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,64 @@
  *
  **********************************************************************
  *
- * $Id: SQLSetConnectAttr.c,v 1.3 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLSetConnectAttr.c,v 1.17 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLSetConnectAttr.c,v $
+ * Revision 1.17  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.16  2006/04/11 10:22:56  lurcher
+ * Fix a data type check
+ *
+ * Revision 1.15  2005/03/01 15:50:50  lurcher
+ * Add Eric's SQLSetConnectAttr patch
+ *
+ * Revision 1.14  2004/06/21 10:01:11  lurcher
+ *
+ * Fix a couple of 64 bit issues
+ *
+ * Revision 1.13  2003/10/30 18:20:46  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.12  2003/03/05 09:48:44  lurcher
+ *
+ * Add some 64 bit fixes
+ *
+ * Revision 1.11  2003/02/27 12:19:39  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.10  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.9  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.8  2002/07/16 13:08:18  lurcher
+ *
+ * Filter attribute values from SQLSetStmtAttr to SQLSetStmtOption to fit
+ * within ODBC 2
+ * Make DSN's double clickable in ODBCConfig
+ *
+ * Revision 1.7  2002/07/04 17:27:56  lurcher
+ *
+ * Small bug fixes
+ *
+ * Revision 1.5  2002/01/30 12:20:02  lurcher
+ *
+ * Add MyODBC 3 driver source
+ *
+ * Revision 1.4  2002/01/10 11:17:20  lurcher
+ *
+ * Allow SQL_ATTR_LOGIN_TIMEOUT to be set when connected to mirror what the
+ * MS DM does
+ *
  * Revision 1.3  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -152,9 +207,21 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLSetConnectAttr.c,v $ $Revision: 1.3 $";
+static char const rcsid[]= "$RCSfile: SQLSetConnectAttr.c,v $ $Revision: 1.17 $";
+
+SQLRETURN SQLSetConnectAttrA( SQLHDBC connection_handle,
+           SQLINTEGER attribute,
+           SQLPOINTER value,
+           SQLINTEGER string_length )
+{
+    return SQLSetConnectAttr( connection_handle,
+           attribute,
+           value,
+           string_length );
+}
 
 SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
            SQLINTEGER attribute,
@@ -170,7 +237,31 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
      */
     if ( attribute == SQL_ATTR_TRACE )
     {
-        if ((SQLINTEGER) value == SQL_OPT_TRACE_OFF )
+        if ((SQLLEN) value != SQL_OPT_TRACE_OFF && 
+            (SQLLEN) value != SQL_OPT_TRACE_ON ) 
+        {
+            if ( __validate_dbc( connection ))
+            {
+                thread_protect( SQL_HANDLE_DBC, connection );
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY024" );
+        
+                __post_internal_error( &connection -> error,
+                    ERROR_HY024, NULL,
+                    connection -> environment -> requested_version );
+        
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            else 
+            {
+                return SQL_INVALID_HANDLE;
+            }
+        }
+
+        if ((SQLLEN) value == SQL_OPT_TRACE_OFF )
         {
             char force_string[ 30 ];
 
@@ -216,11 +307,58 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
     {
         if ( value )
         {
-            if ( log_info.log_file_name )
+            if (((SQLCHAR*)value)[ 0 ] == '\0' ) 
             {
-                free( log_info.log_file_name );
+                if ( __validate_dbc( connection ))
+                {
+                    thread_protect( SQL_HANDLE_DBC, connection );
+                    dm_log_write( __FILE__, 
+                            __LINE__, 
+                            LOG_INFO, 
+                            LOG_INFO, 
+                            "Error: HY024" );
+            
+                    __post_internal_error( &connection -> error,
+                        ERROR_HY024, NULL,
+                        connection -> environment -> requested_version );
+            
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
+                else 
+                {
+                    return SQL_INVALID_HANDLE;
+                }
+            }
+            else 
+            {
+                if ( log_info.log_file_name )
+                {
+                    free( log_info.log_file_name );
+                }
+                log_info.log_file_name = strdup( value );
+            }
+        }
+        else 
+        {
+            if ( __validate_dbc( connection ))
+            {
+                thread_protect( SQL_HANDLE_DBC, connection );
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY024" );
+        
+                __post_internal_error( &connection -> error,
+                    ERROR_HY024, NULL,
+                    connection -> environment -> requested_version );
+        
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            else 
+            {
+                return SQL_INVALID_HANDLE;
             }
-            log_info.log_file_name = strdup( value );
         }
         return SQL_SUCCESS;
     }
@@ -245,14 +383,14 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tStrLen = %p",
+\n\t\t\tConnection = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tStrLen = %d",
                 connection,
                 __con_attr_as_string( s1, attribute ),
                 value, 
-                (void*)string_length );
+                (int)string_length );
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -278,9 +416,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_08003, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else if ( connection -> state == STATE_C3 )
@@ -295,9 +431,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
             ERROR_HY010, NULL,
             connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
     else if ( connection -> state == STATE_C4 ||
             connection -> state == STATE_C5 ||
@@ -315,9 +449,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
         else if ( attribute == SQL_ATTR_PACKET_SIZE )
         {
@@ -331,26 +463,107 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_HY011, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
-        else if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
-        {
-            dm_log_write( __FILE__, 
+    }
+
+    /*
+     * is it a legitimate value
+     */
+    ret = dm_check_connection_attrs( connection, attribute, value );
+
+    if ( ret != SQL_SUCCESS ) 
+    {
+        dm_log_write( __FILE__, 
                     __LINE__, 
                     LOG_INFO, 
                     LOG_INFO, 
-                    "Error: HY011" );
+                    "Error: HY024" );
 
-            __post_internal_error( &connection -> error,
-                ERROR_HY011, NULL,
+        __post_internal_error( &connection -> error,
+                ERROR_HY024, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+    }
 
-            return function_return( connection, SQL_ERROR );
-        }
+    /*
+     * is it a connection attribute or statement, check state of any active connections
+     */
+
+    switch( attribute ) 
+    {
+        /* ODBC 3.x statement attributes are not settable at the connection level */
+        case SQL_ATTR_APP_PARAM_DESC:
+        case SQL_ATTR_APP_ROW_DESC:
+      	case SQL_ATTR_CURSOR_SCROLLABLE:
+      	case SQL_ATTR_CURSOR_SENSITIVITY:
+        case SQL_ATTR_ENABLE_AUTO_IPD:
+        case SQL_ATTR_FETCH_BOOKMARK_PTR:
+        case SQL_ATTR_IMP_PARAM_DESC:
+        case SQL_ATTR_IMP_ROW_DESC:
+        case SQL_ATTR_PARAM_BIND_OFFSET_PTR:
+        case SQL_ATTR_PARAM_BIND_TYPE:
+        case SQL_ATTR_PARAM_OPERATION_PTR:
+        case SQL_ATTR_PARAM_STATUS_PTR:
+        case SQL_ATTR_PARAMS_PROCESSED_PTR:
+        case SQL_ATTR_PARAMSET_SIZE:
+        case SQL_ATTR_ROW_ARRAY_SIZE:
+        case SQL_ATTR_ROW_BIND_OFFSET_PTR:
+        case SQL_ATTR_ROW_OPERATION_PTR:
+        case SQL_ATTR_ROW_STATUS_PTR:
+        case SQL_ATTR_ROWS_FETCHED_PTR:
+            __post_internal_error( &connection -> error,
+                    ERROR_HY092, NULL,
+                    connection -> environment -> requested_version );
+
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+
+      	case SQL_ATTR_CONCURRENCY:
+	  	case SQL_BIND_TYPE:
+      	case SQL_ATTR_CURSOR_TYPE:
+      	case SQL_ATTR_MAX_LENGTH:
+      	case SQL_MAX_ROWS:
+      	case SQL_ATTR_KEYSET_SIZE:
+      	case SQL_ROWSET_SIZE: 
+      	case SQL_ATTR_NOSCAN:
+      	case SQL_ATTR_QUERY_TIMEOUT:
+      	case SQL_ATTR_RETRIEVE_DATA:
+      	case SQL_ATTR_SIMULATE_CURSOR:
+      	case SQL_ATTR_USE_BOOKMARKS:
+            if( __check_stmt_from_dbc_v( connection, 5, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12 )) {
+
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: 24000" );
+
+                __post_internal_error( &connection -> error,
+                        ERROR_24000, NULL,
+                        connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            break;
+
+        default:
+
+            if( __check_stmt_from_dbc_v( connection, 5, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12 )) {
+
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY010" );
+
+                __post_internal_error( &connection -> error,
+                        ERROR_HY010, NULL,
+                        connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            break;
     }
 
     /*
@@ -359,6 +572,15 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
     value = __attr_override( connection, SQL_HANDLE_DBC, attribute, value, &string_length );
 
+    /*
+     * we need to save this even if connected so we can use it for the next connect
+     */
+    if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
+    {
+        connection -> login_timeout = ( SQLLEN ) value;
+        connection -> login_timeout_set = 1;
+    }
+
     /*
      * if connected, call the driver
      * otherwise we need to save the states and set them when we
@@ -372,59 +594,54 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
         if ( attribute == SQL_ATTR_ODBC_CURSORS )
         {
-            connection -> cursors = ( SQLINTEGER ) value;
+            connection -> cursors = ( SQLLEN ) value;
         }
         else if ( attribute == SQL_ATTR_ACCESS_MODE )
         {
-            connection -> access_mode = ( SQLINTEGER ) value;
+            connection -> access_mode = ( SQLLEN ) value;
             connection -> access_mode_set = 1;
         }
         else if ( attribute == SQL_ATTR_ASYNC_ENABLE )
         {
-            connection -> async_enable = ( SQLINTEGER ) value;
+            connection -> async_enable = ( SQLLEN ) value;
             connection -> async_enable_set = 1;
         }
         else if ( attribute == SQL_ATTR_AUTO_IPD )
         {
-            connection -> auto_ipd = ( SQLINTEGER ) value;
+            connection -> auto_ipd = ( SQLLEN ) value;
             connection -> auto_ipd_set = 1;
         }
         else if ( attribute == SQL_ATTR_AUTOCOMMIT )
         {
-            connection -> auto_commit = ( SQLINTEGER ) value;
+            connection -> auto_commit = ( SQLLEN ) value;
             connection -> auto_commit_set = 1;
         }
         else if ( attribute == SQL_ATTR_CONNECTION_TIMEOUT )
         {
-            connection -> connection_timeout = ( SQLINTEGER ) value;
+            connection -> connection_timeout = ( SQLLEN ) value;
             connection -> connection_timeout_set = 1;
         }
-        else if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
-        {
-            connection -> login_timeout = ( SQLINTEGER ) value;
-            connection -> login_timeout_set = 1;
-        }
         else if ( attribute == SQL_ATTR_METADATA_ID )
         {
-            connection -> metadata_id = ( SQLINTEGER ) value;
+            connection -> metadata_id = ( SQLLEN ) value;
             connection -> metadata_id_set = 1;
         }
         else if ( attribute == SQL_ATTR_PACKET_SIZE )
         {
-            connection -> packet_size = ( SQLINTEGER ) value;
+            connection -> packet_size = ( SQLLEN ) value;
             connection -> packet_size_set = 1;
         }
         else if ( attribute == SQL_ATTR_QUIET_MODE )
         {
-            connection -> quite_mode = ( SQLINTEGER ) value;
+            connection -> quite_mode = ( SQLLEN ) value;
             connection -> quite_mode_set = 1;
         }
         else if ( attribute == SQL_ATTR_TXN_ISOLATION )
         {
-            connection -> txn_isolation = ( SQLINTEGER ) value;
+            connection -> txn_isolation = ( SQLLEN ) value;
             connection -> txn_isolation_set = 1;
         }
-        else
+        else if ( attribute != SQL_ATTR_LOGIN_TIMEOUT )
         {
             /*
              * save any unknown attributes untill connect
@@ -441,12 +658,23 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
             }
             else if ( string_length == SQL_NTS )
             {
+                if (!value)
+                {
+                    __post_internal_error( &connection -> error,
+                        ERROR_HY024, "Invalid argument value",
+                        connection -> environment -> requested_version );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
+                else
+                {
                 sa -> str_attr = strdup( value );
                 sa -> str_len = string_length;
             }
+            }
             else
             {
-                sa -> int_attr = ( SQLINTEGER ) value;
+                sa -> int_attr = ( SQLLEN ) value;
+                sa -> str_len = string_length;
             }
             sa -> next = connection -> save_attr;
             connection -> save_attr = sa;
@@ -454,7 +682,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
         sprintf( connection -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( SQL_SUCCESS ));
+                    __get_return_status( SQL_SUCCESS, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -462,9 +690,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return SQL_SUCCESS;
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
     }
     else
     {
@@ -474,6 +700,26 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
             {
                 if ( CHECK_SQLSETCONNECTOPTION( connection ))
                 {
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     ret = SQLSETCONNECTOPTION( connection,
                             connection -> driver_dbc,
                             attribute,
@@ -491,9 +737,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -513,12 +757,32 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 {
                     SQLWCHAR *s1;
 
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                    __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     switch( attribute )
                     {
                       case SQL_ATTR_CURRENT_CATALOG:
                       case SQL_ATTR_TRACEFILE:
                       case SQL_ATTR_TRANSLATE_LIB:
-                        s1 = ansi_to_unicode( value, SQL_NTS );
+                        s1 = ansi_to_unicode_alloc( value, SQL_NTS, connection, NULL );
 
                         ret = SQLSETCONNECTOPTIONW( connection,
                                 connection -> driver_dbc,
@@ -527,6 +791,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
                         if ( s1 )
                             free( s1 );
+
                         break;
 
                       default:
@@ -549,9 +814,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -563,13 +826,13 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                   case SQL_ATTR_CURRENT_CATALOG:
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
-                    s1 = ansi_to_unicode( value, string_length );
+                    s1 = ansi_to_unicode_alloc( value, string_length, connection, NULL );
 
                     ret = SQLSETCONNECTATTRW( connection,
                         connection -> driver_dbc,
                         attribute,
                         s1,
-                        string_length );
+                        string_length == SQL_NTS ? SQL_NTS : string_length * sizeof( SQLWCHAR ));
 
                     if ( s1 )
                         free( s1 );
@@ -590,7 +853,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -606,10 +869,8 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
     if ( attribute == SQL_ATTR_USE_BOOKMARKS && SQL_SUCCEEDED( ret ))
     {
-        connection -> bookmarks_on = (SQLUINTEGER) value;
+        connection -> bookmarks_on = (SQLLEN) value;
     }
 
-    thread_release( SQL_HANDLE_DBC, connection );
-
-    return function_return( connection, ret );
+    return function_return( SQL_HANDLE_DBC, connection, ret );
 }
