@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,59 @@
  *
  **********************************************************************
  *
- * $Id: SQLSetConnectAttrW.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLSetConnectAttrW.c,v 1.14 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLSetConnectAttrW.c,v $
+ * Revision 1.14  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.13  2008/08/29 08:01:39  lurcher
+ * Alter the way W functions are passed to the driver
+ *
+ * Revision 1.12  2007/02/28 15:37:48  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.11  2006/04/18 10:24:47  lurcher
+ * Add a couple of changes from Mark Vanderwiel
+ *
+ * Revision 1.10  2003/10/30 18:20:46  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.9  2003/03/05 09:48:44  lurcher
+ *
+ * Add some 64 bit fixes
+ *
+ * Revision 1.8  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.7  2002/08/23 09:42:37  lurcher
+ *
+ * Fix some build warnings with casts, and a AIX linker mod, to include
+ * deplib's on the link line, but not the libtool generated ones
+ *
+ * Revision 1.6  2002/07/24 08:49:52  lurcher
+ *
+ * Alter UNICODE support to use iconv for UNICODE-ANSI conversion
+ *
+ * Revision 1.5  2002/07/16 13:08:18  lurcher
+ *
+ * Filter attribute values from SQLSetStmtAttr to SQLSetStmtOption to fit
+ * within ODBC 2
+ * Make DSN's double clickable in ODBCConfig
+ *
+ * Revision 1.4  2002/07/04 17:27:56  lurcher
+ *
+ * Small bug fixes
+ *
+ * Revision 1.3  2002/01/30 12:20:02  lurcher
+ *
+ * Add MyODBC 3 driver source
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -72,6 +122,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLSetConnectAttrW.c,v $";
@@ -84,6 +135,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
     DMHDBC connection = (DMHDBC)connection_handle;
     SQLRETURN ret;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
+	SQLWCHAR buffer[ 512 ];
 
     /*
      * doesn't require a handle
@@ -91,7 +143,31 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
 
     if ( attribute == SQL_ATTR_TRACE )
     {
-        if ((SQLINTEGER) value == SQL_OPT_TRACE_OFF )
+        if ((SQLLEN) value != SQL_OPT_TRACE_OFF && 
+            (SQLLEN) value != SQL_OPT_TRACE_ON ) 
+        {
+            if ( __validate_dbc( connection ))
+            {
+                thread_protect( SQL_HANDLE_DBC, connection );
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY024" );
+        
+                __post_internal_error( &connection -> error,
+                    ERROR_HY024, NULL,
+                    connection -> environment -> requested_version );
+        
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            else 
+            {
+                return SQL_INVALID_HANDLE;
+            }
+        }
+
+        if ((SQLLEN) value == SQL_OPT_TRACE_OFF )
         {
             char force_string[ 30 ];
 
@@ -137,12 +213,58 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
     {
         if ( value )
         {
-            if ( log_info.log_file_name )
+            if (((SQLWCHAR*)value)[ 0 ] == 0 ) 
             {
-                free( log_info.log_file_name );
+                if ( __validate_dbc( connection ))
+                {
+                    thread_protect( SQL_HANDLE_DBC, connection );
+                    dm_log_write( __FILE__, 
+                            __LINE__, 
+                            LOG_INFO, 
+                            LOG_INFO, 
+                            "Error: HY024" );
+            
+                    __post_internal_error( &connection -> error,
+                        ERROR_HY024, NULL,
+                        connection -> environment -> requested_version );
+            
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
+                else 
+                {
+                    return SQL_INVALID_HANDLE;
+                }
+            }
+            else 
+            {
+                if ( log_info.log_file_name )
+                {
+                    free( log_info.log_file_name );
+                }
+                log_info.log_file_name = unicode_to_ansi_alloc((SQLWCHAR *) value, SQL_NTS, connection, NULL );
+            }
+        }
+        else 
+        {
+            if ( __validate_dbc( connection ))
+            {
+                thread_protect( SQL_HANDLE_DBC, connection );
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY024" );
+        
+                __post_internal_error( &connection -> error,
+                    ERROR_HY024, NULL,
+                    connection -> environment -> requested_version );
+        
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            else 
+            {
+                return SQL_INVALID_HANDLE;
             }
-            unicode_to_ansi( value, SQL_NTS );
-            log_info.log_file_name = strdup( value );
         }
         return SQL_SUCCESS;
     }
@@ -159,6 +281,36 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHDBC parent_connection;
+
+			parent_connection = find_parent_handle( connection, SQL_HANDLE_DBC );
+
+			if ( parent_connection ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLSETCONNECTATTRW( parent_connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+					return SQLSETCONNECTATTRW( parent_connection, 
+							connection_handle, 
+           					attribute,
+           					value,
+           					string_length );
+				}
+			}
+		}
+#endif
         return SQL_INVALID_HANDLE;
     }
 
@@ -167,14 +319,14 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
     if ( log_info.log_flag )
     {
         sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tStrLen = %d",
+\n\t\t\tConnection = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tStrLen = %d",
                 connection,
                 __con_attr_as_string( s1, attribute ),
                 value, 
-                string_length );
+                (int)string_length );
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -200,9 +352,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                 ERROR_08003, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else if ( connection -> state == STATE_C3 )
@@ -217,9 +367,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
             ERROR_HY010, NULL,
             connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return function_return( connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
     else if ( connection -> state == STATE_C4 ||
             connection -> state == STATE_C5 || 
@@ -237,9 +385,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
         else if ( attribute == SQL_ATTR_PACKET_SIZE )
         {
@@ -253,17 +399,123 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                 ERROR_HY011, NULL,
                 connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_DBC, connection );
-
-            return function_return( connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
+    /*
+     * is it a legitimate value
+     */
+    ret = dm_check_connection_attrs( connection, attribute, value );
+
+    if ( ret != SQL_SUCCESS ) 
+    {
+        dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY024" );
+
+        __post_internal_error( &connection -> error,
+                ERROR_HY024, NULL,
+                connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+    }
+
+    /*
+     * is it a connection attribute or statement, check state of any active connections
+     */
+
+    switch( attribute ) 
+    {
+        /* ODBC 3.x statement attributes are not settable at the connection level */
+        case SQL_ATTR_APP_PARAM_DESC:
+        case SQL_ATTR_APP_ROW_DESC:
+      	case SQL_ATTR_CURSOR_SCROLLABLE:
+      	case SQL_ATTR_CURSOR_SENSITIVITY:
+        case SQL_ATTR_ENABLE_AUTO_IPD:
+        case SQL_ATTR_FETCH_BOOKMARK_PTR:
+        case SQL_ATTR_IMP_PARAM_DESC:
+        case SQL_ATTR_IMP_ROW_DESC:
+        case SQL_ATTR_PARAM_BIND_OFFSET_PTR:
+        case SQL_ATTR_PARAM_BIND_TYPE:
+        case SQL_ATTR_PARAM_OPERATION_PTR:
+        case SQL_ATTR_PARAM_STATUS_PTR:
+        case SQL_ATTR_PARAMS_PROCESSED_PTR:
+        case SQL_ATTR_PARAMSET_SIZE:
+        case SQL_ATTR_ROW_ARRAY_SIZE:
+        case SQL_ATTR_ROW_BIND_OFFSET_PTR:
+        case SQL_ATTR_ROW_OPERATION_PTR:
+        case SQL_ATTR_ROW_STATUS_PTR:
+        case SQL_ATTR_ROWS_FETCHED_PTR:
+            __post_internal_error( &connection -> error,
+                    ERROR_HY092, NULL,
+                    connection -> environment -> requested_version );
+
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+
+      	case SQL_ATTR_CONCURRENCY:
+	  	case SQL_BIND_TYPE:
+      	case SQL_ATTR_CURSOR_TYPE:
+      	case SQL_ATTR_MAX_LENGTH:
+      	case SQL_MAX_ROWS:
+      	case SQL_ATTR_KEYSET_SIZE:
+      	case SQL_ROWSET_SIZE: 
+      	case SQL_ATTR_NOSCAN:
+      	case SQL_ATTR_QUERY_TIMEOUT:
+      	case SQL_ATTR_RETRIEVE_DATA:
+      	case SQL_ATTR_SIMULATE_CURSOR:
+      	case SQL_ATTR_USE_BOOKMARKS:
+            if( __check_stmt_from_dbc_v( connection, 8, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12, STATE_S13, STATE_S14, STATE_S15 )) {
+
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: 24000" );
+
+                __post_internal_error( &connection -> error,
+                        ERROR_24000, NULL,
+                        connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            break;
+
+        default:
+
+            if( __check_stmt_from_dbc_v( connection, 8, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12, STATE_S13, STATE_S14, STATE_S15 )) {
+
+                dm_log_write( __FILE__, 
+                        __LINE__, 
+                        LOG_INFO, 
+                        LOG_INFO, 
+                        "Error: HY010" );
+
+                __post_internal_error( &connection -> error,
+                        ERROR_HY010, NULL,
+                        connection -> environment -> requested_version );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            }
+            break;
+    }
+
     /*
      * is it something overridden
      */
 
-    value = __attr_override( connection, SQL_HANDLE_DBC, attribute, value, &string_length );
+    value = __attr_override_wide( connection, SQL_HANDLE_DBC, attribute, value, &string_length, buffer );
+
+    /*
+     * we need to save this even if connected so we can use it for the next connect
+     */
+    if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
+    {
+        connection -> login_timeout = ( SQLLEN ) value;
+        connection -> login_timeout_set = 1;
+    }
 
     /*
      * if connected, call the driver
@@ -278,59 +530,54 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
 
         if ( attribute == SQL_ATTR_ODBC_CURSORS )
         {
-            connection -> cursors = ( SQLINTEGER ) value;
+            connection -> cursors = ( SQLLEN ) value;
         }
         else if ( attribute == SQL_ATTR_ACCESS_MODE )
         {
-            connection -> access_mode = ( SQLINTEGER ) value;
+            connection -> access_mode = ( SQLLEN ) value;
             connection -> access_mode_set = 1;
         }
         else if ( attribute == SQL_ATTR_ASYNC_ENABLE )
         {
-            connection -> async_enable = ( SQLINTEGER ) value;
+            connection -> async_enable = ( SQLLEN ) value;
             connection -> async_enable_set = 1;
         }
         else if ( attribute == SQL_ATTR_AUTO_IPD )
         {
-            connection -> auto_ipd = ( SQLINTEGER ) value;
+            connection -> auto_ipd = ( SQLLEN ) value;
             connection -> auto_ipd_set = 1;
         }
         else if ( attribute == SQL_ATTR_AUTOCOMMIT )
         {
-            connection -> auto_commit = ( SQLINTEGER ) value;
+            connection -> auto_commit = ( SQLLEN ) value;
             connection -> auto_commit_set = 1;
         }
         else if ( attribute == SQL_ATTR_CONNECTION_TIMEOUT )
         {
-            connection -> connection_timeout = ( SQLINTEGER ) value;
+            connection -> connection_timeout = ( SQLLEN ) value;
             connection -> connection_timeout_set = 1;
         }
-        else if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
-        {
-            connection -> login_timeout = ( SQLINTEGER ) value;
-            connection -> login_timeout_set = 1;
-        }
         else if ( attribute == SQL_ATTR_METADATA_ID )
         {
-            connection -> metadata_id = ( SQLINTEGER ) value;
+            connection -> metadata_id = ( SQLLEN ) value;
             connection -> metadata_id_set = 1;
         }
         else if ( attribute == SQL_ATTR_PACKET_SIZE )
         {
-            connection -> packet_size = ( SQLINTEGER ) value;
+            connection -> packet_size = ( SQLLEN ) value;
             connection -> packet_size_set = 1;
         }
         else if ( attribute == SQL_ATTR_QUIET_MODE )
         {
-            connection -> quite_mode = ( SQLINTEGER ) value;
+            connection -> quite_mode = ( SQLLEN ) value;
             connection -> quite_mode_set = 1;
         }
         else if ( attribute == SQL_ATTR_TXN_ISOLATION )
         {
-            connection -> txn_isolation = ( SQLINTEGER ) value;
+            connection -> txn_isolation = ( SQLLEN ) value;
             connection -> txn_isolation_set = 1;
         }
-        else
+        else if ( attribute != SQL_ATTR_LOGIN_TIMEOUT )
         {
             /*
              * save any unknown attributes untill connect
@@ -347,12 +594,22 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
             }
             else if ( string_length == SQL_NTS )
             {
-                sa -> str_attr = strdup( value );
+                if (!value)
+                {
+                    __post_internal_error( &connection -> error,
+                        ERROR_HY024, "Invalid argument value",
+                        connection -> environment -> requested_version );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
+                else
+                {
+                sa -> str_attr = unicode_to_ansi_alloc( value, string_length, connection, NULL );
                 sa -> str_len = string_length;
             }
+            }
             else
             {
-                sa -> int_attr = ( SQLINTEGER ) value;
+                sa -> int_attr = ( SQLLEN ) value;
             }
             sa -> next = connection -> save_attr;
             connection -> save_attr = sa;
@@ -360,7 +617,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
 
         sprintf( connection -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( SQL_SUCCESS ));
+                    __get_return_status( SQL_SUCCESS, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -368,18 +625,37 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        thread_release( SQL_HANDLE_DBC, connection );
-
-        return SQL_SUCCESS;
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
     }
     else
     {
-        if ( connection -> unicode_driver )
+        if ( connection -> unicode_driver ||
+			CHECK_SQLSETCONNECTATTRW( connection ))
         {
             if ( !CHECK_SQLSETCONNECTATTRW( connection ))
             {
                 if ( CHECK_SQLSETCONNECTOPTIONW( connection ))
                 {
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     ret = SQLSETCONNECTOPTIONW( connection,
                             connection -> driver_dbc,
                             attribute,
@@ -397,10 +673,8 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
-                    }
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
             }
             else
             {
@@ -417,6 +691,28 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
             {
                 if ( CHECK_SQLSETCONNECTOPTION( connection ))
                 {
+                    SQLCHAR *as1 = NULL;
+
+                    /*
+                     * Is it in the legal range of values
+                     */
+
+                    if ( attribute < SQL_CONN_DRIVER_MIN && 
+                            ( attribute > SQL_PACKET_SIZE || attribute < SQL_ACCESS_MODE ))
+                    {
+                        dm_log_write( __FILE__, 
+                                __LINE__, 
+                                LOG_INFO, 
+                                LOG_INFO, 
+                                "Error: HY092" );
+
+                        __post_internal_error( &connection -> error,
+                                ERROR_HY092, NULL,
+                                connection -> environment -> requested_version );
+
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    }
+
                     switch( attribute )
                     {
                       case SQL_ATTR_CURRENT_CATALOG:
@@ -424,7 +720,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                       case SQL_ATTR_TRANSLATE_LIB:
                         if ( value )
                         {
-                            unicode_to_ansi( value, SQL_NTS );
+                            as1 = (SQLCHAR*) unicode_to_ansi_alloc( value, SQL_NTS, connection, NULL );
                         }
                         break;
                     }
@@ -432,19 +728,9 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                     ret = SQLSETCONNECTOPTION( connection,
                             connection -> driver_dbc,
                             attribute,
-                            value );
+                            as1 ? as1 : value );
 
-                    switch( attribute )
-                    {
-                      case SQL_ATTR_CURRENT_CATALOG:
-                      case SQL_ATTR_TRACEFILE:
-                      case SQL_ATTR_TRANSLATE_LIB:
-                        if ( SQL_SUCCEEDED( ret ) && value )
-                        {
-                            ansi_back_to_unicode((char*) value, SQL_NTS );
-                        }
-                        break;
-                    }
+                    if ( as1 ) free( as1 );
                 }
                 else
                 {
@@ -458,13 +744,13 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    thread_release( SQL_HANDLE_DBC, connection );
-
-                    return function_return( connection, SQL_ERROR );
-                    }
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
             }
             else
             {
+                SQLCHAR *as1 = NULL;
+
                 switch( attribute )
                 {
                   case SQL_ATTR_CURRENT_CATALOG:
@@ -472,26 +758,34 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
                   case SQL_ATTR_TRANSLATE_LIB:
                     if ( value )
                     {
-                        unicode_to_ansi( value, SQL_NTS );
+                        if ( string_length > 0 )
+                        {
+                            as1 = (SQLCHAR*) unicode_to_ansi_alloc( value, string_length, connection, NULL );
+                        }
+                        else if ( string_length == SQL_NTS )
+                        {
+                            as1 = (SQLCHAR*) unicode_to_ansi_alloc( value, SQL_NTS, connection, NULL );
+                        }
+                    }
+
+                    ret = SQLSETCONNECTATTR( connection,
+                            connection -> driver_dbc,
+                            attribute,
+                            as1 ? as1 : value,
+                            string_length / sizeof( SQLWCHAR ));
+
+                    if ( as1 ) 
+                    {
+                        free( as1 );
                     }
                     break;
-                }
 
-                ret = SQLSETCONNECTATTR( connection,
-                        connection -> driver_dbc,
-                        attribute,
-                        value,
-                        string_length );
-
-                switch( attribute )
-                {
-                  case SQL_ATTR_CURRENT_CATALOG:
-                  case SQL_ATTR_TRACEFILE:
-                  case SQL_ATTR_TRANSLATE_LIB:
-                    if ( SQL_SUCCEEDED( ret ) && value )
-                    {
-                        ansi_back_to_unicode((char*) value, SQL_NTS );
-                    }
+                  default:
+                    ret = SQLSETCONNECTATTR( connection,
+                            connection -> driver_dbc,
+                            attribute,
+                            value,
+                            string_length );
                     break;
                 }
             }
@@ -501,7 +795,7 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
         {
             sprintf( connection -> msg, 
                     "\n\t\tExit:[%s]",
-                        __get_return_status( ret ));
+                        __get_return_status( ret, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -517,10 +811,8 @@ SQLRETURN SQLSetConnectAttrW( SQLHDBC connection_handle,
 
     if ( attribute == SQL_ATTR_USE_BOOKMARKS && SQL_SUCCEEDED( ret ))
     {
-        connection -> bookmarks_on = (SQLUINTEGER) value;
+        connection -> bookmarks_on = (SQLULEN) value;
     }
 
-    thread_release( SQL_HANDLE_DBC, connection );
-
-    return function_return( connection, ret );
+    return function_return( SQL_HANDLE_DBC, connection, ret );
 }
