@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,59 @@
  *
  **********************************************************************
  *
- * $Id: SQLSetStmtAttr.c,v 1.2 2001/12/13 13:00:32 lurcher Exp $
+ * $Id: SQLSetStmtAttr.c,v 1.16 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLSetStmtAttr.c,v $
+ * Revision 1.16  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.15  2009/02/04 09:30:02  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.14  2007/12/17 13:13:03  lurcher
+ * Fix a couple of descriptor typo's
+ *
+ * Revision 1.13  2007/02/12 11:49:34  lurcher
+ * Add QT4 support to existing GUI parts
+ *
+ * Revision 1.12  2006/04/27 16:39:50  lurcher
+ * fix missing return from SQLSetStmtAttr changes
+ *
+ * Revision 1.11  2006/04/24 08:42:10  lurcher
+ * Handle resetting statement descriptors to implicit values, by passing in NULL or the implicit descrptor  to SQLSetStmtAttr with the attribute SQL_ATTR_APP_PARAM_DESC or SQL_ATTR_APP_ROW_DESC. Also catch trying to call SQLGetDescField on a closed connection
+ *
+ * Revision 1.10  2005/11/23 08:29:16  lurcher
+ * Add cleanup in postgres driver
+ *
+ * Revision 1.9  2003/10/30 18:20:46  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.8  2003/03/05 09:48:45  lurcher
+ *
+ * Add some 64 bit fixes
+ *
+ * Revision 1.7  2003/02/27 12:19:40  lurcher
+ *
+ * Add the A functions as well as the W
+ *
+ * Revision 1.6  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.5  2002/09/18 14:49:32  lurcher
+ *
+ * DataManagerII additions and some more threading fixes
+ *
+ * Revision 1.3  2002/07/16 13:08:18  lurcher
+ *
+ * Filter attribute values from SQLSetStmtAttr to SQLSetStmtOption to fit
+ * within ODBC 2
+ * Make DSN's double clickable in ODBCConfig
+ *
  * Revision 1.2  2001/12/13 13:00:32  lurcher
  *
  * Remove most if not all warnings on 64 bit platforms
@@ -157,9 +207,21 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
-static char const rcsid[]= "$RCSfile: SQLSetStmtAttr.c,v $ $Revision: 1.2 $";
+static char const rcsid[]= "$RCSfile: SQLSetStmtAttr.c,v $ $Revision: 1.16 $";
+
+SQLRETURN SQLSetStmtAttrA( SQLHSTMT statement_handle,
+           SQLINTEGER attribute,
+           SQLPOINTER value,
+           SQLINTEGER string_length )
+{
+    return SQLSetStmtAttr( statement_handle,
+           attribute,
+           value,
+           string_length );
+}
 
 SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
            SQLINTEGER attribute,
@@ -190,10 +252,10 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, "\n\t\tEntry:\
-            \n\t\t\tStatement = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tStrLen = %d",
+\n\t\t\tStatement = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tStrLen = %d",
                 statement,
                 __stmt_attr_as_string( s1, attribute ),
                 value, 
@@ -232,9 +294,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                  ERROR_HY011, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
         else if ( statement -> state == STATE_S4 ||
                 statement -> state == STATE_S5 ||
@@ -251,15 +311,16 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                  ERROR_24000, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
         else if ( statement -> state == STATE_S8 ||
                 statement -> state == STATE_S9 ||
                 statement -> state == STATE_S10 ||
                 statement -> state == STATE_S11 ||
-                statement -> state == STATE_S12 )
+                statement -> state == STATE_S12 ||
+                statement -> state == STATE_S13 ||
+                statement -> state == STATE_S14 ||
+                statement -> state == STATE_S15 )
         {
             if ( statement -> prepared )
             {
@@ -273,9 +334,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                      ERROR_HY011, NULL,
                      statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
             else
             {
@@ -289,9 +348,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                      ERROR_HY010, NULL,
                      statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
     }
@@ -313,13 +370,12 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                  ERROR_HY010, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
-    if ( !CHECK_SQLSETSTMTATTR( statement -> connection ) &&
+    if ( (!CHECK_SQLSETSTMTATTR( statement -> connection ) &&
+          !CHECK_SQLSETSTMTATTRW( statement -> connection )) &&
         !CHECK_SQLSETSTMTOPTION( statement -> connection ))
     {
         dm_log_write( __FILE__, 
@@ -332,9 +388,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                 ERROR_IM001, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
@@ -345,13 +399,94 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
     {
         DMHDESC desc = ( DMHDESC ) value;
 
+		/*
+		 * needs to reset to implicit descriptor, this is safe
+		 * without a validate, as the value is either null, or the
+		 * same as a descriptor we know is valid
+		 */
+
+		if ( desc == NULL || desc == statement -> implicit_ard ) 
+		{
+			DRV_SQLHDESC drv_desc = NULL;
+
+			ret = SQL_SUCCESS;
+			
+			if ( desc == statement -> implicit_ard )
+			{
+				drv_desc = statement -> implicit_ard -> driver_desc;
+			}
+
+        	if ( CHECK_SQLSETSTMTATTR( statement -> connection ))
+        	{
+           	 	ret = SQLSETSTMTATTR( statement -> connection,
+           	     	statement -> driver_stmt,
+           	     	attribute,
+           	     	drv_desc,
+           	     	0 );
+        	}
+        	else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+        	{
+            	ret = SQLSETSTMTATTRW( statement -> connection,
+                	statement -> driver_stmt,
+                	attribute,
+                	statement -> implicit_ard -> driver_desc,
+                	0 );
+        	}
+            else
+            {
+                ret = SQLSETSTMTOPTION( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    statement -> implicit_ard -> driver_desc );
+            }
+
+			if ( ret != SQL_SUCCESS ) 
+			{
+    			if ( log_info.log_flag )
+    			{
+        			sprintf( statement -> msg, 
+                			"\n\t\tExit:[%s]",
+                    			__get_return_status( ret, s1 ));
+			
+        			dm_log_write( __FILE__, 
+                			__LINE__, 
+                			LOG_INFO, 
+                			LOG_INFO, 
+                			statement -> msg );
+    			}
+			
+    			return function_return( SQL_HANDLE_STMT, statement, ret ); 
+			}
+			
+			/*
+			 * copy DM descriptor
+			 */
+
+			statement -> apd = statement -> implicit_apd;
+
+    		if ( log_info.log_flag )
+    		{
+        		sprintf( statement -> msg, 
+                		"\n\t\tExit:[%s]",
+                    		__get_return_status( ret, s1 ));
+
+        		dm_log_write( __FILE__, 
+                		__LINE__, 
+                		LOG_INFO, 
+                		LOG_INFO, 
+                		statement -> msg );
+    		}
+		
+    		return function_return( SQL_HANDLE_STMT, statement, ret ); 
+		}
+
         if ( !__validate_desc( desc ))
         {
             thread_release( SQL_HANDLE_STMT, statement );
 
             sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                    __get_return_status( SQL_INVALID_HANDLE ));
+                    __get_return_status( SQL_INVALID_HANDLE, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -375,9 +510,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     ERROR_HY017, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         if ( desc -> connection !=
@@ -393,9 +526,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     ERROR_HY024, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         /*
@@ -403,17 +534,99 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
          */
         value = ( SQLPOINTER ) desc -> driver_desc;
         statement -> ard = desc;
+        desc -> associated_with = statement;
     }
 
     if ( attribute == SQL_ATTR_APP_PARAM_DESC )
     {
         DMHDESC desc = ( DMHDESC ) value;
 
+		/*
+		 * needs to reset to implicit descriptor, this is safe
+		 * without a validate, as the value is either null, or the
+		 * same as a descriptor we know is valid
+		 */
+
+		if ( desc == NULL || desc == statement -> implicit_apd ) 
+		{
+			DRV_SQLHDESC drv_desc = NULL;
+
+			ret = SQL_SUCCESS;
+			
+			if ( desc == statement -> implicit_apd )
+			{
+				drv_desc = statement -> implicit_apd -> driver_desc;
+			}
+
+        	if ( CHECK_SQLSETSTMTATTR( statement -> connection ))
+        	{
+           	 	ret = SQLSETSTMTATTR( statement -> connection,
+           	     	statement -> driver_stmt,
+           	     	attribute,
+                	statement -> implicit_apd -> driver_desc,
+           	     	0 );
+        	}
+        	else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+        	{
+            	ret = SQLSETSTMTATTRW( statement -> connection,
+                	statement -> driver_stmt,
+                	attribute,
+                	statement -> implicit_apd -> driver_desc,
+                	0 );
+        	}
+            else
+            {
+                ret = SQLSETSTMTOPTION( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    drv_desc );
+            }
+
+			if ( ret != SQL_SUCCESS ) 
+			{
+    			if ( log_info.log_flag )
+    			{
+        			sprintf( statement -> msg, 
+                			"\n\t\tExit:[%s]",
+                    			__get_return_status( ret, s1 ));
+			
+        			dm_log_write( __FILE__, 
+                			__LINE__, 
+                			LOG_INFO, 
+                			LOG_INFO, 
+                			statement -> msg );
+    			}
+			
+    			return function_return( SQL_HANDLE_STMT, statement, ret ); 
+			}
+			
+			/*
+			 * copy DM descriptor
+			 */
+
+			statement -> apd = statement -> implicit_apd;
+
+    		if ( log_info.log_flag )
+    		{
+        		sprintf( statement -> msg, 
+                		"\n\t\tExit:[%s]",
+                    		__get_return_status( ret, s1 ));
+
+        		dm_log_write( __FILE__, 
+                		__LINE__, 
+                		LOG_INFO, 
+                		LOG_INFO, 
+                		statement -> msg );
+    		}
+		
+    		return function_return( SQL_HANDLE_STMT, statement, ret ); 
+		}
+
         if ( !__validate_desc( desc ))
         {
             sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                    __get_return_status( SQL_INVALID_HANDLE ));
+                    __get_return_status( SQL_INVALID_HANDLE, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -439,9 +652,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     ERROR_HY017, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         if ( desc -> connection !=
@@ -457,9 +668,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     ERROR_HY024, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         /*
@@ -467,6 +676,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
          */
         value = ( SQLPOINTER ) desc -> driver_desc;
         statement -> apd = desc;
+        desc -> associated_with = statement;
     }
 
     /*
@@ -475,7 +685,43 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
 
     if ( attribute == SQL_ATTR_METADATA_ID )
     {
-        statement -> metadata_id = (int) value;
+        statement -> metadata_id = (SQLLEN) value;
+    }
+
+    if ( attribute == SQL_ATTR_IMP_ROW_DESC || 
+        attribute == SQL_ATTR_IMP_PARAM_DESC )
+    {
+        dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY017" );
+
+        __post_internal_error( &statement -> error,
+                ERROR_HY017, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    /*
+     * is it a legitimate value
+     */
+    ret = dm_check_statement_attrs( statement, attribute, value );
+
+    if ( ret != SQL_SUCCESS ) 
+    {
+        dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY024" );
+
+        __post_internal_error( &statement -> error,
+                ERROR_HY024, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
@@ -493,7 +739,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
             CHECK_SQLEXTENDEDFETCH( statement -> connection ) &&
             !CHECK_SQLFETCHSCROLL( statement -> connection ))
     {
-        statement -> fetch_bm_ptr = (SQLUINTEGER*) value;
+        statement -> fetch_bm_ptr = (SQLULEN*) value;
         /*
          * pass on if required
          */
@@ -534,6 +780,14 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     value,
                     string_length );
             }
+            else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+            {
+                SQLSETSTMTATTRW( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    value,
+                    string_length );
+            }
             else
             {
                 ret = SQLSETSTMTOPTION( statement -> connection,
@@ -547,7 +801,7 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
     else if ( attribute == SQL_ATTR_ROWS_FETCHED_PTR &&
 	      statement -> connection -> driver_act_ver == SQL_OV_ODBC2 )
     {
-        statement -> row_ct_ptr = (SQLUINTEGER*) value;
+        statement -> row_ct_ptr = (SQLULEN*) value;
         /*
          * pass on if required
          */
@@ -561,6 +815,14 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                     value,
                     string_length );
             }
+            else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+            {
+                SQLSETSTMTATTRW( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    value,
+                    string_length );
+            }
             else
             {
                 ret = SQLSETSTMTOPTION( statement -> connection,
@@ -575,9 +837,9 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
             statement -> connection -> driver_act_ver == SQL_OV_ODBC2 )
     {
         /*
-         * save this incase we need it in SQLExtendedFetch
+         * save this in case we need it in SQLExtendedFetch
          */
-        statement -> row_array_size = (SQLUINTEGER) value;
+        statement -> row_array_size = (SQLULEN) value;
 
         if ( CHECK_SQLSETSTMTATTR( statement -> connection ))
         {
@@ -587,6 +849,14 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                 value,
                 string_length );
         }
+        else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+        {
+            ret = SQLSETSTMTATTRW( statement -> connection,
+                statement -> driver_stmt,
+                SQL_ROWSET_SIZE,
+                value,
+                string_length );
+        }
         else
         {
             ret = SQLSETSTMTOPTION( statement -> connection,
@@ -603,12 +873,40 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                 value,
                 string_length );
     }
-    else
+    else if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
     {
-        ret = SQLSETSTMTOPTION( statement -> connection,
+        ret = SQLSETSTMTATTRW( statement -> connection,
                 statement -> driver_stmt,
                 attribute,
-                value );
+                value,
+                string_length );
+    }
+    else
+    {
+        /*
+         * Is it in the legal range of values
+         */
+
+        if ( attribute < SQL_STMT_DRIVER_MIN && 
+                ( attribute > SQL_ROW_NUMBER || attribute < SQL_QUERY_TIMEOUT ))
+        {
+            dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY092" );
+
+            __post_internal_error( &statement -> error,
+                    ERROR_HY092, NULL,
+                    statement -> connection -> environment -> requested_version );
+
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        }
+
+        ret = SQLSETSTMTOPTION( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    value );
     }
 
     /*
@@ -617,14 +915,14 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
 
     if ( attribute == SQL_ATTR_USE_BOOKMARKS && SQL_SUCCEEDED( ret ))
     {
-        statement -> bookmarks_on = (SQLUINTEGER) value;
+        statement -> bookmarks_on = (SQLULEN) value;
     }
 
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -633,7 +931,5 @@ SQLRETURN SQLSetStmtAttr( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    thread_release( SQL_HANDLE_STMT, statement );
-
-    return function_return( statement, ret ); 
+    return function_return( SQL_HANDLE_STMT, statement, ret ); 
 }
