@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -27,9 +27,43 @@
  *
  **********************************************************************
  *
- * $Id: SQLSetStmtAttrW.c,v 1.1.1.1 2001/10/17 16:40:07 lurcher Exp $
+ * $Id: SQLSetStmtAttrW.c,v 1.10 2009/02/18 17:59:08 lurcher Exp $
  *
  * $Log: SQLSetStmtAttrW.c,v $
+ * Revision 1.10  2009/02/18 17:59:08  lurcher
+ * Shift to using config.h, the compile lines were making it hard to spot warnings
+ *
+ * Revision 1.9  2009/02/04 09:30:02  lurcher
+ * Fix some SQLINTEGER/SQLLEN conflicts
+ *
+ * Revision 1.8  2008/08/29 08:01:39  lurcher
+ * Alter the way W functions are passed to the driver
+ *
+ * Revision 1.7  2007/02/28 15:37:48  lurcher
+ * deal with drivers that call internal W functions and end up in the driver manager. controlled by the --enable-handlemap configure arg
+ *
+ * Revision 1.6  2006/04/18 10:24:47  lurcher
+ * Add a couple of changes from Mark Vanderwiel
+ *
+ * Revision 1.5  2003/10/30 18:20:46  lurcher
+ *
+ * Fix broken thread protection
+ * Remove SQLNumResultCols after execute, lease S4/S% to driver
+ * Fix string overrun in SQLDriverConnect
+ * Add initial support for Interix
+ *
+ * Revision 1.4  2003/03/05 09:48:45  lurcher
+ *
+ * Add some 64 bit fixes
+ *
+ * Revision 1.3  2002/12/05 17:44:31  lurcher
+ *
+ * Display unknown return values in return logging
+ *
+ * Revision 1.2  2002/05/28 13:30:34  lurcher
+ *
+ * Tidy up for AIX
+ *
  * Revision 1.1.1.1  2001/10/17 16:40:07  lurcher
  *
  * First upload to SourceForge
@@ -54,6 +88,7 @@
  *
  **********************************************************************/
 
+#include <config.h>
 #include "drivermanager.h"
 
 static char const rcsid[]= "$RCSfile: SQLSetStmtAttrW.c,v $";
@@ -66,6 +101,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
     DMHSTMT statement = (DMHSTMT) statement_handle;
     SQLRETURN ret;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
+	SQLWCHAR buffer[ 512 ];
 
     /*
      * check statement
@@ -79,6 +115,36 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                     LOG_INFO, 
                     "Error: SQL_INVALID_HANDLE" );
 
+#ifdef WITH_HANDLE_REDIRECT
+		{
+			DMHSTMT parent_statement;
+
+			parent_statement = find_parent_handle( statement, SQL_HANDLE_STMT );
+
+			if ( parent_statement ) {
+        		dm_log_write( __FILE__, 
+                	__LINE__, 
+                    	LOG_INFO, 
+                    	LOG_INFO, 
+                    	"Info: found parent handle" );
+
+				if ( CHECK_SQLSETSTMTATTRW( parent_statement -> connection ))
+				{
+        			dm_log_write( __FILE__, 
+                		__LINE__, 
+                   		 	LOG_INFO, 
+                   		 	LOG_INFO, 
+                   		 	"Info: calling redirected driver function" );
+
+                	return  SQLSETSTMTATTRW( parent_statement -> connection,
+							statement_handle,
+							attribute,
+							value,
+							string_length );
+				}
+			}
+		}
+#endif
         return SQL_INVALID_HANDLE;
     }
 
@@ -87,10 +153,10 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, "\n\t\tEntry:\
-            \n\t\t\tStatement = %p\
-            \n\t\t\tAttribute = %s\
-            \n\t\t\tValue = %p\
-            \n\t\t\tStrLen = %d",
+\n\t\t\tStatement = %p\
+\n\t\t\tAttribute = %s\
+\n\t\t\tValue = %p\
+\n\t\t\tStrLen = %d",
                 statement,
                 __stmt_attr_as_string( s1, attribute ),
                 value, 
@@ -129,9 +195,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                  ERROR_HY011, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
         else if ( statement -> state == STATE_S4 ||
                 statement -> state == STATE_S5 ||
@@ -148,15 +212,16 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                  ERROR_24000, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
         else if ( statement -> state == STATE_S8 ||
                 statement -> state == STATE_S9 ||
                 statement -> state == STATE_S10 ||
                 statement -> state == STATE_S11 ||
-                statement -> state == STATE_S12 )
+                statement -> state == STATE_S12 ||
+                statement -> state == STATE_S13 ||
+                statement -> state == STATE_S14 ||
+                statement -> state == STATE_S15 )
         {
             if ( statement -> prepared )
             {
@@ -170,9 +235,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                      ERROR_HY011, NULL,
                      statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
             else
             {
@@ -186,9 +249,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                      ERROR_HY010, NULL,
                      statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-
-                return function_return( statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
         }
     }
@@ -210,28 +271,45 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                  ERROR_HY010, NULL,
                  statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
-    if ( !CHECK_SQLSETSTMTATTRW( statement -> connection ))
-    {
-        dm_log_write( __FILE__, 
-                __LINE__, 
-                LOG_INFO, 
-                LOG_INFO, 
-                "Error: IM001" );
-
-        __post_internal_error( &statement -> error,
-                ERROR_IM001, NULL,
-                statement -> connection -> environment -> requested_version );
-
-        thread_release( SQL_HANDLE_STMT, statement );
-
-        return function_return( statement, SQL_ERROR );
-    }
+    if ( statement -> connection -> unicode_driver ||
+		    CHECK_SQLSETSTMTATTRW( statement -> connection ))
+	{
+    	if ( !CHECK_SQLSETSTMTATTRW( statement -> connection ))
+    	{
+        	dm_log_write( __FILE__, 
+                	__LINE__, 
+                	LOG_INFO, 
+                	LOG_INFO, 
+                	"Error: IM001" );
+	
+        	__post_internal_error( &statement -> error,
+                	ERROR_IM001, NULL,
+                	statement -> connection -> environment -> requested_version );
+	
+        	return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    	}
+	}
+	else
+	{
+    	if ( !CHECK_SQLSETSTMTATTR( statement -> connection ))
+    	{
+        	dm_log_write( __FILE__, 
+                	__LINE__, 
+                	LOG_INFO, 
+                	LOG_INFO, 
+                	"Error: IM001" );
+	
+        	__post_internal_error( &statement -> error,
+                	ERROR_IM001, NULL,
+                	statement -> connection -> environment -> requested_version );
+	
+        	return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    	}
+	}
 
     /*
      * map descriptors to our copies
@@ -241,6 +319,87 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
     {
         DMHDESC desc = ( DMHDESC ) value;
 
+   		/*
+		 * needs to reset to implicit descriptor, this is safe
+		 * without a validate, as the value is either null, or the
+		 * same as a descriptor we know is valid
+		 */
+
+		if ( desc == NULL || desc == statement -> implicit_ard ) 
+		{
+			DRV_SQLHDESC drv_desc = NULL;
+
+			ret = SQL_SUCCESS;
+			
+			if ( desc == statement -> implicit_ard )
+			{
+				drv_desc = statement -> implicit_ard -> driver_desc;
+			}
+
+        	if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+        	{
+           	 	ret = SQLSETSTMTATTRW( statement -> connection,
+           	     	statement -> driver_stmt,
+           	     	attribute,
+           	     	statement -> implicit_ard -> driver_desc,
+           	     	0 );
+        	}
+        	else if ( CHECK_SQLSETSTMTATTR( statement -> connection ))
+        	{
+            	ret = SQLSETSTMTATTR( statement -> connection,
+                	statement -> driver_stmt,
+                	attribute,
+                	drv_desc,
+                	0 );
+        	}
+            else
+            {
+                ret = SQLSETSTMTOPTION( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    statement -> implicit_ard -> driver_desc );
+            }
+
+			if ( ret != SQL_SUCCESS ) 
+			{
+    			if ( log_info.log_flag )
+    			{
+        			sprintf( statement -> msg, 
+                			"\n\t\tExit:[%s]",
+                    			__get_return_status( ret, s1 ));
+			
+        			dm_log_write( __FILE__, 
+                			__LINE__, 
+                			LOG_INFO, 
+                			LOG_INFO, 
+                			statement -> msg );
+    			}
+			
+    			return function_return( SQL_HANDLE_STMT, statement, ret ); 
+			}
+			
+			/*
+			 * copy DM descriptor
+			 */
+
+			statement -> apd = statement -> implicit_apd;
+
+    		if ( log_info.log_flag )
+    		{
+        		sprintf( statement -> msg, 
+                		"\n\t\tExit:[%s]",
+                    		__get_return_status( ret, s1 ));
+
+        		dm_log_write( __FILE__, 
+                		__LINE__, 
+                		LOG_INFO, 
+                		LOG_INFO, 
+                		statement -> msg );
+    		}
+		
+    		return function_return( SQL_HANDLE_STMT, statement, ret ); 
+		}
+
         if ( !__validate_desc( desc ))
         {
             thread_release( SQL_HANDLE_STMT, statement );
@@ -261,9 +420,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                     ERROR_HY017, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         if ( desc -> connection !=
@@ -279,9 +436,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                     ERROR_HY024, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         /*
@@ -289,17 +444,99 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
          */
         value = ( SQLPOINTER ) desc -> driver_desc;
         statement -> ard = desc;
+        desc -> associated_with = statement;
     }
 
     if ( attribute == SQL_ATTR_APP_PARAM_DESC )
     {
         DMHDESC desc = ( DMHDESC ) value;
 
+		/*
+		 * needs to reset to implicit descriptor, this is safe
+		 * without a validate, as the value is either null, or the
+		 * same as a descriptor we know is valid
+		 */
+
+		if ( desc == NULL || desc == statement -> implicit_apd ) 
+		{
+			DRV_SQLHDESC drv_desc = NULL;
+
+			ret = SQL_SUCCESS;
+			
+			if ( desc == statement -> implicit_apd )
+			{
+				drv_desc = statement -> implicit_apd -> driver_desc;
+			}
+
+        	if ( CHECK_SQLSETSTMTATTRW( statement -> connection ))
+        	{
+           	 	ret = SQLSETSTMTATTRW( statement -> connection,
+           	     	statement -> driver_stmt,
+           	     	attribute,
+                	statement -> implicit_apd -> driver_desc,
+           	     	0 );
+        	}
+        	else if ( CHECK_SQLSETSTMTATTR( statement -> connection ))
+        	{
+            	ret = SQLSETSTMTATTR( statement -> connection,
+                	statement -> driver_stmt,
+                	attribute,
+                	statement -> implicit_apd -> driver_desc,
+                	0 );
+        	}
+            else
+            {
+                ret = SQLSETSTMTOPTION( statement -> connection,
+                    statement -> driver_stmt,
+                    attribute,
+                    drv_desc );
+            }
+
+			if ( ret != SQL_SUCCESS ) 
+			{
+    			if ( log_info.log_flag )
+    			{
+        			sprintf( statement -> msg, 
+                			"\n\t\tExit:[%s]",
+                    			__get_return_status( ret, s1 ));
+			
+        			dm_log_write( __FILE__, 
+                			__LINE__, 
+                			LOG_INFO, 
+                			LOG_INFO, 
+                			statement -> msg );
+    			}
+			
+    			return function_return( SQL_HANDLE_STMT, statement, ret ); 
+			}
+			
+			/*
+			 * copy DM descriptor
+			 */
+
+			statement -> apd = statement -> implicit_apd;
+
+    		if ( log_info.log_flag )
+    		{
+        		sprintf( statement -> msg, 
+                		"\n\t\tExit:[%s]",
+                    		__get_return_status( ret, s1 ));
+
+        		dm_log_write( __FILE__, 
+                		__LINE__, 
+                		LOG_INFO, 
+                		LOG_INFO, 
+                		statement -> msg );
+    		}
+		
+    		return function_return( SQL_HANDLE_STMT, statement, ret ); 
+		}
+
         if ( !__validate_desc( desc ))
         {
             sprintf( statement -> msg, 
                     "\n\t\tExit:[%s]",
-                    __get_return_status( SQL_INVALID_HANDLE ));
+                    __get_return_status( SQL_INVALID_HANDLE, s1 ));
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -325,9 +562,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                     ERROR_HY017, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         if ( desc -> connection !=
@@ -343,9 +578,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                     ERROR_HY024, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            thread_release( SQL_HANDLE_STMT, statement );
-
-            return function_return( statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         /*
@@ -353,6 +586,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
          */
         value = ( SQLPOINTER ) desc -> driver_desc;
         statement -> apd = desc;
+        desc -> associated_with = statement;
     }
 
     /*
@@ -361,14 +595,54 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
 
     if ( attribute == SQL_ATTR_METADATA_ID )
     {
-        statement -> metadata_id = (int) value;
+#ifdef HAVE_PTRDIFF_T
+        statement -> metadata_id = (ptrdiff_t) value;
+#else
+        statement -> metadata_id = (SQLINTEGER) value;
+#endif
+    }
+
+    if ( attribute == SQL_ATTR_IMP_ROW_DESC || 
+        attribute == SQL_ATTR_IMP_PARAM_DESC )
+    {
+        dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY017" );
+
+        __post_internal_error( &statement -> error,
+                ERROR_HY017, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    /*
+     * is it a legitimate value
+     */
+    ret = dm_check_statement_attrs( statement, attribute, value );
+
+    if ( ret != SQL_SUCCESS ) 
+    {
+        dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: HY011" );
+
+        __post_internal_error( &statement -> error,
+                ERROR_HY024, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
      * is it something overridden
      */
 
-    value = __attr_override( statement, SQL_HANDLE_STMT, attribute, value, &string_length );
+    value = __attr_override_wide( statement, SQL_HANDLE_STMT, attribute, value, &string_length, buffer );
 
     /*
      * does the call need mapping from 3 to 2
@@ -379,7 +653,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
             CHECK_SQLEXTENDEDFETCH( statement -> connection ) &&
             !CHECK_SQLFETCHSCROLL( statement -> connection ))
     {
-        statement -> fetch_bm_ptr = (SQLINTEGER*) value;
+        statement -> fetch_bm_ptr = (SQLULEN*) value;
         ret = SQL_SUCCESS;
     }
     else if ( attribute == SQL_ATTR_ROW_STATUS_PTR &&
@@ -391,7 +665,7 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
     else if ( attribute == SQL_ATTR_ROWS_FETCHED_PTR &&
 	      statement -> connection -> driver_act_ver == SQL_OV_ODBC2 )
     {
-        statement -> row_ct_ptr = (SQLUINTEGER*) value;
+        statement -> row_ct_ptr = (SQLULEN*) value;
         ret = SQL_SUCCESS;
     }
     else if ( attribute == SQL_ATTR_ROW_ARRAY_SIZE &&
@@ -405,11 +679,26 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
     }
     else 
     {
-        ret = SQLSETSTMTATTRW( statement -> connection,
+    	if ( statement -> connection -> unicode_driver )
+		{
+        	ret = SQLSETSTMTATTRW( statement -> connection,
                 statement -> driver_stmt,
                 attribute,
                 value,
                 string_length );
+		}
+		else
+		{
+			/*
+			 * I can't find any string values, so we don't need to translate
+			 */
+
+        	ret = SQLSETSTMTATTR( statement -> connection,
+                statement -> driver_stmt,
+                attribute,
+                value,
+                string_length );
+		}
     }
 
     /*
@@ -418,14 +707,14 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
 
     if ( attribute == SQL_ATTR_USE_BOOKMARKS && SQL_SUCCEEDED( ret ))
     {
-        statement -> bookmarks_on = (SQLUINTEGER) value;
+        statement -> bookmarks_on = (SQLULEN) value;
     }
 
     if ( log_info.log_flag )
     {
         sprintf( statement -> msg, 
                 "\n\t\tExit:[%s]",
-                    __get_return_status( ret ));
+                    __get_return_status( ret, s1 ));
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -434,7 +723,5 @@ SQLRETURN SQLSetStmtAttrW( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    thread_release( SQL_HANDLE_STMT, statement );
-
-    return function_return( statement, ret ); 
+    return function_return( SQL_HANDLE_STMT, statement, ret ); 
 }
