@@ -1,12 +1,21 @@
 #ifndef _DRIVERMANAGER_H
 #define _DRIVERMANAGER_H
 
-#define ODBCVER 0x0351
+#define ODBCVER 0x0380
 
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_PWD_H
 #include <pwd.h>
+#endif
 #include <ltdl.h>
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
+#ifdef HAVE_TIME_H
+#include <time.h>
+#endif
 
 #ifdef HAVE_SYNCH_H
 #include <synch.h>
@@ -32,10 +41,39 @@
 #include <sqlucode.h>
 #include "__stats.h"
 
+/*
+ * iconv support
+ */
+
+#ifdef HAVE_ICONV
+#include <stdlib.h>
+#include <iconv.h>
+#endif
+
+#ifdef UNICODE_ENCODING
+#define DEFAULT_ICONV_ENCODING      UNICODE_ENCODING 
+#else
+#define DEFAULT_ICONV_ENCODING      "auto-search"
+#endif
+
 #define ERROR_PREFIX        "[unixODBC]"
 #define DM_ERROR_PREFIX     "[Driver Manager]"
 #define LOG_MESSAGE_LEN     128         /* length of string to display in log */
 
+/*
+ * SQLSetStmt/ConnectionAttr limits
+ */
+
+#define SQL_CONN_DRIVER_MIN     20000
+#define SQL_STMT_DRIVER_MIN     20000
+
+/*
+ * its possible that the driver has a different definition of a handle to the driver
+ * manager, DB2 64bit is a example of this
+ */
+
+#define DRV_SQLHANDLE			SQLHANDLE
+#define DRV_SQLHDESC			SQLHDESC
 
 /*
  * DEFAULT FILE NAMES
@@ -81,6 +119,10 @@
 #define STATE_S11       11
 #define STATE_S12       12
 
+#define STATE_S13       13          /* SQLExecute/SQLExecDirect/SQLMoreResult  returned SQL_PARAM_DATA_AVAILABLE */
+#define STATE_S14       14          /* SQL_PARAM_DATA_AVAILABLE version of S9, Must Get */
+#define STATE_S15       15          /* SQL_PARAM_DATA_AVAILABLE version of S10, Can Get */
+
 #define STATE_D0        0
 #define STATE_D1i       1
 #define STATE_D1e       2
@@ -111,7 +153,7 @@ typedef struct error
 {
     SQLWCHAR    sqlstate[ 6 ];
     SQLWCHAR    *msg;
-    int         native_error;
+    SQLINTEGER  native_error;
     int         return_val;
     SQLRETURN   diag_column_number_ret;
     SQLRETURN   diag_row_number_ret;
@@ -120,7 +162,7 @@ typedef struct error
     SQLRETURN   diag_connection_name_ret;
     SQLRETURN   diag_server_name_ret;
     SQLINTEGER  diag_column_number;
-    SQLINTEGER  diag_row_number;
+    SQLLEN      diag_row_number;
     SQLWCHAR    diag_class_origin[ 128 ];
     SQLWCHAR    diag_subclass_origin[ 128 ];
     SQLWCHAR    diag_connection_name[ 128 ];
@@ -146,18 +188,18 @@ typedef struct error_head
     EHEADER     sql_diag_head;
     void        *owning_handle;
     int         handle_type;
-    SQLINTEGER  return_code;
+    SQLRETURN   return_code;
     SQLINTEGER  header_set;
     SQLRETURN   diag_cursor_row_count_ret;
     SQLRETURN   diag_dynamic_function_ret;
     SQLRETURN   diag_dynamic_function_code_ret;
     SQLRETURN   diag_number_ret;
     SQLRETURN   diag_row_count_ret;
-    SQLINTEGER  diag_cursor_row_count;
+    SQLLEN      diag_cursor_row_count;
     SQLWCHAR    diag_dynamic_function[ 128 ];
     SQLINTEGER  diag_dynamic_function_code;
-    SQLINTEGER  diag_number;
-    SQLINTEGER  diag_row_count;
+    SQLLEN      diag_number;
+    SQLLEN      diag_row_count;
 } EHEAD;
 
 struct log_structure
@@ -165,6 +207,9 @@ struct log_structure
     char    *program_name;
     char    *log_file_name;
     int     log_flag;
+    int     pid_logging;            /* the log path specifies a directory, and a */
+                                    /* log file per pid is created */
+
 };
 
 extern struct log_structure log_info;
@@ -178,7 +223,7 @@ struct save_attr
     int                 attr_type;
     char                *str_attr;
     int                 str_len;
-    int                 int_attr;
+    SQLLEN              int_attr;
     struct save_attr    *next;
 };
 
@@ -207,7 +252,22 @@ int __parse_attribute_string( struct attr_struct *attr_str,
     char *str, int str_len );
 void __release_attr_str( struct attr_struct *attr_str );
 void __set_attributes( void *handle, int type );
+void __set_local_attributes( void *handle, int type );
 void *__attr_override( void *handle, int type, int attribute, void * value, SQLINTEGER *string_length );
+void *__attr_override_wide( void *handle, int type, int attribute, void * value, SQLINTEGER *string_length, SQLWCHAR *buffer );
+
+/*
+ * use this to maintain a list of the drivers that are loaded under this env,
+ * and to decide if we want to call SQLAllocHandle( SQL_ENV ) om them
+ */
+
+struct env_lib_struct
+{
+    char            *lib_name;
+    DRV_SQLHANDLE   env_handle;
+    int             count;
+    struct env_lib_struct   *next;
+};
 
 typedef struct environment
 {
@@ -215,6 +275,7 @@ typedef struct environment
     struct environment *next_class_list;/* static list of all env handles */
     char            msg[ LOG_MSG_MAX ];	/* buff to format msgs */
     int             state;              /* state of environment */
+    int             version_set;        /* whether ODBC version has been set */
     SQLINTEGER      requested_version;  /* SQL_OV_ODBC2 or SQL_OV_ODBC3 */
     int             connection_count;   /* number of hdbc of this env */
     int             sql_driver_count;   /* used for SQLDrivers */
@@ -224,8 +285,16 @@ typedef struct environment
     int             fetch_mode;         /* for SQLDataSources */
     int             entry;
     void            *sh;                /* statistics handle */
+    int             driver_act_ver;     /* real version of the driver */
+    struct env_lib_struct *env_lib_list;/* use this to avoid multiple AllocEnv in the driver */
 } *DMHENV;
 
+
+#ifdef FAST_HANDLE_VALIDATE
+    struct statement;
+#endif
+
+
 /*
  * connection pooling attributes
  */
@@ -234,10 +303,15 @@ typedef struct connection
 {
     int             type;               /* magic number */
     struct connection *next_class_list; /* static list of all dbc handles */
-    char            msg[ LOG_MSG_MAX ];	/* buff to format msgs */
+    char            msg[ LOG_MSG_MAX ]; /* buff to format msgs */
     int             state;              /* state of connection */
-    DMHENV          environment;         /* environment that own's the
+    DMHENV          environment;        /* environment that own's the
                                            connection */
+#ifdef FAST_HANDLE_VALIDATE
+    struct statement *statements;       /* List of statements owned by this 
+                                           connection */
+#endif
+    
     void            *dl_handle;         /* handle of the loaded lib */
     char            dl_name[ 256 ];     /* name of loaded lib */
     struct driver_func *functions;      /* entry points */
@@ -245,8 +319,8 @@ typedef struct connection
     struct driver_func  fini_func;
     int             unicode_driver;     /* do we use the W functions in the */
                                         /* driver ? */
-    SQLHANDLE       driver_env;         /* environment handle in client */
-    SQLHANDLE       driver_dbc;         /* connection handle in client */
+    DRV_SQLHANDLE 	driver_env;         /* environment handle in client */
+    DRV_SQLHANDLE   driver_dbc;         /* connection handle in client */
     int             driver_version;     /* required version of the connected */
                                         /* driver */
     int             driver_act_ver;     /* real version of the driver */
@@ -269,7 +343,7 @@ typedef struct connection
     int             metadata_id_set;
     int             packet_size;
     int             packet_size_set;
-    int             quite_mode;
+    SQLLEN          quite_mode;
     int             quite_mode_set;
     int             txn_isolation;
     int             txn_isolation_set;
@@ -304,10 +378,21 @@ typedef struct connection
     char            password[ 128 ];
     int             password_length;
     char            cli_year[ 5 ];
-    struct attr_struct  env_attribute;     /* Extended attribute set info */
+    struct attr_struct  env_attribute;      /* Extended attribute set info */
     struct attr_struct  dbc_attribute;
     struct attr_struct  stmt_attribute;
-    struct save_attr    *save_attr;          /* SQLConnectAttr before connect */
+    struct save_attr    *save_attr;         /* SQLConnectAttr before connect */
+#ifdef HAVE_ICONV
+    iconv_t         iconv_cd_uc_to_ascii;   /* in and out conversion descriptor */
+    iconv_t         iconv_cd_ascii_to_uc;
+    char            unicode_string[ 64 ];   /* name of unicode conversion */
+#endif
+    struct env_lib_struct *env_list_ent;    /* pointer to reference in the env list */
+    char            probe_sql[ 512 ];       /* SQL to use to check a pool is valid */
+	int				threading_level;		/* level of thread protection the DM proves */
+	int				cbs_found;				/* Have we queried the driver for the effect of a */
+	SQLSMALLINT		ccb_value;				/* COMMIT or a ROLLBACK */
+	SQLSMALLINT		crb_value;
 } *DMHDBC;
 
 typedef struct connection_pool
@@ -333,12 +418,18 @@ typedef struct descriptor
 {
     int             type;               /* magic number */
     struct descriptor *next_class_list; /* static list of all desc handles */
-    char            msg[ LOG_MSG_MAX ];	/* buff to format msgs */
+    char            msg[ LOG_MSG_MAX ]; /* buff to format msgs */
     int             state;              /* state of descriptor */
+
+#ifdef FAST_HANDLE_VALIDATE
+    struct descriptor *prev_class_list;/* static list of all desc handles */
+#endif    
+
     EHEAD           error;              /* keep track of errors */
-    SQLHDESC        driver_desc;        /* driver descriptor */
+    DRV_SQLHDESC    driver_desc;        /* driver descriptor */
     DMHDBC          connection;         /* DM connection that owns this */
     int             implicit;           /* created by a AllocStmt */
+    void            *associated_with;   /* statement that this is a descriptor of */
 #ifdef HAVE_LIBPTH
     pth_mutex_t     mutex;              /* protect the object */
 #elif HAVE_LIBPTHREAD
@@ -351,12 +442,17 @@ typedef struct descriptor
 typedef struct statement
 {
     int             type;               /* magic number */
-    struct statement *next_class_list; /* static list of all stmt handles */
-    char            msg[ LOG_MSG_MAX ];	/* buff to format msgs */
+    struct statement *next_class_list;  /* static list of all stmt handles */
+    char            msg[ LOG_MSG_MAX ]; /* buff to format msgs */
     int             state;              /* state of statement */
+#ifdef FAST_HANDLE_VALIDATE
+    struct statement *prev_class_list;  /* static list of all stmt handles */
+    struct statement *next_conn_list;   /* Single linked list storing statements 
+                                           owned by "connection" connection */
+#endif
     DMHDBC          connection;         /* DM connection that owns this */
-    SQLHANDLE       driver_stmt;        /* statement in the driver */
-    SQLSMALLINT     numcols;            /* Number of columns in the result set */
+    DRV_SQLHANDLE   driver_stmt;        /* statement in the driver */
+    SQLSMALLINT     hascols;            /* is there a result set */
     int             prepared;           /* the statement has been prepared */
     int             interupted_func;    /* current function running async */
                                         /* or NEED_DATA */
@@ -373,17 +469,22 @@ typedef struct statement
     DMHDESC         implicit_apd;
     DMHDESC         implicit_ird;
     DMHDESC         implicit_ard;
-    SQLUINTEGER     *fetch_bm_ptr;      /* Saved for ODBC3 to ODBC2 mapping */ 
-    SQLUINTEGER     *row_ct_ptr;        /* row count ptr */
+    SQLULEN		    *fetch_bm_ptr;      /* Saved for ODBC3 to ODBC2 mapping */ 
+    SQLULEN     	*row_ct_ptr;        /* row count ptr */
     SQLUSMALLINT    *row_st_arr;        /* row status array */
-    SQLUINTEGER     row_array_size;
+    SQLULEN     	row_array_size;
+	SQLPOINTER      valueptr;           /* Default buffer for SQLParamData() */
+
 #ifdef HAVE_LIBPTH
     pth_mutex_t     mutex;              /* protect the object */
 #elif HAVE_LIBPTHREAD
     pthread_mutex_t mutex;              /* protect the object */
 #elif HAVE_LIBTHREAD
     mutex_t mutex;              		/* protect the object */
+
 #endif
+
+    int             eod;                /* when in S6 has EOD been returned */
 } *DMHSTMT;
 
 #if defined ( HAVE_LIBPTHREAD ) || defined ( HAVE_LIBTHREAD ) || defined ( HAVE_LIBPTH )
@@ -394,9 +495,15 @@ typedef struct statement
 #define TS_LEVEL3   3           /* protection on a environment level */
 #endif
 
+void mutex_lib_entry( void );
+void mutex_lib_exit( void );
+
 void mutex_pool_entry( void );
 void mutex_pool_exit( void );
 
+void mutex_iconv_entry( void );
+void mutex_iconv_exit( void );
+
 typedef struct connection_pair
 {
     char            *name;
@@ -408,7 +515,7 @@ typedef struct connection_pair
  * defined down here to get the DMHDBC definition
  */
 
-void __handle_attr_extensions( DMHDBC connection, char *dsn );
+void __handle_attr_extensions( DMHDBC connection, char *dsn, char *driver_name );
 
 /*
  * handle allocation functions
@@ -423,6 +530,8 @@ int __validate_dbc( DMHDBC );
 void __release_dbc( DMHDBC connection );
 
 DMHSTMT __alloc_stmt( void );
+void __register_stmt ( DMHDBC connection, DMHSTMT statement );
+void __set_stmt_state ( DMHDBC connection, SQLSMALLINT cb_value );
 int __validate_stmt( DMHSTMT );
 void __release_stmt( DMHSTMT );
 
@@ -442,16 +551,29 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
            SQLHANDLE handle );
 
-int __connect_part_one( DMHDBC connection, char *driver_lib, char *driver_name );
+SQLRETURN __SQLGetInfo( SQLHDBC connection_handle,
+		 	SQLUSMALLINT info_type,
+			SQLPOINTER info_value,
+			SQLSMALLINT buffer_length,
+			SQLSMALLINT *string_length );
+
+int __connect_part_one( DMHDBC connection, char *driver_lib, char *driver_name, int *warnings );
 void __disconnect_part_one( DMHDBC connection );
 int __connect_part_two( DMHDBC connection );
 void __disconnect_part_two( DMHDBC connection );
+void __disconnect_part_three( DMHDBC connection );
+void __disconnect_part_four( DMHDBC connection );
 DMHDBC __get_dbc_root( void );
 
 void  __check_for_function( DMHDBC connection,
         SQLUSMALLINT function_id,
         SQLUSMALLINT *supported );
 
+int __clean_stmt_from_dbc( DMHDBC connection );
+int __clean_desc_from_dbc( DMHDBC connection );
+void __map_error_state( char * state, int requested_version );
+void __map_error_state_w( SQLWCHAR * wstate, int requested_version );
+
 /*
  * mapping from ODBC 2 <-> 3 datetime types
  */
@@ -481,21 +603,26 @@ typedef enum error_id
     ERROR_25000,
     ERROR_25S01,
     ERROR_S1000,
+    ERROR_S1003,
     ERROR_S1010,
     ERROR_S1011,
     ERROR_S1107,
     ERROR_S1108,
     ERROR_S1C00,
     ERROR_HY001,
+    ERROR_HY003,
     ERROR_HY004,
+    ERROR_HY007,
     ERROR_HY009,
     ERROR_HY010,
     ERROR_HY011,
+    ERROR_HY012,
     ERROR_HY013,
     ERROR_HY017,
     ERROR_HY024,
     ERROR_HY090,
     ERROR_HY092,
+    ERROR_HY095,
     ERROR_HY097,
     ERROR_HY098,
     ERROR_HY099,
@@ -505,6 +632,7 @@ typedef enum error_id
     ERROR_HY105,
     ERROR_HY106,
     ERROR_HY110,
+    ERROR_HY111,
     ERROR_HYC00,
     ERROR_IM001,
     ERROR_IM002,
@@ -515,30 +643,59 @@ typedef enum error_id
     ERROR_IM012,
     ERROR_SL004,
     ERROR_SL009,
-    ERROR_SL010
+    ERROR_SL010,
+    ERROR_SL008,
+    ERROR_HY000,
+    ERROR_IM011
 } error_id;
 
-#define function_return(h,r)    function_return_ex(h,r,FALSE)
+#define IGNORE_THREAD       (-1)
+
+#define function_return(l,h,r)    function_return_ex(l,h,r,FALSE)
+
+#define SUBCLASS_ODBC           0
+#define SUBCLASS_ISO            1
 
 void __post_internal_error( EHEAD *error_handle,
         error_id, char *txt, int connection_mode );
+void __post_internal_error_api( EHEAD *error_handle,
+        error_id, char *txt, int connection_mode, int calling_api );
 void __post_internal_error_ex( EHEAD *error_handle,
         SQLCHAR *sqlstate,
         SQLINTEGER native_error,
-        SQLCHAR *message_text );
+        SQLCHAR *message_text,
+        int class_origin,
+        int subclass_origin );
+void __post_internal_error_ex_noprefix( EHEAD *error_handle,
+        SQLCHAR *sqlstate,
+        SQLINTEGER native_error,
+        SQLCHAR *message_text,
+        int class_origin,
+        int subclass_origin );
 void __post_internal_error_ex_w( EHEAD *error_handle,
         SQLWCHAR *sqlstate,
         SQLINTEGER native_error,
-        SQLWCHAR *message_text );
-int function_return_ex( void * handle, int ret_code, int save_to_diag );
+        SQLWCHAR *message_text,
+        int class_origin,
+        int subclass_origin );
+void __post_internal_error_ex_w_noprefix( EHEAD *error_handle,
+        SQLWCHAR *sqlstate,
+        SQLINTEGER native_error,
+        SQLWCHAR *message_text,
+        int class_origin,
+        int subclass_origin );
+int function_return_nodrv( int level, void *handle, int ret_code );
+int function_return_ex( int level, void * handle, int ret_code, int save_to_diag );
 void function_entry( void *handle );
 void setup_error_head( EHEAD *error_header, void *handle, int handle_type );
 void clear_error_head( EHEAD *error_header );
-SQLWCHAR *ansi_to_unicode( SQLCHAR *str, SQLINTEGER len );
-SQLWCHAR *ansi_to_unicode_in_place( SQLCHAR *str, SQLINTEGER len );
-SQLWCHAR *ansi_back_to_unicode( char *str, SQLINTEGER len );
-char *unicode_to_ansi( SQLWCHAR *str, SQLINTEGER len );
-char * __get_return_status( SQLRETURN ret );
+SQLWCHAR *ansi_to_unicode_copy( SQLWCHAR * dest, char *src, SQLINTEGER buffer_len, DMHDBC connection, int *wlen );
+SQLWCHAR *ansi_to_unicode_alloc( SQLCHAR *str, SQLINTEGER len, DMHDBC connection, int *wlen );
+char *unicode_to_ansi_copy( char* dest, int dest_len, SQLWCHAR *src, SQLINTEGER len, DMHDBC connection, int *clen );
+char *unicode_to_ansi_alloc( SQLWCHAR *str, SQLINTEGER len, DMHDBC connection, int *clen );
+int unicode_setup( DMHDBC connection );
+void unicode_shutdown( DMHDBC connection );
+char * __get_return_status( SQLRETURN ret, SQLCHAR *buffer );
 char * __sql_as_text( SQLINTEGER type );
 char * __c_as_text( SQLINTEGER type );
 char * __string_with_length( SQLCHAR *out, SQLCHAR *str, SQLINTEGER len );
@@ -548,12 +705,14 @@ char * __wstring_with_length( SQLCHAR *out, SQLWCHAR *str, SQLINTEGER len );
 char * __wstring_with_length_pass( SQLCHAR *out, SQLWCHAR *str, SQLINTEGER len );
 char * __wstring_with_length_hide_pwd( SQLCHAR *out, SQLWCHAR *str, SQLINTEGER len );
 SQLWCHAR *wide_strcpy( SQLWCHAR *str1, SQLWCHAR *str2 );
+SQLWCHAR *wide_strncpy( SQLWCHAR *str1, SQLWCHAR *str2, int buffer_length );
 SQLWCHAR *wide_strcat( SQLWCHAR *str1, SQLWCHAR *str2 );
 SQLWCHAR *wide_strdup( SQLWCHAR *str1 );
 int wide_strlen( SQLWCHAR *str1 );
 int wide_ansi_strncmp( SQLWCHAR *str1, char *str2, int len );
 char * __get_pid( SQLCHAR *str );
-char * __ptr_as_string( SQLCHAR *s, SQLINTEGER *ptr );
+char * __iptr_as_string( SQLCHAR *s, SQLINTEGER *ptr );
+char * __ptr_as_string( SQLCHAR *s, SQLLEN *ptr );
 char * __sptr_as_string( SQLCHAR *s, SQLSMALLINT *ptr );
 char * __info_as_string( SQLCHAR *s, SQLINTEGER typ );
 void __clear_internal_error( struct error *error_handle );
@@ -561,6 +720,8 @@ char * __data_as_string( SQLCHAR *s, SQLINTEGER type,
         SQLLEN *ptr, SQLPOINTER buf );
 char * __sdata_as_string( SQLCHAR *s, SQLINTEGER type, 
         SQLSMALLINT *ptr, SQLPOINTER buf );
+char * __idata_as_string( SQLCHAR *s, SQLINTEGER type, 
+        SQLINTEGER *ptr, SQLPOINTER buf );
 char * __col_attr_as_string( SQLCHAR *s, SQLINTEGER type );
 char * __fid_as_string( SQLCHAR *s, SQLINTEGER fid );
 char * __con_attr_as_string( SQLCHAR *s, SQLINTEGER type );
@@ -570,8 +731,34 @@ char * __desc_attr_as_string( SQLCHAR *s, SQLINTEGER type );
 char * __diag_attr_as_string( SQLCHAR *s, SQLINTEGER type );
 char * __type_as_string( SQLCHAR *s, SQLSMALLINT type );
 DMHDBC __get_connection( EHEAD * head );
-SQLHANDLE __get_driver_handle( EHEAD * head );
+DRV_SQLHANDLE __get_driver_handle( EHEAD * head );
 int __get_version( EHEAD * head );
+int dm_check_connection_attrs( DMHDBC connection, SQLINTEGER attribute, SQLPOINTER value );
+int dm_check_statement_attrs( DMHSTMT statement, SQLINTEGER attribute, SQLPOINTER value );
+int __check_stmt_from_dbc( DMHDBC connection, int state );
+#define MAX_STATE_ARGS  8
+int __check_stmt_from_dbc_v( DMHDBC connection, int statecount, ... );
+int __check_stmt_from_desc( DMHDESC desc, int state );
+int __check_stmt_from_desc_ird( DMHDESC desc, int state );
+
+/* 
+ * These are passed to the cursor lib as helper functions
+ */
+
+struct driver_helper_funcs
+{
+    void (*__post_internal_error_ex)( EHEAD *error_header,
+            SQLCHAR *sqlstate,
+            SQLINTEGER native_error,
+            SQLCHAR *message_text,
+            int class_origin,
+            int subclass_origin );
+
+    void (*__post_internal_error)( EHEAD *error_handle,
+        error_id id, char *txt, int connection_mode );
+    void (*dm_log_write)( char *function_name, int line, int type, int severity,
+        char *message );
+};
 
 /*
  * thread protection funcs
@@ -591,6 +778,12 @@ void thread_release( int type, void *handle );
 
 void dbc_change_thread_support( DMHDBC connection, int level );
 
+#ifdef WITH_HANDLE_REDIRECT
+
+void *find_parent_handle( DRV_SQLHANDLE hand, int type );
+
+#endif
+
 /*
  * lookup functions
  */
@@ -601,7 +794,7 @@ char *__find_lib_name( char *dsn, char *lib_name, char *driver_name );
  * setup the cursor library
  */
 
-SQLRETURN CLConnect( DMHDBC connection );
+SQLRETURN SQL_API CLConnect( DMHDBC connection, struct driver_helper_funcs * );
 
 /*
  * connection string functions
@@ -631,27 +824,21 @@ void __release_conn( struct con_struct *con_str );
 void __get_attr( char ** cp, char ** keyword, char ** value );
 struct con_pair * __get_pair( char ** cp );
 int __append_pair( struct con_struct *con_str, char *kword, char *value );
+void __handle_attr_extensions_cs( DMHDBC connection, struct con_struct *con_str );
+void __strip_from_pool( DMHENV env );
 
 /*
  * the following two are part of a effort to get a particular unicode driver working
  */
 
-/*
- * set this if you are using a unicode driver that expects to get data in little 
- * endian order
- *
-#define BROKEN_UNICODE              1
- */
-
 SQLINTEGER map_ca_odbc3_to_2( SQLINTEGER field_identifier );
 SQLINTEGER map_ca_odbc2_to_3( SQLINTEGER field_identifier );
 
 /*
- * set this if your unicode driver is rearly ANSI
- * 
-#define BROKEN_UNICODE_SUPPORT      1
+ * check the type passed to SQLBindCol is a valid C_TYPE
  */
-#define BROKEN_UNICODE_SUPPORT      1
+
+int check_target_type( int c_type, int connection_mode);
 
 /*
  * entry exit functions in drivers
@@ -664,7 +851,7 @@ SQLINTEGER map_ca_odbc2_to_3( SQLINTEGER field_identifier );
  * driver manager logging functions
  */
 
-void dm_log_open( char *program_name, char *log_file );
+void dm_log_open( char *program_name, char *log_file, int pid_logging );
 
 void dm_log_write( char *function_name, int line, int type, int severity, char *message );
 void dm_log_write_diag( char *message );
@@ -847,9 +1034,10 @@ void return_to_pool( DMHDBC connection );
 #define SQLDRIVERCONNECT(con,dbc,wh,ics,sl1,ocs,bl,sl2p,dc)\
                                     (con->functions[22].func)\
                                         (dbc,wh,ics,sl1,ocs,bl,sl2p,dc)
+
 #define CHECK_SQLDRIVERCONNECTW(con) (con->functions[22].funcW!=NULL)
 #define SQLDRIVERCONNECTW(con,dbc,wh,ics,sl1,ocs,bl,sl2p,dc)\
-                                    (con->functions[22].funcW)\
+                                     (con->functions[22].funcW)\
                                         (dbc,wh,ics,sl1,ocs,bl,sl2p,dc)
 
 #define DM_SQLDRIVERS               23
@@ -866,7 +1054,7 @@ void return_to_pool( DMHDBC connection );
                                         (env,dbc,stmt,st,nat,msg,mm,pcb)
 #define CHECK_SQLERRORW(con)         (con->functions[25].funcW!=NULL)
 #define SQLERRORW(con,env,dbc,stmt,st,nat,msg,mm,pcb)\
-                                    (con->functions[25].funcW)\
+                                     (con->functions[25].funcW)\
                                         (env,dbc,stmt,st,nat,msg,mm,pcb)
 
 #define DM_SQLEXECDIRECT            26
@@ -1193,7 +1381,6 @@ void return_to_pool( DMHDBC connection );
 #define SQLSETSCROLLOPTIONS(con,stmt,fc,cr,rs)\
                                     (con->functions[69].func)\
                                         (stmt,fc,cr,rs)
-
 #define DM_SQLSETSTMTATTR           70
 #define CHECK_SQLSETSTMTATTR(con)   (con->functions[70].func!=NULL)
 #define SQLSETSTMTATTR(con,stmt,attr,vp,sl)\
@@ -1210,6 +1397,11 @@ void return_to_pool( DMHDBC connection );
                                     (con->functions[71].func)\
                                         (stmt,op,val)
 
+#define CHECK_SQLSETSTMTOPTIONW(con)   (con->functions[71].funcW!=NULL)
+#define SQLSETSTMTOPTIONW(con,stmt,op,val)\
+                                    (con->functions[71].funcW)\
+                                        (stmt,op,val)
+
 #define DM_SQLSPECIALCOLUMNS        72
 #define CHECK_SQLSPECIALCOLUMNS(con)    (con->functions[72].func!=NULL)
 #define SQLSPECIALCOLUMNS(con,stmt,it,cn,nl1,sn,nl2,tn,nl3,s,n)\
@@ -1266,4 +1458,10 @@ void return_to_pool( DMHDBC connection );
                                     (con->functions[77].funcW)\
                                         (typ,han,rn,st,nat,msg,bl,tlp)
 
+#define DM_SQLCANCELHANDLE          78
+#define CHECK_SQLCANCELHANDLE(con)  (con->functions[78].func!=NULL)
+#define SQLCANCELHANDLE(con,typ,han)\
+                                    (con->functions[78].func)\
+                                        (typ,han)
+
 #endif
